package edu.vub.crime;

import edu.vub.crime.grammar.attributes.datastructures.TypedObject;
import edu.vub.crime.util.Logging;

/** 
 * The abstract Action class can be subclassed by applications to respond to changes in the Fact Space.
 * Action subclass instances can be passed explicitly to Events when manually constructing rules from
 * Java. Alternatively they can be instantiated when their class name is used literally in a CRIME source
 * file.
 * 
 * @brief Action needs to be subclassed to attribute custom reactions to changes in the Fact Space.
 */
public abstract class Action {
	
	/** 
	 * This instance variable is initialized when the Action is tied to an Event instance, it points to
	 * the CRIME engine that will evaluate the rule to which it will respond.
	 */
	protected CrimeEngine itsEngine;
	
	/**
	 * This method is invoked whenever the prerequisites for triggering this action are met.
	 * @param arguments contains constants and/or variables that are passed from the rule.
	 */
	public abstract void activated(TypedObject arguments);
	
	
	/** 
	 * This method is invoked when certain preconditions for triggering this action no longer
	 * hold. The Action can then decide whether to reconsider the actions it performed previously.
	 * @param arguments contains the same combination values that were passed when Action was activated.
	 */
	public abstract void deactivate(TypedObject arguments);
		
	/** 
	 * Correctly sets the engine for this action, when tying the action to an Event.
	 */
	public void initializeEngine(CrimeEngine theEngine) {
		if(itsEngine == null) {
			itsEngine = theEngine;
		} else {
			Logging.Engine_LOG.fatal("Error initializing Action of class: " + this.getClass());
			Logging.Engine_LOG.fatal("Engine was already set while initializing.");
			throw new RuntimeException("Action Initialization: Engine was already set while initializing");
		}
	}
}