package edu.vub.crime;

import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.attributes.constant.Type;
import edu.vub.crime.grammar.attributes.datastructures.Pair;
import edu.vub.crime.grammar.attributes.datastructures.TypedObject;

/** <pre>
 * This method handles the asserting and retracting of a QualifiedFact 
 * (a fact which belongs to a specific fact space).
 * Such a fact is added to the network (when the rule, which contains
 * this action as part of its left-hand-side, is triggered), and removed
 * when the conditions of that rule are no longer met.
 * In that case, all actions in the left-hand-side of the rule should
 * be reversed.
 * </pre> */
public class AssertQualifiedFact extends Action {

	// A field name which is not legal in pure CRIME is used to store 
	// the name of the space in which the fact should be published.
	public  static final String SPACE_NAME  = ":Space"; 
	private static final Vector SPACE_INDEX = new Vector();
	
	static { SPACE_INDEX.add(SPACE_NAME); }
	
	/** <pre>
	 * This method is called when the right-hand-side of the rule
	 * (from which this action is part of its left-hand-side) is 
	 * satisfied.  This method will cause the asserting of a fact,
	 * whose attributes are given as arguments to this method,
	 * to the network.
	 * </pre> */
	public void activated(TypedObject arguments) {
		createFact(arguments).assert(itsEngine);
	}

	/** <pre>
	 * This method will cause the removal of a fact.
	 * The arguments of that fact are given as arguments.
	 * </pre> */
	public void deactivate(TypedObject arguments) {
		createFact(arguments).retract(itsEngine);
	}

	/** <pre>
	 * Create a fact, given its arguments.
	 * </pre> */
	private GenericFact createFact(TypedObject arguments) {
		try {
			Type name = (Type) arguments.getType();
			Type space = (Type) arguments.getAttribute(GenericFact.SPACE_IDX.iterator());
			GenericFact fact = ExpressionFactory.qualifiedFact(space.toString(), name.toString());
			
			for (Iterator attributeIt 		= arguments.iterator(); attributeIt.hasNext();) {
				Pair 	  currentAttribute	= (Pair) attributeIt.next();
				String 	  currentName		= currentAttribute.getName();
				
				if(!currentName.equals(SPACE_NAME))
					fact.addNamedAttribute(currentName, currentAttribute.getValue());
			}

			return fact;
		} catch (IncorrectAttributeIndexException e) {
			throw new RuntimeException("Could not construct the fact to be asserted", e);
		}
	}
}