package edu.vub.crime;

import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.Rule;
import edu.vub.crime.memory.Memory;
import edu.vub.crime.network.ConnectedCrimeEngine;
import edu.vub.crime.rete.CompileNetworkVisitor;
import edu.vub.crime.rete.agenda.Activation;
import edu.vub.crime.rete.agenda.Agenda;
import edu.vub.crime.rete.agenda.Scheduler;
import edu.vub.crime.rete.data.DataFactory;
import edu.vub.crime.rete.nodes.NodesFactory;
import edu.vub.crime.rete.nodes.RootNode;
import edu.vub.crime.util.Logging;

/** <pre>
 * The CrimeEngine is the actual engine which will
 * built and maintain the Rete network.
 * The rule which is used for explanation is : 
 *
 *         likeMyMusic(?genre, ?number) :-
 *              genreILike(?genre),
 *              findall(?person, (
 *                         isPresent(?person, LivingRoom),
 *                         prefers(?person, ?genre)),
 *                      ?similarTastes),
 *              length(?similarTastes, ?number).
 *              
 * In the example, we consider that I like rock and pop.
 * Alice likes rock and Bob likes rock and pop, so the memory contains four facts : 
 *        __________________________
 *        | MEMORY                 |
 *        |========================|
 *        | genreILike(Rock)       |
 *        | genreILike(Pop)        |
 *        | prefers(Alice, Rock)   |
 *        | prefers(Bob, Rock)     |
 *        | prefers(Bob, Pop)      |
 *        |________________________|
 *  
 *  This engine also inserts tokens, formed from these facts, in the Rete network.
 *  This is realised by inserting the token in the root node of the network,
 *  which has a reference in this engine. 
 *  The Rete network at this moment is depicted below.
 *  
 *         *     FILTER: genreILike           FILTER: isPresent           FILTER: prefers
 *             |                             |                          |
 *              \                    FILTER: LivingRoom                 |
 *               \                           |                          |
 *                \                           \                        /
 *                 \                           \                      /   
 *                  \             LEFT MEMORY   JOIN: ?person = ? person     RIGHT MEMORY 
 *                   \            ===========            |                   ============
 *                    \                                 /                    Alice, Rock
 *                     \                               /                     Bob, Rock
 *                      \                             /                      Bob, Pop
 *         LEFT MEMORY       JOIN: ?genre = ?genre      RIGHT MEMORY
 *         ===========                |                 ============
 *         Rock                       |
 *         Pop                     FINDALL       CACHE       
 *                                    |          =====
 *                                    |          Rock, [Alice, Bob] 
 *                                    |          Pop, [Bob]
 *                                    |
 *                                  LENGTH
 *                                    | <add, genreILike(Rock), similarTastes({Alice, Bob}), number(2)>
 *                                    | <add, genreILike(Pop), similarTastes({Bob}0, number(1)>
 *                                    |
 *                                PRODUCTION      
 * 
 * When the token < add, genreILike(Rock), similarTastes({Alice, Bob}), number(2) > is inserted
 * into this ProductionNode, a new Activation will be added to the agenda.
 * The agenda contains all actions that still need to be executed by the engine.
 * This activation consists of the actions of the left-hand-side of the rule whose conditions are met.
 * In the example, there's only one action namely the assertion of a fact likeMyMusic(Rock, 2).
 * So this action is added to an activation which is added to the agenda.
 * The agenda is depicted below (supposed that the agenda was empty before the addition of this activation):
 *        ______________________________
 *        | AGENDA                     |
 *        |============================|
 *        | likeMyMusic(Rock, 2)       |
 *        |____________________________|
 *        
 * When the second token < add, genreILike(Pop), similarTastes({Bob}0, number(1) > is added to the engine
 * and passed through the network, the ProductionNode will add a new Activation to the agenda.
 *        ______________________________
 *        | AGENDA                     |
 *        |============================|
 *        | likeMyMusic(Rock, 2)       |
 *        | likeMyMusic(Pop, 1)        |
 *        |____________________________|
 *        
 * A specified scheduler will determine in which order these activations in the agenda will
 * be executed.
 * </pre> */
public class CrimeEngine implements RuleBase {

	/** <pre>
	 * rootNode_ : the root node of the rete network
	 *             Every token which will be passed through the network,
	 *             is inserted to this node.
	 * agenda_ : contains all Activations (actions of a rule) that still 
	 *           need to be executed
	 * scheduler_ : determines the order in which the Activations of the agenda
	 *              will be executed
	 * busy_ : boolean (true when busy, false otherwise)
	 * memory_ : contains all the facts which are inserted (as tokens) in the network
	 * packageName_ : the current package to search for actions
	 * currentTime : the current time
	 * </pre> */
	public static String PACKAGE_DECL;
	protected	RootNode	itsRootNode;
	protected	Agenda		itsAgenda;
	protected	Scheduler	itsScheduler;
	private 	Memory		itsMemory;

    public CrimeEngine() {
		this(NodesFactory.rootNode(), CrimeFactory.createAgenda(), CrimeFactory.createMemory(), true);
	}
	
	public CrimeEngine(Agenda theAgenda, boolean shouldStartScheduler) {
		this(NodesFactory.rootNode(), theAgenda, CrimeFactory.createMemory(), shouldStartScheduler);		
	}

	public CrimeEngine(boolean shouldStartScheduler) {
		this(NodesFactory.rootNode(), CrimeFactory.createAgenda(), CrimeFactory.createMemory(), shouldStartScheduler);		
	}
	
	private CrimeEngine(RootNode theRootNode, Agenda theAgenda, Memory theMemory, boolean shouldStartScheduler) {
		itsRootNode = theRootNode;
		itsAgenda	= theAgenda;
		itsMemory	= theMemory;

		if(shouldStartScheduler) 
			itsScheduler = CrimeFactory.createScheduler(itsAgenda, this);
		else
			itsScheduler = null;
	}
	
	/**
	 * Asserting a fact will create a new (positive) token that will be
	 * subsequently inserted in the root node of the RETE network.
	 */
	public void assert(GenericFact theFact) {
		Logging.Engine_LOG.info("assert " + theFact.toString());
		
		itsMemory.insert(theFact);
		itsRootNode.insert(DataFactory.addToken(theFact));
	}
	
	/**
	 * Asserting a fact in a public fact space will not only assert the fact
	 * locally, but will also cause nearby CRIME engines to be notified of the
	 * presence of the fact.
	 * <p>
	 * Note that this engine is not a connected one, hence the implementation
	 * of this method will simply assert the fact locally.
	 * 
	 * @see ConnectedCrimeEngine#assertShared(GenericFact) for a networked 
	 * implementation which will propagate the facts to nearby engines.
	 */
	public void assertShared(GenericFact theFact) {
		// Since this is not a connected engine simply forward to the default
		// semantics of assertion for a fact.
		assert(theFact);
	}
	
	/**
	 * Retracting a fact will create a new (negative) token that will be
	 * subsequently inserted in the root node of the RETE network.
	 */
	public void retract(GenericFact theFact) {
		Logging.Engine_LOG.info("retract " + theFact.toString());

		itsRootNode.insert(DataFactory.removeToken(theFact));
	}

	/**
	 * Retracting a fact from a public fact space will not only retract the fact
	 * locally, but will also affect nearby CRIME engines, causing then to retract
	 * the fact as well.
	 * <p>
	 * Note that this engine is not a connected one, hence the implementation
	 * of this method will only retract the fact locally.
	 * 
	 * @see ConnectedCrimeEngine#retractShared(GenericFact) for a networked 
	 * implementation which will contact nearby engines.
	 */
	public void retractShared(GenericFact theFact) {
		retract(theFact);
	}
	
	/** 
	 * Deploys a rule, which implies compiling it so that its prerequisites are compiled
	 * to a rete network where each fact is compiled to a filter node which are joined
	 * one by one using join nodes. The consequence part of the rule is then compiled to
	 * a single production node which is triggered by the rete network representing the 
	 * prerequisites of the rule.
	 * <p>
	 * The compilation of a rule is performed by the {@link CompileNetworkVisitor}. 
	 */
	public void deploy(Rule rule) {
		rule.accept(new CompileNetworkVisitor(this, itsRootNode));
	}

	/** <pre>
	 * A future request, not yet implemented,
	 * the general idea is to remove the rule but also
	 * remove all it's derivations. 
	 * The later is not that easy for the moment.
	 * </pre> */
	public void undeploy(Rule theRule) {
		itsRootNode.undeploy(theRule);
	}
	
	/** 
	 * This hook method is called before an activation is fired, allowing derived classes
	 * to determine the boundaries of every activation, without having to pass this as a 
	 * parameter to methods such as <tt>assert</tt> and <tt>retract</tt> explicitly.
	 * 
	 * @param theActivation the activation which will be fired
	 */
	public void beginActivation(Activation theActivation) {

	}
	
	/** 
	 * This hook method is called after an activation has been fired, allowing derived 
	 * classes to determine the boundaries of every activation, without having to pass this 
	 * as a parameter to methods such as <tt>assert</tt> and <tt>retract</tt> explicitly.
	 * 
	 * @param theActivation the activation which has been completed
	 */
	public void endActivation(Activation theActivation) {
		
	}
	
	/** <pre>
	 * When starting up the engine
	 * Activations are halted until all rules are parsed.
	 * Then all activations in the agenda will be loaded
	 * this is done with the load_all method
	 * </pre> */
	public void load_all() {
//		if(! scheduler_.isAlive())
//			scheduler_.start();
	}
	
	/** 
	 * Schedules the activation in the Agenda to be handled by the Scheduler
	 */
	public void insertActivation(Activation theActivation) {
		itsAgenda.schedule(theActivation);	       
	}

	public Memory getMemory() {
		return itsMemory;
	}
	
	public Agenda getAgenda() {
		return itsAgenda;
	}
	
	public RootNode getRootNode() {
		return itsRootNode;
	}
	
}