package edu.vub.crime;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Properties;

import edu.vub.crime.memory.Memory;
import edu.vub.crime.network.NetworkLayer;
import edu.vub.crime.rete.agenda.Agenda;
import edu.vub.crime.rete.agenda.Scheduler;
import edu.vub.crime.util.Logging;

public class CrimeFactory {
	
	public static final Properties factoryClasses; 
	
	static {
		// intialize the Logging API
		factoryClasses = new Properties();
		
		try {
			factoryClasses.load(CrimeFactory.class.getResourceAsStream("crime.props"));
		} catch (IOException e) {
			System.err.println("Could not read properties file: " + e.getMessage());
		}
	}
	
	private static Object genericInitialise(String type, Class[] constructorParamTypes, Object[] constructorArguments) {
		String 	qualifiedClassName = (String) factoryClasses.get(type);
		
		try {
			Class engineClass = Class.forName(qualifiedClassName);
			
			return engineClass.getConstructor(constructorParamTypes).newInstance(constructorArguments);
		} catch (ClassNotFoundException e) {
			Logging.Engine_LOG.fatal(type + " class (" + qualifiedClassName + ") does not exist", e);
			throw new RuntimeException(type + " class (" + qualifiedClassName + ") does not exist");
		} catch (NoSuchMethodException e) {
			Logging.Engine_LOG.fatal(type + " class (" + qualifiedClassName + ") does not have a constructor with the given arguments", e);
			throw new RuntimeException(type + " class (" + qualifiedClassName + ") does not have a constructor with the given arguments");
		} catch (InstantiationException e) {
			Logging.Engine_LOG.fatal(type + " class (" + qualifiedClassName + ") could not be instantiated", e);
			throw new RuntimeException(type + " class (" + qualifiedClassName + ") could not be instantiated");
		} catch (IllegalAccessException e) {
			Logging.Engine_LOG.fatal(type + " class (" + qualifiedClassName + ") does not have a suitable public constructor", e);
			throw new RuntimeException(type + " class (" + qualifiedClassName + ") does not have a suitable public constructor");
		} catch (IllegalArgumentException e) {
			Logging.Engine_LOG.fatal(type + " class (" + qualifiedClassName + ") constructor was supplied with illegal arguments", e);
			throw new RuntimeException(type + " class (" + qualifiedClassName + ") constructor was supplied with illegal arguments");
		} catch (SecurityException e) {
			Logging.Engine_LOG.fatal(type + " class (" + qualifiedClassName + ") constructor was not accessible", e);
			throw new RuntimeException(type + " class (" + qualifiedClassName + ") constructor was not accessible");
		} catch (InvocationTargetException e) {
			Logging.Engine_LOG.fatal(type + " class (" + qualifiedClassName + ") constructor was not accessible", e);
			throw new RuntimeException(type + " class (" + qualifiedClassName + ") constructor was not accessible");
		}		
	}
	
	public static CrimeEngine createEngine() {
		return (CrimeEngine) genericInitialise("Engine", new Class[0], new Object[0]);
	}
	
	public static NetworkLayer createNetworkLayer(CrimeEngine theEngine) {
		return (NetworkLayer) genericInitialise("Network", new Class[] { CrimeEngine.class }, new Object[] { theEngine });
	}
	
	public static Memory createMemory() {
		return (Memory) genericInitialise("Memory", new Class[0], new Object[0]);
	}
	public static Agenda createAgenda() {
		return (Agenda) genericInitialise("Agenda", new Class[0], new Object[0]);
	}

	public static Scheduler createScheduler(Agenda theAgenda, CrimeEngine theEngine) {
		return (Scheduler) genericInitialise("Scheduler", new Class[] { Agenda.class, CrimeEngine.class }, new Object[] { theAgenda, theEngine });
	}
}
