package edu.vub.crime;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;

import antlr.CommonAST;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.parser.EvaluationVisitor;
import edu.vub.crime.parser.LexerImpl;
import edu.vub.crime.parser.ParserImpl;
import edu.vub.crime.parser.TreeWalkerImpl;

/** 
 * Starts a command line crime engine with a given input file.
 * Whether a networked version of the engine is started depends
 * on the arguments given in the crime.props property file.
 */
public class Main extends Thread {	

	private static	final String		_STOP_ = "/q";	
	private 		final String		itsInputFile;
	private 		final CrimeEngine	itsReteEngine;
	
	private Main(String theInputFile) {
		itsInputFile	= theInputFile;
		itsReteEngine	= CrimeFactory.createEngine();
    }
	
	/** 
	 * Parse an input string to a corresponding {@link ExpressionList}.
	 */
	public static ExpressionList parse(String input) {
		StringReader reader = new StringReader(input);
		return parse(reader);
	}
	
	/** 
	 * Parse input from a {@link Reader} to a corresponding {@link ExpressionList}.
	 */
	public static ExpressionList parse(Reader theInputReader) {
		ExpressionList theParsedInput	= new ExpressionList();

		try {
			LexerImpl	crimeLexer		= new LexerImpl(theInputReader);
			ParserImpl	crimeParser		= new ParserImpl(crimeLexer);
			
			crimeParser.program();
			
			CommonAST theParsedAST		= (CommonAST)crimeParser.getAST();
			TreeWalkerImpl crimeWalker	= new TreeWalkerImpl();
			theParsedInput				= crimeWalker.program(theParsedAST);
		} catch(Exception e) {
			e.printStackTrace();
		}
		
		return theParsedInput;
	}

	/** 
	 * Main loop: starts the evaluation by parsing the input file and feeding it to
	 * the CRIME engine. Subsequently runs the commandline function,  which prompts
	 * the user for additional facts. 
	 */
	public void run() {
		try {
			InputStreamReader fileReader	= new FileReader(itsInputFile);
			parse(fileReader).accept(new EvaluationVisitor(itsReteEngine, true));
			
			itsReteEngine.load_all();
			
			commandLine();		
		} 
		catch(Exception e) {
    		e.printStackTrace();
		}
	}

	/**
	 * Gives the user the ability to enter expressions directly through the command line.
	 * There is not kept a history of commands and even the left and right arrows want work
	 * A lot of improvement to the command line interface could be done here.
	 */
	public void commandLine() {
		BufferedReader		theReader	= new BufferedReader(new InputStreamReader(System.in));

		while(true) {
			try {
				System.out.print(">> ");
				
				String 		theInput	= theReader.readLine();
				
				if(theInput.equals(_STOP_)) System.exit(0);
				
				do_(theInput, itsReteEngine);
			} catch (Exception e) {
				e.printStackTrace();
			} 
		}
	}

	/**
	 * Parse an input string and feed it to the engine.
	 */
	public static void do_(String input, CrimeEngine itsReteEngine) {
		ExpressionList theParsedInput = parse(input);
		
		// Evaluate the input, which may lead to the assertion of facts or the introduction of new rules.
		theParsedInput.accept(new EvaluationVisitor(itsReteEngine, true));
	}
	
	/** <pre>
	 * Retract the fact that is given in it's string representation.
	 * </pre> */
	public static void undo_(String input, CrimeEngine itsReteEngine) {
		ExpressionList theParsedInput = parse(input);
		
		// Unevaluate the input which may lead to the retraction of facts and the removal of rules.
		theParsedInput.accept(new EvaluationVisitor(itsReteEngine, false));
	}
	
	/** 
	 * Starts a command-line crime engine.
	 */
	public static void main(String args[]) {
		String theInputFile = args[0];
		Main theCrimeEngine = new Main(theInputFile);
		
		theCrimeEngine.start();
	}	
}