package edu.vub.crime;

import edu.vub.crime.grammar.Event;
import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.Rule;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.parser.CrimeParser;
import edu.vub.crime.parser.EvaluationVisitor;
import edu.vub.crime.parser.IndexingVisitor;
import edu.vub.crime.rete.agenda.InlineScheduler;
import edu.vub.crime.rete.agenda.UserActivation;
import edu.vub.crime.rete.filter.FilterFactory;
import edu.vub.crime.rete.nodes.FilterNode;
import edu.vub.crime.rete.nodes.NodesFactory;
import edu.vub.crime.rete.nodes.aux.TestFlowAttribute;
import edu.vub.crime.rete.nodes.aux.TestNode;

/**
 * This unit test verifies the correct behavior of standard Crime rules, using an inline
 * scheduler, which provides precise control over when activations are executed. 
 * 
 * @see ThreadedRuleTest a counterpart unit test which uses instrumentation and a threaded 
 * scheduler to perform similar tests.
 * 
 * @author smostinc
 */
public class SteppingRuleTest extends CrimeTestCase {

	private Rule compileRule(String theRuleText) {
		return (Rule) CrimeParser.parse(theRuleText).getExpression(0);
	}
	
	/**
	 *  Tests the correctness of a rule which needs to assert a single private fact when triggered.
	
	 */
	public void testPrivateAssertingRule() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the required fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("curriculum");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{ // Install a new rule in the Crime engine
			Main.do_(
				"curriculum(name = ?student, degree = ComputerScience, curriculum = ?curriculum) :-" +
				"  bagof(?course, [ attending(student = ?student, course = Course( title = ?course )) ], ?curriculum).",
			
				itsEngine);
		}
		
		{	// Assert a single fact and verify whether the rule triggers
			
			// Reset the token counters of the test node
			theNodeToBeTriggered.reset();
			
			// Compile the following assertion an create a user activation which is scheduled for execution in the agenda
			Main.do_("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).", itsEngine);
			
			// Execute the assertion of the fact.
			assertTrue(itsScheduler.step());
			
			// Verify that the rule triggered and an activation was placed in the agenda.
			assertTrue(itsScheduler.step());
			
			// Verify that the fact was correctly added by testing whether the test node was triggered (1 assertion, 0 retractions)
			theNodeToBeTriggered.assertTriggered(1, 0);
		}
		
		{	// Retract the same fact and verify whether the derived fact is properly retracted as well
			
			// Reset the token counters of the test node
			theNodeToBeTriggered.reset();

			// Compile the following retraction an create a user activation which is scheduled for execution in the agenda
			Main.undo_("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).", itsEngine);
			
			// Execute the retraction of the fact.
			assertTrue(itsScheduler.step());
			
			// Verify that the rule triggered (to undo its consequent) and an activation was placed in the agenda.
			assertTrue(itsScheduler.step());
			
			// Verify that the fact was correctly removed by testing whether the test node was triggered (0 assertions, 1 retraction)
			theNodeToBeTriggered.assertTriggered(0, 1);
		}
	}
	
	public void testPublicAssertingRule() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the required fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("curriculum");
			
			FilterNode theSpaceFilter = NodesFactory.filterNode();
			theSpaceFilter.addNewFilter(FilterFactory.filterSpace("vub"));
			
			theFilterNode.addChild(theSpaceFilter);
			theSpaceFilter.addChild(theNodeToBeTriggered);
		}
		
		{ // Install a new rule in the Crime engine
			Main.do_(
				"vub -> curriculum(name = ?student, degree = ComputerScience, curriculum = ?curriculum) :-" +
				"  bagof(?course, [ attending(student = ?student, course = Course( title = ?course )) ], ?curriculum).",
			
				itsEngine);
		}
		
		{	// Assert a single fact and verify whether the rule triggers
			
			// Reset the token counters of the test node
			theNodeToBeTriggered.reset();
			
			// Compile the following assertion an create a user activation which is scheduled for execution in the agenda
			Main.do_("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).", itsEngine);
			
			// Execute the assertion of the fact.
			assertTrue(itsScheduler.step());
			
			// Verify that the rule triggered and an activation was placed in the agenda.
			assertTrue(itsScheduler.step());
			
			// Verify that the fact was correctly added by testing whether the test node was triggered (1 assertion, 0 retractions)
			theNodeToBeTriggered.assertTriggered(1, 0);
		}
		
		{	// Retract the same fact and verify whether the derived fact is properly retracted as well
			
			// Reset the token counters of the test node
			theNodeToBeTriggered.reset();

			// Compile the following retraction an create a user activation which is scheduled for execution in the agenda
			Main.undo_("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).", itsEngine);
			
			// Execute the retraction of the fact.
			assertTrue(itsScheduler.step());
			
			// Verify that the rule triggered (to undo its consequent) and an activation was placed in the agenda.
			assertTrue(itsScheduler.step());
			
			// Verify that the fact was correctly removed by testing whether the test node was triggered (0 assertions, 1 retraction)
			theNodeToBeTriggered.assertTriggered(0, 1);
		}
	}
	
	
	/**
	 * This method tests the correct behavior of a rule which uses a combination of findall and length
	 * to count the number of users that are present in a particular room.
	 * <p>
	 * This test method uses a findall operator without any preceding query.
	 */
	public void testLiteralFindAllRule() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the required fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("how_crowded");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{	// Set up the working memory such that we have a few facts matching the rule we will define
			
			// Provide some input facts for the test.
			Main.do_("location(user = Stijn, 		location = Lounge).",	itsEngine);
			Main.do_("location(user = Tom, 			location = Lounge).",	itsEngine);
			Main.do_("location(user = Wolf, 		location = Bench).",	itsEngine);
			Main.do_("location(user = Andoni, 		location = Cockpit).",	itsEngine);
			Main.do_("location(user = Christophe, 	location = Bench).",	itsEngine);
			Main.do_("location(user = Elisa, 		location = Cockpit).",	itsEngine);
			
			// Perform all assertions
			while(itsScheduler.step());
		}
		
		{	// Define a rule which passes a variable from its prerequisite to the conclusion
			
			// Define a rule which counts the number of people in my current location
			Main.do_(
					"how_crowded(no_users = ?count, location = Lounge) :-" +
					"  findall(?name, [ location(user = ?name, location = Lounge) ], ?users)," +
					"  length(?users, ?count).", 
					
					itsEngine);
			
			// Note that upon installing the rule different tokens will be propagated to the findall 
			// node, who will in response first propagate a token containing a list with a match, and 
			// subsequently retract this fact to propagate an updated fact with a list of two users 
			// in the Lounge. The agenda will filter out complementary activations, such that only 
			// the last assertion remains active.
			
			// Verify that the rule has triggered
			assertTrue(itsScheduler.step());
			
			// verify that only a single activation was scheduled
			assertFalse(itsScheduler.step());
			
			// Consequently only a single fact was asserted
			theNodeToBeTriggered.assertTriggered(1, 0);
		}
		
		{	// Remove a fact to test the retract/assert performed by the findall node
			
			// Reset the test node
			theNodeToBeTriggered.reset();
			
			Main.undo_("location(user = Stijn, 		location = Lounge).",	itsEngine);

			// Perform the retraction of the location fact
			assertTrue(itsScheduler.step());
			
			// Perform the retraction of the aggregate fact.
			assertTrue(itsScheduler.step());
			
			// Ensure that the retraction of the of fact is indeed performed first 
			theNodeToBeTriggered.assertTriggered(0, 1);

			// Perform the assertion of the new aggregate fact.
			assertTrue(itsScheduler.step());
			
			// Ensure that both retraction and assertion are performed correctly
			theNodeToBeTriggered.assertTriggered(1, 1);
		}
	}
	
	
	/**
	 * This method tests the correct behavior of a rule which uses a combination of findall and length
	 * to count the number of users that are present in a particular room.
	 * <p>
	 * This test method uses a findall operator whose query is parametrized by variables bound by a
	 * preceding query.
	 */
	public void testParametrizedFindAllRule() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the required fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("how_crowded");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{	// Set up the working memory such that we have a few facts matching the rule we will define
			
			// Provide some input facts for the test.
			Main.do_("location(user = Stijn, 		location = Lounge).",	itsEngine);
			Main.do_("location(user = Tom, 			location = Lounge).",	itsEngine);
			Main.do_("location(user = Wolf, 		location = Bench).",	itsEngine);
			Main.do_("location(user = Andoni, 		location = Cockpit).",	itsEngine);
			Main.do_("location(user = Christophe, 	location = Bench).",	itsEngine);
			Main.do_("location(user = Elisa, 		location = Cockpit).",	itsEngine);
			
			// Perform all assertions
			while(itsScheduler.step());
		}
		
		{	// Define a rule which passes a variable from its prerequisite to the conclusion
			
			// Define a rule which counts the number of people in my current location
			Main.do_(
					"how_crowded(no_users = ?count, location = ?room) :-" +
					"  my_location(location = ?room)," +
					"  findall(?name, [ location(user = ?name, location = ?room) ], ?users)," +
					"  length(?users, ?count).", 
					
					itsEngine);
			
			// Assert a fact to indicate what my current location is
			Main.do_("my_location(location = Lounge).", itsEngine);
			
			// Perform the assertion
			assertTrue(itsScheduler.step());
			
			// Note that upon asserting the my_location fact different combined tokens will be 
			// propagated to the findall node. The node will in response first propagate a token 
			// containing a list with a match, and subsequently retract this fact to propagate an 
			// updated fact with a list of two users in the Lounge. The agenda will filter out 
			// complementary activations, such that only the last assertion remains active.
			
			// Verify that the rule has triggered
			assertTrue(itsScheduler.step());
			
			// verify that only a single activation was scheduled
			assertFalse(itsScheduler.step());
			
			// Consequently only a single fact was asserted
			theNodeToBeTriggered.assertTriggered(1, 0);
		}
		
		{	// Remove a fact to test the retract/assert performed by the findall node
			
			// Reset the test node
			theNodeToBeTriggered.reset();
			
			Main.undo_("location(user = Stijn, location = Lounge).", itsEngine);

			// Perform the retraction of the location fact
			assertTrue(itsScheduler.step());
			
			// Perform the retraction of the aggregate fact.
			assertTrue(itsScheduler.step());
			
			// Ensure that the retraction of the of fact is indeed performed first 
			theNodeToBeTriggered.assertTriggered(0, 1);

			// Perform the assertion of the new aggregate fact.
			assertTrue(itsScheduler.step());
			
			// Ensure that both retraction and assertion are performed correctly
			theNodeToBeTriggered.assertTriggered(1, 1);
		}
		
		{	// Remove a fact matching the preceding query which will retract the corresponding fact
			
			// Reset the test node
			theNodeToBeTriggered.reset();
			
			Main.undo_("my_location(location = Lounge).", itsEngine);

			// Perform the retraction of the location fact
			assertTrue(itsScheduler.step());
			
			// Perform the retraction of the aggregate fact.
			assertTrue(itsScheduler.step());
			
			// Ensure that the retraction of the of fact is indeed performed
			theNodeToBeTriggered.assertTriggered(0, 1);

			// Ensure that there are no more scheduled activations
			assertFalse(itsScheduler.step());
		}
	}
	
	/**
	 * This method tests the correct behavior of a rule which uses a combination of bagof and length
	 * to count the number of users that are present in a particular type of room.
	 * <p>
	 * This test method uses a bagof operator without any preceding query.
	 */
	public void testLiteralBagOfRule() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the required fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("how_crowded");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{	// Set up the working memory such that we have a few facts matching the rule we will define
			
			// Provide some input facts for the test.
			Main.do_("location(user = Stijn, 		location = Lounge).",	itsEngine);
			Main.do_("location(user = Tom, 			location = Lounge).",	itsEngine);
			Main.do_("location(user = Wolf, 		location = Bench).",	itsEngine);
			Main.do_("location(user = Andoni, 		location = Cockpit).",	itsEngine);
			Main.do_("location(user = Christophe, 	location = Bench).",	itsEngine);
			Main.do_("location(user = Elisa, 		location = Cockpit).",	itsEngine);
			
			// Perform all assertions
			while(itsScheduler.step());
		}
		
		{	// Define a rule which passes a variable from its prerequisite to the conclusion
			
			// Define a rule which counts the number of people in my current location
			Main.do_(
					"how_crowded(no_users = ?count, location = ?room) :-" +
					"  bagof(?name, [ location(user = ?name, location = ?room) ], ?users)," +
					"  length(?users, ?count).", 
					
					itsEngine);
			
			// Note that upon installing the rule for each of the different room types, multiple tokens 
			// will be propagated to the bagof node. The intermediate steps where for a particular room
			// type there is but a single user present are immediately retracted. 
			// The  agenda will filter out complementary activations, such that for each room type only
			// the last assertion remains active. (three in total)
			
			// Verify that the rule has triggered precisely three times (one for each room type).
			for (int i = 0; i < 3; i++)
				assertTrue(itsScheduler.step());
			
			// Verify that no additional activations were scheduled
			assertFalse(itsScheduler.step());
			
			// Consequently only a single fact was asserted
			theNodeToBeTriggered.assertTriggered(3, 0);
		}
		
		{	// Remove a fact to test the retract/assert performed by the bagof node
			
			// Reset the test node
			theNodeToBeTriggered.reset();
			
			Main.undo_("location(user = Stijn, location = Lounge).", itsEngine);

			// Perform the retraction of the location fact
			assertTrue(itsScheduler.step());
			
			// Perform the retraction of the aggregate fact.
			assertTrue(itsScheduler.step());
			
			// Ensure that the retraction of that fact is indeed performed first 
			theNodeToBeTriggered.assertTriggered(0, 1);

			// Perform the assertion of the new aggregate fact.
			assertTrue(itsScheduler.step());
			
			// Ensure that both retraction and assertion are performed correctly
			theNodeToBeTriggered.assertTriggered(1, 1);
		}
	}
	
	
	/**
	 * This method tests the correct behavior of a rule which uses a combination of bagof and length
	 * to count the number of users that are present in a particular room.
	 * <p>
	 * This test method uses a bagof operator whose query is parametrized by variables bound by a
	 * preceding query.
	 */
	public void testParametrizedBagOfRule() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the required fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("how_crowded");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{	// Set up the working memory such that we have a few facts matching the rule we will define
			
			// Provide some input facts for the test.
			Main.do_("location(user = Stijn, 		location = Lounge).",	itsEngine);
			Main.do_("location(user = Tom, 			location = Lounge).",	itsEngine);
			Main.do_("location(user = Wolf, 		location = Bench).",	itsEngine);
			Main.do_("location(user = Andoni, 		location = Cockpit).",	itsEngine);
			Main.do_("location(user = Christophe, 	location = Bench).",	itsEngine);
			Main.do_("location(user = Elisa, 		location = Cockpit).",	itsEngine);
			
			// Perform all assertions
			while(itsScheduler.step());
		}
		
		{	// Define a rule which passes a variable from its prerequisite to the conclusion
			
			// Define a rule which counts the number of people in my current location
			Main.do_(
					"how_crowded(no_users = ?count, location = ?room) :-" +
					"  my_location(location = ?room)," +
					"  bagof(?name, [ location(user = ?name, location = ?room) ], ?users)," +
					"  length(?users, ?count).", 
					
					itsEngine);
			
			// Assert a fact to indicate what my current location is
			Main.do_("my_location(location = Lounge).", itsEngine);
			
			// Perform the assertion
			assertTrue(itsScheduler.step());
			
			// Note that upon asserting the my_location fact different combined tokens will be 
			// propagated to the bagof node. The node will in response first propagate a token 
			// containing a list with a match, and subsequently retract this fact to propagate an 
			// updated fact with a list of two users in the Lounge. The agenda will filter out 
			// complementary activations, such that only the last assertion remains active.
			
			// Verify that the rule has triggered
			assertTrue(itsScheduler.step());
			
			// verify that only a single activation was scheduled
			assertFalse(itsScheduler.step());
			
			// Consequently only a single fact was asserted
			theNodeToBeTriggered.assertTriggered(1, 0);
		}
		
		{	// Remove a fact to test the retract/assert performed by the findall node
			
			// Reset the test node
			theNodeToBeTriggered.reset();
			
			Main.undo_("location(user = Stijn, location = Lounge).", itsEngine);

			// Perform the retraction of the location fact
			assertTrue(itsScheduler.step());
			
			// Perform the retraction of the aggregate fact.
			assertTrue(itsScheduler.step());
			
			// Ensure that the retraction of the of fact is indeed performed first 
			theNodeToBeTriggered.assertTriggered(0, 1);

			// Perform the assertion of the new aggregate fact.
			assertTrue(itsScheduler.step());
			
			// Ensure that both retraction and assertion are performed correctly
			theNodeToBeTriggered.assertTriggered(1, 1);
		}
		
		{	// Remove a fact matching the preceding query which will retract the corresponding fact
			
			// Reset the test node
			theNodeToBeTriggered.reset();
			
			Main.undo_("my_location(location = Lounge).", itsEngine);

			// Perform the retraction of the location fact
			assertTrue(itsScheduler.step());
			
			// Perform the retraction of the aggregate fact.
			assertTrue(itsScheduler.step());
			
			// Ensure that the retraction of the of fact is indeed performed
			theNodeToBeTriggered.assertTriggered(0, 1);

			// Ensure that there are no more scheduled activations
			assertFalse(itsScheduler.step());
		}
	}
	
	/**
	 * This test method verifies whether complimentary activations cancel out one another if they are both
	 * scheduled in the agenda (i.e. before either one of them gets executed by the scheduler).
	 */
	public void testComplementaryUserActivation() {
		// Assert a fact indicating the user's preferences
		Main.do_("user_preferences(volume = 60.0, playlist = Eighties).", itsEngine);

		// Retract the fact indicating the user's preferences
		Main.undo_("user_preferences(volume = 60.0, playlist = Eighties).", itsEngine);

		// Ensure that the corresponding activations canceled each other out and that they are no longer in the agenda.
		assertFalse(itsScheduler.step());
	}
	
	/**
	 * This test method verifies whether complimentary activations cancel out one another if they are both
	 * scheduled in the agenda (i.e. before either one of them gets executed by the scheduler).
	 */
	public void testComplementaryRuleActivation() {
		{	// To allow testing for a complementary rule activation, we need to set up an engine which uses a LIFO agenda
			
			// Ensure that the factory creates a LIFO agenda
			CrimeFactory.factoryClasses.setProperty("Agenda", "edu.vub.crime.rete.agenda.LIFOAgenda");
			
			// Recreate the engine and the scheduler manually
			itsEngine		= CrimeFactory.createEngine();
			itsRootNode		= itsEngine.itsRootNode;
			itsScheduler	= (InlineScheduler) itsEngine.itsScheduler;
		}
		
		// Register the action used in the example as one that causes the test to fail if either asserted or retracted
		ExpressionFactory.registerAbbreviation("createHomeDirectory", "edu.vub.crime.aux.FailUponAssertionAndRetraction");
		
		// Install a rule which creates a home directory for every registered user.
		Main.do_(
			":createHomeDirectory(user = ?name) :-" +
			"  registered(user = ?name).", 
			itsEngine);
		
		// Register a new user, which pushes an activation on top of the stack.
		// AGENDA = [UserActivation: +[registered(user = snostinc)]] E
		Main.do_("registered(user = snostinc).", itsEngine);
		
		// Perform the user activation, which will trigger the rule.
		// AGENDA = [RuleActivation: +[:createHomeDirectory(user = snostinc)]] E
		assertTrue(itsScheduler.step());
		
		// Noting the typo, we unregister the user, which pushes an activation on top of the stack.
		// AGENDA = [UserActivation: -[registered(user = snostinc)]] [RuleActivation: +[:createHomeDirectory(user = snostinc)]] E
		Main.undo_("registered(user = snostinc).", itsEngine);

		// Perform the user activation, which will trigger the rule.
		// AGENDA = [RuleActivation: -[:createHomeDirectory(user = snostinc)]] [RuleActivation: +[:createHomeDirectory(user = snostinc)]] E
		// => AGENDA = E (Complementary activations are removed)
		assertTrue(itsScheduler.step());
		
		// Verify that the agenda is empty as the complementary rule activations have been removed.
		assertFalse(itsScheduler.step());		
	}
	
	/**
	 * This test method verifies that when rules are undeployed, new facts can no longer cause the rule to trigger.
	 */
	public void testDynamicDeployment_StopTriggers() {
		
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the required fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("curriculum");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		Rule theCurriculumRule = compileRule(
				"curriculum(name = ?student, degree = ComputerScience, curriculum = ?curriculum) :-" +
				"  bagof(?course, [ attending(student = ?student, course = Course( title = ?course )) ], ?curriculum).");
			
		itsEngine.deploy(theCurriculumRule);
			
		{	// Assert a single fact and verify whether the rule triggers
			
			// Reset the token counters of the test node
			theNodeToBeTriggered.reset();
			
			// Compile the following assertion an create a user activation which is scheduled for execution in the agenda
			Main.do_("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).", itsEngine);
			
			// Execute the assertion of the fact.
			assertTrue(itsScheduler.step());
			
			// Verify that the rule triggered and an activation was placed in the agenda.
			assertTrue(itsScheduler.step());
			
			// Verify that the fact was correctly added by testing whether the test node was triggered (1 assertion, 0 retractions)
			theNodeToBeTriggered.assertTriggered(1, 0);
		}
		
		{	// Removing the rule from the system does not remove the facts derived by this rule
			
			// Reset the token counters of the test node
			theNodeToBeTriggered.reset();

			itsEngine.undeploy(theCurriculumRule);
			
			// Verify that no activations were created as a result of removing the rule 
			assertFalse(itsScheduler.step());
			
			// And as a consequence the test node will not have been triggered
			theNodeToBeTriggered.assertTriggered(false);
		}
		
		{	// Asserting a single fact after the rule was undeployed has no effect
			
			// Reset the token counters of the test node
			theNodeToBeTriggered.reset();
			
			// Compile the following assertion an create a user activation which is scheduled for execution in the agenda
			Main.do_("attending(student = Stijn, course = Course( title = \"Compilers\", profile = SoftwareEngineering)).", itsEngine);
			
			// Execute the assertion of the fact.
			assertTrue(itsScheduler.step());
			
			// Verify that no additional activations were placed in the agenda.
			assertFalse(itsScheduler.step());
			
			// And as a consequence the test node will not have been triggered
			theNodeToBeTriggered.assertTriggered(false);
		}
	}

	
	/**
	 * This test method verifies whether newly deployed rules correctly trigger based on existing facts
	 * in the engine's working memory.
	 */
	public void testDynamicDeployment_OldFacts() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the required fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("curriculum");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{	// Assert a single fact for which there is currently no corresponding rule.
			
			// Compile the following assertion an create a user activation which is scheduled for execution in the agenda
			Main.do_("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).", itsEngine);
			
			// Execute the assertion of the fact.
			assertTrue(itsScheduler.step());
		}
		
		{ // Install a new rule in the Crime engine
			
			// Reset the token counters of the test node
			theNodeToBeTriggered.reset();
			
			// Add a rule which may trigger with facts already in the working memory
			Main.do_(
				"curriculum(name = ?student, degree = ComputerScience, curriculum = ?curriculum) :-" +
				"  bagof(?course, [ attending(student = ?student, course = Course( title = ?course )) ], ?curriculum).",
			
				itsEngine);
			
			// Verify that the rule triggered and an activation was placed in the agenda.
			assertTrue(itsScheduler.step());
			
			// Verify that the fact was correctly added by testing whether the test node was triggered (1 assertion, 0 retractions)
			theNodeToBeTriggered.assertTriggered(1, 0);

		}		
	}

	/**
	 * When constant attributes are used in the conclusion of a rule, they are represented using an attribute picker.
	 * This test method mingles all forms of constant structures and values to test these attribute picker's construction.
	 */
	public void testAttributePickers() {
		Main.do_(
				"preferences(" +
				"  user = ?user," +
				"  volume = 48.5," +
				"  presets = [" +
				"    [0,  `recorder]," +
				"    [1,  `een]," +
				"    [2,  `canvas]," +
				"    [3,  `bbc1]," +
				"    [4,  `bbc3]," +
				"    [40, `radio1,  options(screen = \"off\")]," +
				"    [50, `jukebox, options(screen = \"off\",      playlist = Generic)]," +
				"    [51, `jukebox, options(screen = \"animated\", playlist = [Party, Eighties])]]," +
				"  packages = [ `default, `music_box | ?custom ]) :-" +
				"  start(user = ?user, additionalPackages = ?custom).", itsEngine);
		
		Main.do_("start(user = Stijn, additionalPackages = [ `comics, `scifi ]).", itsEngine);
		
		// Perform the assertion
		assertTrue(itsScheduler.step());
		
		// Verify whether the rule got triggered
		assertTrue(itsScheduler.step());
	}
	
	/**
	 * This method tests the rules of a grounded negation (i.e. one containing only constant attributes)
	 * as the first (and in this case, only) clause of a rule.
	 */
	public void testGroundedNegationTriggersInAbsenceOfData() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ 	// Set up a test node to verify whether the resulting fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("default_prefences");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{	// Verify that in the absence of preferences, the default preferences are asserted.

			// This rule asserts the default preferences when no preferences are given. 
			Main.do_(
					"default_prefences(" +
					"  volume = 48.5," +
					"  playlist = Generic) :-" +
					"  not user_preferences().", itsEngine);
			
			// Verify that in a system without facts, this rule already yields a match.
			assertTrue(itsScheduler.step());

			// Verify that the default_preferences fact was added
			theNodeToBeTriggered.assertTriggered(1, 0);
		}
		
		{	// Verify that when user preferences are added to the working memory the default ones are retracted. 

			// Reset the test node
			theNodeToBeTriggered.reset();
			
			// Assert a fact indicating the user's preferences
			Main.do_("user_preferences(volume = 60.0, playlist = Eighties).", itsEngine);
			
			// Ensure that the assertion is made which will cause the rule to retract its conclusion.
			assertTrue(itsScheduler.step());

			// Verify that an activation exists to retract the default preferences.
			assertTrue(itsScheduler.step());

			// Verify that the default_preferences fact was removed
			theNodeToBeTriggered.assertTriggered(0, 1);
		}		
	}
	
	/**
	 * This method tests the rules of a grounded negation (i.e. one containing only constant attributes)
	 * as the first (and in this case, only) clause of a rule.
	 */
	public void testGroundedNegationDoesNotTriggerInPresenceOfData() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the resulting fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("default_prefences");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{	// Test the behavior in a system which has the negated fact in its working memory

			// Assert a fact indicating the user's preferences
			Main.do_("user_preferences(volume = 60.0, playlist = Eighties).", itsEngine);
			
			// Ensure that the assertion is made prior to deploying the rule.
			assertTrue(itsScheduler.step());
			
			// This rule asserts the default preferences when no preferences are given. 
			Main.do_(
					"default_prefences(" +
					"  volume = 48.5," +
					"  playlist = Generic) :-" +
					"  not user_preferences().", itsEngine);
			
			// Verify that in a system with user preference facts, this rule does nothing.
			assertFalse(itsScheduler.step());

			// And hence, the test node was not triggered
			theNodeToBeTriggered.assertTriggered(false);
		}
		
		{	// Retract the fact that is negated in the rule to check whether this causes the rule to trigger
			
			// Retract the fact indicating the user's preferences
			Main.undo_("user_preferences(volume = 60.0, playlist = Eighties).", itsEngine);
			
			// Ensure that the retraction is made which in its turn will trigger the rule.
			assertTrue(itsScheduler.step());

			// Verify the assertion of the default preferences
			assertTrue(itsScheduler.step());

			// Finally, ensure that the default_preferences fact was added
			theNodeToBeTriggered.assertTriggered(1, 0);
		}
	}
	
	/**
	 * Tests the behavior of the Rete network when facts are published which contain dataflow attributes.
	 */
	public void testDataFlowAssertion() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{	// Set up a test node to verify whether the resulting fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("user_occupation");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{	// Install a rule which translates location facts to occupation facts
			
			// This rule derives what a user is doing based on the office he is occupying
			Main.do_(
				"user_occupation(user = ?name, occupation = ?work) :-" +
				"  location(user = ?name, location = ?room)," +
				"  room_occupation(location = ?room, occupation = ?work).", 
				itsEngine);
			
			// Sample facts mapping an occupation to a room, based on the unchained working concept.
			Main.do_("room_occupation(location = Bench, 		occupation = \"Interruptable work\").", itsEngine);
			Main.do_("room_occupation(location = Lounge, 		occupation = \"Brainstorming/Informal Meetings\").", itsEngine);
			Main.do_("room_occupation(location = Cockpit, 		occupation = \"Non-interruptable work\").", itsEngine);
			Main.do_("room_occupation(location = MeetingRoom,	occupation = \"Formal Meeting\").", itsEngine);
			
			// Assert all facts
			while(itsScheduler.step());
		}
		
		TestFlowAttribute 	theCurrentLocation;
		
		{	// Create and assert a fact which uses a dataflow attribute to represent the user's location
			
			// Reset the test node
			theNodeToBeTriggered.reset();

			GenericFact	theUsersLocation = ExpressionFactory.unqualifiedFact("location");
			
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theUsersLocation);
			
			// Attribute 1 is a text attribute containing the user's name
			Attribute		 theUserName = ExpressionFactory.attribute("Stijn Mostinckx");
			theFilterPrescriptionIndexer.nextAttribute("user", theUserName);
			
			// Attribute 2 is a dataflow attribute initially set to the Bench type
			Attribute theInitialLocation = ExpressionFactory.type("Bench");
					  theCurrentLocation = new TestFlowAttribute(theInitialLocation);	
			theFilterPrescriptionIndexer.nextAttribute("location" ,theCurrentLocation);
			
			// Schedule an activation that will assert the fact with the initial value for the dataflow attribute
			itsEngine.itsAgenda.schedule(new UserActivation(theUsersLocation, true));
			
			// Perform the activation
			assertTrue(itsScheduler.step());
			
			// Verify that the rule has triggered
			assertTrue(itsScheduler.step());
			
			// Verify that the derived fact was asserted
			theNodeToBeTriggered.assertTriggered(1, 0);
		}
		
		{	// Reset the user's current location to a new valid location, ensure that the previous derived fact is retracted 
			// and a new one is properly asserted.
			
			// Reset the test node
			theNodeToBeTriggered.reset();
			
			// Set the value of the dataflow attribute
			theCurrentLocation.setCurrentValue(ExpressionFactory.type("Lounge"));
			
			// Verify the existence of an activation to retract the previous fact
			assertTrue(itsScheduler.step());
			
			// Verify the existence of an activation to assert the new (flattened) fact
			assertTrue(itsScheduler.step());

			// Verify the retraction of the previous derived fact
			assertTrue(itsScheduler.step());
			
			// Verify the assertion of the newly derived fact
			assertTrue(itsScheduler.step());

			theNodeToBeTriggered.assertTriggered(1, 1);
		}
	}
	
	/**
	 * Tests the behavior of the Rete network when rules contain dataflow attributes. 
	 */
	public void testDataFlowMatching() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{	// Set up a test node to verify whether the resulting fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("IM_Status");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		ExpressionList theParsedRuleExpression 
								= Main.parse("IM_Status(user = ?name, status = \"busy\") :- location(user = ?name).");
		
		Rule theRule 			= (Rule) theParsedRuleExpression.getExpression(0);
		GenericFact	theFilter	= (GenericFact) theRule.getPrerequisite().getExpression(0);
		
		TestFlowAttribute theFlowAttribute;
		
		{ 	// Set up the initial version of the rule which always matches the user as busy.
			
			// All attributes of the filter fact should be indexed. This is done by the following visitor
			IndexingVisitor theFilterIndexer = new IndexingVisitor(theFilter);
						
			// Install a dataflow attribute currently set to the root of the type hierarchy (it matches any valid type)
			Attribute theProfile = ExpressionFactory.type("root");
					  theFlowAttribute = new TestFlowAttribute(theProfile);
			theFilterIndexer.nextAttribute("location", theFlowAttribute);
		}
		
		{	// Verify whether the rule works properly
			
			// Evaluate the rule
			theParsedRuleExpression.accept(new EvaluationVisitor(itsEngine, true));
			
			// Add a user in the Lounge
			Main.do_("location(user = Stijn, location = Lounge).", itsEngine);
			
			// Perform the assertion
			assertTrue(itsScheduler.step());
			
			// Verify that the agenda contains a rule activation with a derived fact.
			assertTrue(itsScheduler.step());
			
			// Add a user in the Cockpit
			Main.do_("location(user = Tom, location = Cockpit).", itsEngine);
			
			// Perform the assertion
			assertTrue(itsScheduler.step());
			
			// Verify that the agenda contains a rule activation with a derived fact.
			assertTrue(itsScheduler.step());
			
			// Verify that both derived facts were properly asserted
			theNodeToBeTriggered.assertTriggered(2, 0);
		}
		
		{	// Ensure that the network responds adequately to a change in the dataflow attribute's value
			
			// Reset the test node
			theNodeToBeTriggered.reset();
			
			// Change the rule such that only people working in the Cockpit are automatically marked as busy
			theFlowAttribute.setCurrentValue(ExpressionFactory.type("Cockpit"));
			
			// Verify that the reevaluation occurs by means of a scheduled activation to avoid concurrency issues
			assertTrue(itsScheduler.step());
			
			// Perform a single retraction of derived information
			assertTrue(itsScheduler.step());
			
			// Ensure that the retraction of "location(user = Tom, location = Cockpit)" and its subsequent
			// re-assertion as part of re-evaluating the rule have canceled each other out.
			assertFalse(itsScheduler.step());
			
			// Verify that consequently only a single retraction of the derived fact was performed
			theNodeToBeTriggered.assertTriggered(0, 1);
			
		}
	}
	
	/**
	 * This method tests what happens if a variable that is passed from prerequisite to the conclusion contains
	 * a dataflow attribute. The expected semantics are that the dataflow attribute is flattened to its current
	 * value, since changes in the dataflow attribute will trigger the rule anew, which automatically results in
	 * the assertion of an adapted conclusion.
	 */
	public void testDataFlowPassedToConclusion() {
		{	// Define a rule which passes a variable from its prerequisite to the conclusion
			
			// Define a rule which counts the number of people in my current location
			Main.do_(
					"how_crowded(no_users = ?count, location = ?room) :-" +
					"  my_location(location = ?room)," +
					"  findall(?name, [ location(user = ?name, location = ?room) ], ?users)," +
					"  length(?users, ?count).", 
					
					itsEngine);
			
			// Provide some input facts for the test.
			Main.do_("location(user = Stijn, 		location = Lounge).",	itsEngine);
			Main.do_("location(user = Tom, 			location = Lounge).",	itsEngine);
			Main.do_("location(user = Wolf, 		location = Bench).",	itsEngine);
			Main.do_("location(user = Andoni, 		location = Cockpit).",	itsEngine);
			Main.do_("location(user = Christophe, 	location = Bench).",	itsEngine);
			Main.do_("location(user = Elisa, 		location = Cockpit).",	itsEngine);
			
			// Perform all assertions
			while(itsScheduler.step());
		}
		
		TestFlowAttribute 	theCurrentLocation;
		
		{	// Create and assert a fact which uses a dataflow attribute to represent the user's current location
			
			GenericFact	theUsersLocation = ExpressionFactory.unqualifiedFact("my_location");
			
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theUsersLocation);
						
			// Attribute 2 is a dataflow attribute initially set to the Bench type
			Attribute theInitialLocation = ExpressionFactory.type("Lounge");
					  theCurrentLocation = new TestFlowAttribute(theInitialLocation);	
			theFilterPrescriptionIndexer.nextAttribute("location" ,theCurrentLocation);
			
			// Schedule an activation that will assert the fact with the initial value for the dataflow attribute
			itsEngine.itsAgenda.schedule(new UserActivation(theUsersLocation, true));
			
			// Perform the activation
			assertTrue(itsScheduler.step());
			
			// Verify that the rule has triggered
			assertTrue(itsScheduler.step());
		}
		
		{	// Setting the value of the dataflow attribute should cause a reassert of the newly flattened base fact
			// However,  the derived fact should be retracted and reasserted by means of a rule activation, instead
			// of by means of a direct reassertion caused by the dataflow attribute.
			
			theCurrentLocation.setCurrentValue(ExpressionFactory.type("Cockpit"));
			
			// Verify that there are only 2 pending updates, to wit:
			// retract my_location(location = Lounge)
			// assert  my_location(location = Cockpit)
			assertEquals(2, itsEngine.getAgenda().size());
			
			while(itsScheduler.step());
		}
	}
	
	/**
	 * This method tests what happens if a dataflow attribute is used literally in the conclusion of a rule. The
	 * expected semantics is that the dataflow attribute remains as is, such that changes to it will be reflected
	 * in changes to the derived fact.
	 * <p>
	 * This test also ensures that if the preconditions to the rule are no longer met that the derived fact is
	 * properly retracted, even after the dataflow attribute's value has changed.
	 */
	public void testDataFlowLiterallyInConclusion() {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{	// Set up a test node to verify whether the resulting fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("location");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		ExpressionList theParsedRuleExpression 
								= Main.parse("location(user = ?name) :- signed_in(user = ?name).");
		
		Rule theRule 			= (Rule) theParsedRuleExpression.getExpression(0);
		Event theConclusion		= (Event) theRule.getConsequence().getExpression(0);
		
		TestFlowAttribute theFlowAttribute;
		
		{ 	// Set up the rule such that a dataflow attribute is added to its conclusion tracking
			// the user's current whereabouts.
			
			// All attributes of the conclusion fact should be indexed. This is done by the following visitor
			IndexingVisitor theFilterIndexer = new IndexingVisitor(theConclusion);
						
			// Install a dataflow attribute currently set to the Lounge
			Attribute theProfile = ExpressionFactory.type("Lounge");
					  theFlowAttribute = new TestFlowAttribute(theProfile);
			theFilterIndexer.nextAttribute("location", theFlowAttribute);
		}
		
		{	// Verify whether the rule works properly
			
			// Evaluate the rule
			theParsedRuleExpression.accept(new EvaluationVisitor(itsEngine, true));
			
			// Have a user of the system sign in.
			Main.do_("signed_in(user = Stijn).", itsEngine);
						
			// Perform the assertion
			assertTrue(itsScheduler.step());
			
			// Verify that the agenda contains a rule activation with a derived fact.
			assertTrue(itsScheduler.step());
			
			// Verify that the derived fact (with the dataflow attribute) was asserted.
			theNodeToBeTriggered.assertTriggered(1, 0);
		}
		
		{	// Verify that no-op changes to the dataflow attribute (setting its value to exactly the same
			// attribute) are properly handled by the agenda which will allow the retraction of the previous
			// flattened fact and the reassertion of the new fact (which has an identical flattened version)
			// to cancel out one another.
			
			// Reset the test node
			theNodeToBeTriggered.reset();
						
			// Change the value of the dataflow attribute
			theFlowAttribute.setCurrentValue(ExpressionFactory.type("Lounge"));
						
			// Verify the absence of an both an activation to retract the previous fact and one
			// assert the new (flattened) fact
			assertFalse(itsScheduler.step());

			// As a consequence, the test node has not been triggered
			theNodeToBeTriggered.assertTriggered(false);
		}
		
		{	// Verify that real changes to the dataflow attribute are reflected in the derived fact
			
			// Reset the test node
			theNodeToBeTriggered.reset();
						
			// Change the value of the dataflow attribute
			theFlowAttribute.setCurrentValue(ExpressionFactory.type("MeetingRoom"));
			
			// Verify the existence of an activation to retract the previous fact
			assertTrue(itsScheduler.step());
			
			// Verify the existence of an activation to assert the new (flattened) fact
			assertTrue(itsScheduler.step());

			theNodeToBeTriggered.assertTriggered(1, 1);
		}
		
		{	// Verify that if the preconditions to the rule no longer hold, the fact is properly retracted
			
			// Reset the test node
			theNodeToBeTriggered.reset();
			
			Main.undo_("signed_in(user = Stijn).", itsEngine);
			
			// Perform the retraction
			assertTrue(itsScheduler.step());
			
			// Verify that the agenda contains a rule deactivation with a derived fact.
			assertTrue(itsScheduler.step());
			
			// Verify that the derived fact (with the dataflow attribute) was retracted.
			theNodeToBeTriggered.assertTriggered(0, 1);		
		}

	}
}
