package edu.vub.crime;

import java.util.Properties;

import junit.framework.TestCase;
import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.rete.agenda.ThreadedScheduler;
import edu.vub.crime.rete.nodes.FilterNode;
import edu.vub.crime.rete.nodes.RootNode;
import edu.vub.crime.rete.nodes.aux.TestNode;

/**
 * This unit test verifies the correct behavior of standard Crime rules, using an threaded 
 * scheduler which executes activations concurrently with the test code. 
 * 
 * @see SteppingRuleTest for a test suite which tests the finer semantic points where control
 * over the interleaving of activations is crucial.
 * 
 * @author smostinc
 */
public class ThreadedRuleTest extends TestCase {
	
	private CrimeEngine			itsEngine;
	private RootNode			itsRootNode;
	private ThreadedScheduler	itsScheduler;
	
	public ThreadedRuleTest() {
		Properties theClassesUnderTest = CrimeFactory.factoryClasses;
		
		theClassesUnderTest.setProperty("Engine",		"edu.vub.crime.CrimeEngine");
		theClassesUnderTest.setProperty("Memory",		"edu.vub.crime.memory.LocalMemory");
		theClassesUnderTest.setProperty("Network",		"edu.vub.crime.network.NetworkLayer");
		theClassesUnderTest.setProperty("Agenda",		"edu.vub.crime.rete.agenda.FIFOAgenda");
		theClassesUnderTest.setProperty("Scheduler",	"edu.vub.crime.rete.agenda.ThreadedScheduler");
		
		itsEngine		= CrimeFactory.createEngine();
		itsRootNode		= itsEngine.itsRootNode;
		itsScheduler	= (ThreadedScheduler) itsEngine.itsScheduler;
	}
	
	/**
	 * This test method is a truncated version of {@link SteppingRuleTest#testPrivateAssertingRule()}. It is
	 * used primarily to test whether closing the agenda allows a threaded scheduler to end properly. Such a
	 * scheduler is typically performing a blocking get() on the agenda when it is closed, requiring the use
	 * of InterruptedExceptions to unblock it and try again.  When doing so,  the scheduler will notice that
	 * the agenda has been closed.
	 */
	public void testFastFIFO_PrivateAssertingRule() throws Exception {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the required fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("curriculum");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{ // Install a new rule in the Crime engine
			Main.do_(
				"curriculum(name = ?student, degree = ComputerScience, curriculum = ?curriculum) :-" +
				"  bagof(?course, [ attending(student = ?student, course = Course( title = ?course )) ], ?curriculum).",
			
				itsEngine);
		}
		
		{	// Assert a single fact and verify whether the rule triggers
			
			// Compile the following assertion an create a user activation which is scheduled for execution in the agenda
			Main.do_("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).", itsEngine);
			
		}
		
		synchronized(this) { this.wait(500); }

		// Close is typically called after the threaded scheduler has finished processing all activations
		// this method thus tests the use of InterruptedExceptions to allow such a scheduler to finish when
		// it would ordinarily be blocked waiting for a new element to be put in the agenda
		itsEngine.itsAgenda.close();

		itsScheduler.getExecutor().join();
		
		// Verify that the fact was correctly added by testing whether the test node was triggered (1 assertion, 0 retractions)
		theNodeToBeTriggered.assertTriggered(1, 0);
	}
	
	/**
	 * This test method performs a similar test than {@link #testFastFIFO_PrivateAssertingRule()}, albeit with
	 * a LIFO agenda.
	 */
	public void testFastLIFO_PrivateAssertingRule() throws Exception {
		{	// Install a LIFO agenda
			
			// Ensure that the factory creates a threaded scheduler and a LIFO agenda
			CrimeFactory.factoryClasses.setProperty("Scheduler",	"edu.vub.crime.rete.agenda.ThreadedScheduler");
			CrimeFactory.factoryClasses.setProperty("Agenda", 		"edu.vub.crime.rete.agenda.LIFOAgenda");
			
			// Recreate the engine and the scheduler manually
			itsEngine		= CrimeFactory.createEngine();
			itsRootNode		= itsEngine.itsRootNode;
			itsScheduler	= (ThreadedScheduler) itsEngine.itsScheduler;
		}
		
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the required fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("curriculum");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{ // Install a new rule in the Crime engine
			Main.do_(
				"curriculum(name = ?student, degree = ComputerScience, curriculum = ?curriculum) :-" +
				"  bagof(?course, [ attending(student = ?student, course = Course( title = ?course )) ], ?curriculum).",
			
				itsEngine);
		}
		
		{	// Assert a single fact and verify whether the rule triggers
			
			// Compile the following assertion an create a user activation which is scheduled for execution in the agenda
			Main.do_("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).", itsEngine);
			
		}
		
		synchronized(this) { this.wait(500); }

		// Close is typically called after the threaded scheduler has finished processing all activations
		// this method thus tests the use of InterruptedExceptions to allow such a scheduler to finish when
		// it would ordinarily be blocked waiting for a new element to be put in the agenda
		itsEngine.itsAgenda.close();

		itsScheduler.getExecutor().join();
		
		// Verify that the fact was correctly added by testing whether the test node was triggered (1 assertion, 0 retractions)
		theNodeToBeTriggered.assertTriggered(1, 0);
	}
	
	
	/**
	 * This test method verifies whether complimentary activations cancel out one another if they are both
	 * scheduled in the agenda (i.e. before either one of them gets executed by the scheduler).
	 */
	public void testComplementaryRuleActivation() throws Exception {
		{	// To allow testing for a complementary rule activation, we need to set up an engine which uses a LIFO agenda
			
			// Ensure that the factory creates a threaded scheduler and a LIFO agenda
			CrimeFactory.factoryClasses.setProperty("Scheduler",	"edu.vub.crime.aux.SlowerScheduler");
			CrimeFactory.factoryClasses.setProperty("Agenda", 		"edu.vub.crime.rete.agenda.LIFOAgenda");
			
			// Recreate the engine and the scheduler manually
			itsEngine		= CrimeFactory.createEngine();
			itsRootNode		= itsEngine.itsRootNode;
			itsScheduler	= (ThreadedScheduler) itsEngine.itsScheduler;
		}
		
		// Register the action used in the example as one that causes the test to fail if either asserted or retracted
		ExpressionFactory.registerAbbreviation("createHomeDirectory", "edu.vub.crime.aux.FailUponAssertionAndRetraction");
		
		// Install a rule which creates a home directory for every registered user.
		Main.do_(
			":createHomeDirectory(user = ?name) :-" +
			"  registered(user = ?name).", 
			itsEngine);
		
		// Register a new user, which pushes an activation on top of the stack.
		// AGENDA = [UserActivation: +[registered(user = snostinc)]] E
		Main.do_("registered(user = snostinc).", itsEngine);
		
		// Wait a bit to ensure that the following activation is pushed after the rule activation
		synchronized (this) { this.wait(100); }
		
		// Noting the typo, we unregister the user, which pushes an activation on top of the stack.
		// AGENDA = [UserActivation: -[registered(user = snostinc)]] [RuleActivation: +[:createHomeDirectory(user = snostinc)]] E
		Main.undo_("registered(user = snostinc).", itsEngine);

		itsEngine.itsAgenda.close();
		
		itsScheduler.getExecutor().join();
	}
	
	/**
	 * Truncated version of {@link SteppingRuleTest#testGroundedNegationDoesNotTriggerInPresenceOfData()}
	 * which verifies that a grounded negation as the first clause of a rule does not trigger when matching
	 * facts exist in the working memory prior to the installation of the rule.
	 * <p>
	 * The original rule also tests what happens when that fact is subsequently removed.
	 */
	public void testGroundedNegationDoesNotTriggerInPresenceOfData() throws Exception {
		TestNode theNodeToBeTriggered = new TestNode();
		
		{ // Set up a test node to verify whether the resulting fact was added
			
			FilterNode theFilterNode = itsRootNode.getChild("default_prefences");
			theFilterNode.addChild(theNodeToBeTriggered);
		}
		
		{	// Test the behavior in a system which has the negated fact in its working memory
			
			// Assert a fact indicating the user's preferences
			Main.do_("user_preferences(volume = 60.0, playlist = Eighties).", itsEngine);
			
			// Ensure that the assertion is made prior to deploying the rule.
			synchronized(this) { this.wait(500); }
			
			// Define SignalError : test fails if activated
			ExpressionFactory.registerAbbreviation("SignalError", "edu.vub.crime.aux.FailUponAssertion");
			
			// This rule asserts the default preferences when no preferences are given. 
			Main.do_(
					":SignalError()," +
					"default_prefences(" +
					"  volume = 48.5," +
					"  playlist = Generic) :-" +
					"  not user_preferences().", itsEngine);
		}
		
		itsEngine.itsAgenda.close();
		
		itsScheduler.getExecutor().join();
		
		// And hence, the test node was not triggered
		theNodeToBeTriggered.assertTriggered(false);
	}
}
