/**
 * 
 */
package edu.vub.crime.dataflow;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import java.util.Stack;

import edu.vub.crime.CrimeEngine;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.AttributeVisitor;
import edu.vub.crime.grammar.attributes.CompoundAttribute;
import edu.vub.crime.grammar.attributes.ConstantAttribute;
import edu.vub.crime.grammar.attributes.datastructures.List;
import edu.vub.crime.grammar.attributes.datastructures.Pair;
import edu.vub.crime.grammar.attributes.datastructures.TypedObject;
import edu.vub.crime.grammar.utils.AttributePicker;
import edu.vub.crime.rete.agenda.UserActivation;

/**
 * The introduction of dataflow attributes implies that facts may change over time. This needs to be
 * accounted for when compiling such facts into filter nodes, as well as when asserting a fact which
 * contains a dataflow attribute. Rather than publishing the fact with the dataflow attribute in it,
 * a flattened version is published where every dataflow attribute is replaced by its current value.
 * <p>
 * The flattened version of such a fact is compiled by CompileAssertionVisitor instances. This class
 * also maintains the required infrastructure to ensure that when a dataflow attribute changes, the
 * asserted fact is retracted and replaced by a more current version.
 *  
 * @author smostinc
 *
 */
public class CompileAssertionVisitor extends AttributeVisitor {
	
	private Stack 	itsConsumers;
	private boolean	containsDataflows;
	
	private final GenericFact theOriginalFact;
	private final GenericFact theFlattenedFact;
	private final CrimeEngine theEngineInWhichItWillBeAsserted;
	private final ListenerSynchronizationChannel itsSynchronizationChannel;
		
	private static class ListFiller extends CompoundAttribute {
		
		private final List theFlattenedList;
		
		public ListFiller(List theFlattenedList) {
			this.theFlattenedList = theFlattenedList;
		}

		public void addAttribute(Attribute theAttribute) {
			theFlattenedList.add(theAttribute);
		}
		
		public Attribute getAttribute(Iterator indexIterator) {	return null; }
		
		public AttributePicker makeAttributePicker(ExpressionList facts) { return null; }
		
		public Object clone() { return this; }
		
		public boolean equals(Object o) { return false; }
		
		public void accept(AttributeVisitor theVisitor) {}
	}
	
	private static class ObjectProducer extends CompoundAttribute {
		
		public TypedObject theProducedObject;
		
		public void addAttribute(Attribute theAttribute) {
			theProducedObject = new TypedObject(theAttribute);
		}

		public Attribute getAttribute(Iterator indexIterator) {	return null; }
		
		public AttributePicker makeAttributePicker(ExpressionList facts) { return null; }
		
		public Object clone() { return this; }
		
		public boolean equals(Object o) { return false; }
		
		public void accept(AttributeVisitor theVisitor) {}
	}
	
	private static class PairProducer extends CompoundAttribute {
		
		private final Pair theOriginalPair;
		private final TypedObject theFlattenedObject;
		
		public PairProducer(TypedObject theFlattenedObject, Pair theOriginalPair) {
			this.theFlattenedObject = theFlattenedObject;
			this.theOriginalPair    = theOriginalPair;
		}

		public void addAttribute(Attribute theAttribute) {
			theFlattenedObject.addField(
					new Pair(
							theOriginalPair.getName(),
							theOriginalPair.getOperator(),
							theAttribute));
		}
		
		public Attribute getAttribute(Iterator indexIterator) {	return null; }
		
		public AttributePicker makeAttributePicker(ExpressionList facts) { return null; }
		
		public Object clone() { return this; }
		
		public boolean equals(Object o) { return false; }
		
		public void accept(AttributeVisitor theVisitor) {}
	}
	
	private static class NamedAttributeProducer extends CompoundAttribute {
		
		private final String 		theAttributeName;
		private final GenericFact	theFlattenedFact;
		
		public NamedAttributeProducer(GenericFact theFlattenedFact, String theAttributeName) {
			this.theFlattenedFact	= theFlattenedFact;
			this.theAttributeName	= theAttributeName;
		}

		public void addAttribute(Attribute theAttribute) {
			theFlattenedFact.addNamedAttribute(theAttributeName, theAttribute);
		}
		
		public Attribute getAttribute(Iterator indexIterator) {	return null; }
		
		public AttributePicker makeAttributePicker(ExpressionList facts) { return null; }
		
		public Object clone() { return this; }
		
		public boolean equals(Object o) { return false; }
		
		public void accept(AttributeVisitor theVisitor) {}
	}
	
	/*
	 * When a fact is asserted which has multiple dataflow attributes,  each of these attributes
	 * can evolve independently. To ensure that facts are only retracted when the first dataflow
	 * attribute changes, this class is used to synchronize the different listeners.
	 */
	protected static class ListenerSynchronizationChannel {
		
		public boolean someListenerHasTriggered = false;
	}
	
	private static class DataflowListener implements ActionListener {
		
		private final Attribute				theStoredValue;
		private final DataFlowAttribute		theObservedFlow;
		private final GenericFact	theOriginalFact;
		private final GenericFact	theFlattenedFact;
		private final CrimeEngine	theEngineInWhichItWillBeAsserted;
		private final ListenerSynchronizationChannel theSynchronizationChannel;
		
		public DataflowListener(
				DataFlowAttribute    theObservedFlow,
				GenericFact theOriginalFact,
				GenericFact theFlattenedFact,
				CrimeEngine theEngineInWhichItWillBeAsserted,
				ListenerSynchronizationChannel theChannel) {
			this.theStoredValue  = theObservedFlow.getCurrentValue();
			this.theObservedFlow = theObservedFlow;
			this.theOriginalFact = theOriginalFact;
			this.theFlattenedFact = theFlattenedFact;
			this.theEngineInWhichItWillBeAsserted = theEngineInWhichItWillBeAsserted;
			this.theSynchronizationChannel = theChannel;
		}

		public void actionPerformed(ActionEvent e) {
			if(theStoredValue.equals(theObservedFlow.getCurrentValue()))
				return; // If the value has not changed, ignore the update.
			
			theObservedFlow.unsubscribe(this);
			
			if(theSynchronizationChannel.someListenerHasTriggered)
				return; // this was a left-over listener as some other listener has already retracted the fact
			
			// Schedule an activation to remove the flattened fact from the working memory.
			// It is crucial to use an activation to avoid concurrency problems when using a threaded scheduler.
			// theFlattenedFact.retract(theEngineInWhichItWillBeAsserted);
			theEngineInWhichItWillBeAsserted.getAgenda().schedule(
					new UserActivation(theFlattenedFact, false));
			
			// Schedule a new assertion of the original fact, which will then be flattened again.
			// Again we use an activation to avoid concurrency problems.
			theEngineInWhichItWillBeAsserted.getAgenda().schedule(
					new UserActivation(theOriginalFact, true));
			
			theSynchronizationChannel.someListenerHasTriggered = true;
		}
	}
	
	public CompileAssertionVisitor(
			GenericFact theOriginalFact,
			GenericFact theFlattenedFact,
			CrimeEngine theEngineInWhichItWillBeAsserted) {
		super("flatten", " into a constant Attribute");
		
		this.itsConsumers = new Stack();
		this.theOriginalFact = theOriginalFact;
		this.theFlattenedFact = theFlattenedFact;
		this.theEngineInWhichItWillBeAsserted = theEngineInWhichItWillBeAsserted;
		this.itsSynchronizationChannel = new ListenerSynchronizationChannel();

		TypedObject theArguments = theFlattenedFact.getRepresentation();
		
		itsConsumers.push(theArguments);
	}

	public static GenericFact compileFact(
			GenericFact theFactToBeAsserted, 
			CrimeEngine theEngineInWhichItWillBeAsserted) {

		GenericFact theFlattenedFact = new GenericFact(
				theFactToBeAsserted.getSpace(),
				theFactToBeAsserted.getLabel());
		
		CompileAssertionVisitor theVisitor = 
			new CompileAssertionVisitor(theFactToBeAsserted, theFlattenedFact, theEngineInWhichItWillBeAsserted);
				
		for (Iterator attributeIt = theFactToBeAsserted.attributeIterator(); attributeIt.hasNext();) {
			Pair currentAttribute = (Pair) attributeIt.next();
			
			theVisitor.itsConsumers.push(new NamedAttributeProducer(theFlattenedFact, currentAttribute.getName()));

			currentAttribute.getValue().accept(theVisitor);

			theVisitor.itsConsumers.pop();
		}
		
		return (theVisitor.containsDataflows)?theFlattenedFact:theFactToBeAsserted;
	}
	
	// Data-driven attributes
	
	public void visit(DataFlowAttribute theFlow) {
		containsDataflows = true;
		
		theFlow.subscribe(
				new DataflowListener(theFlow, theOriginalFact, theFlattenedFact, theEngineInWhichItWillBeAsserted, itsSynchronizationChannel));
		
		Attribute			theAttribute = theFlow.getCurrentValue();
		CompoundAttribute	theConsumer  = (CompoundAttribute)itsConsumers.peek();
		
		theConsumer.addAttribute(theAttribute);
	}
	
	// Compound attributes

	public void visit(TypedObject theTypedObject) {
		TypedObject theFlattenedObject;
		
		{
			ObjectProducer theObjectProducer = new ObjectProducer();
		
			itsConsumers.push(theObjectProducer);
		
			theTypedObject.getType().accept(this);
		
			itsConsumers.pop();
		
			theFlattenedObject = theObjectProducer.theProducedObject;
		}
		
		for (Iterator fieldIt = theTypedObject.iterator(); fieldIt.hasNext();) {
			Pair currentField = (Pair) fieldIt.next();
			
			itsConsumers.push(new PairProducer(theFlattenedObject, currentField));

			currentField.getValue().accept(this);
			
			itsConsumers.pop();
		}
		
		CompoundAttribute theConsumer = (CompoundAttribute)itsConsumers.peek();
		
		theConsumer.addAttribute(theFlattenedObject);
	}
	
	public void visit(List theList) {
		List theFlattenedList = new List();
		
		itsConsumers.push(new ListFiller(theFlattenedList));
		
		for (Iterator attributeIt = theList.iterator(); attributeIt.hasNext();) {
			Attribute currentAttr = (Attribute) attributeIt.next();
			
			currentAttr.accept(this);
		}
		
		itsConsumers.pop();
		
		CompoundAttribute theConsumer = (CompoundAttribute)itsConsumers.peek();
		
		theConsumer.addAttribute(theFlattenedList);
	}
	
	// Constant Attributes
	
	public void visit(ConstantAttribute theConstant) {
		CompoundAttribute theConsumer = (CompoundAttribute)itsConsumers.peek();
		
		theConsumer.addAttribute(theConstant);
	}
}