package edu.vub.crime.dataflow;

import java.awt.event.ActionListener;
import java.util.Vector;

import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.AttributeVisitor;
import edu.vub.crime.grammar.utils.AttributePicker;
import edu.vub.crime.grammar.utils.UtilsFactory;

/**
 * This abstract class represents an attribute which is likely to change during the course of the program.
 * If such an attribute is used in a fact that is asserted, the CRIME engine will make sure that a derived
 * fact is published which contains the current value of the dataflow attribute.  If the attribute's value
 * changes, the original fact will be retracted automatically and the updated version will be asserted.
 * <p>
 * Similarly if such an attribute is used in a query expression,  the query will be completely reevaluated
 * when the value is changed.
 * <p>
 * Subclasses are required to notify all subscribers upon updating the value. 
 * 
 * @author smostinc
 */
public abstract class DataFlowAttribute extends Attribute {
	
	protected Vector itsSubscribers = new Vector();
	
	public void accept(AttributeVisitor theVisitor) {
		theVisitor.visit(this);
	}
	
	public AttributePicker makeAttributePicker(ExpressionList facts) {
		return UtilsFactory.ctePicker(this);
	}

	public abstract Attribute getCurrentValue();
	
	public void subscribe(ActionListener theListener) {
		itsSubscribers.add(theListener);
	}
	
	public void unsubscribe(ActionListener theListener) {
		itsSubscribers.remove(theListener);
	}
	
	public String toString() {
		return "->" + getCurrentValue() + "";
	}
	
	public Object clone() {
		return this;
	}
	
}
