package edu.vub.crime.grammar;

import java.io.Serializable;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.Variable;
import edu.vub.crime.grammar.attributes.datastructures.Operator;
import edu.vub.crime.grammar.attributes.datastructures.Pair;
import edu.vub.crime.grammar.attributes.datastructures.TypedObject;
import edu.vub.crime.rete.filter.Join;
import edu.vub.crime.rete.nodes.Node;

/** 
 * This class represents the common interface of all valid expressions in the CRIME
 * coordination language. It offers methods used by the parser to check the validity
 * of the expression, as well as a mechanism to evaluate the expression given an 
 * inference engine.
 * 
 * @brief Superclass of all valid expressions in the CRIME coordination language.
 */
public abstract class Expression implements Serializable {

	public Expression() {}
	
	public Set variables() { return new HashSet(); };
			
	/**
	 * This method is used to create a {@link Join} constraint for a particular variable, should it occur in this
	 * fact. Join constraints are used to ensure that when tokens are joined by the Rete network, all variables 
	 * which are shared between two tokens have the same binding.
	 * <p>
	 * Rather than encoding this behavior in the compilation process itself, it is represented as a method on the
	 * Expression hierarchy to allow language primitives (such as bagof and findall) to provide a dedicated 
	 * implementation.
	 * <p>
	 * Since in a single token all variable bindings are by construction equal, this method returns a boolean to 
	 * indicate whether the expression has installed a new Join constraint. If this is the case, the compiler can
	 * skip directly to the next variable.
	 * 
	 * @param theVariable a Variable which needs to be Joined
	 * @param theIndexIntheLeftToken the index of this Expression in the left token
	 * @param theIndexInTheRightToken the index of the fact in the right token which contains theVariable
	 * @param theNodeInWhichToAdd the node where the Join constraints should be added to
	 * @return a boolean signaling when a constraint was added
	 */
	public boolean compileJoinConstraint(Variable theVariable, int theIndexIntheLeftToken, int theIndexInTheRightToken, Node theNodeInWhichToAdd) {
		for(Iterator variableIt = variables().iterator(); variableIt.hasNext();) {
			Variable currentvariable = (Variable)variableIt.next();
			
			if(currentvariable.equals(theVariable)) {
				Vector myVariableIndex 		= currentvariable.getIndex();
				Vector inputVariableIndex 	= theVariable.getIndex();
				
				Join theConstraint = new Join(
						theIndexIntheLeftToken, theIndexInTheRightToken, 
						myVariableIndex, inputVariableIndex, Operator.EQL); 
				
				theNodeInWhichToAdd.addNewJoin(theConstraint);
				
				return true;				
			}
		}
		return false;
	}

	public boolean compileJoinConstraint(Property theProperty, int theIndexIntheLeftToken, int theIndexInTheRightToken, Node theNodeInWhichToAdd) {
		for(Iterator variableIt = variables().iterator(); variableIt.hasNext();) {
			Variable currentvariable = (Variable)variableIt.next();
			
			if(currentvariable.equals(theProperty.getVariable())) {
				Stack myVariableIndex 		= (Stack)currentvariable.getIndex().clone();
				Stack theOtherVariableIndex = null;
				TypedObject thePropertyObject = theProperty.getObject();
				
				while(!thePropertyObject.isEmpty()) {
					Iterator 	theMembers	= thePropertyObject.iterator();
					Pair 		theMember	= (Pair) theMembers.next();
					Attribute 	theData		= theMember.getValue();
					
					myVariableIndex.push(theMember.getName());
					
					if(theData instanceof Variable) {
						theOtherVariableIndex = (Stack)theData.getIndex();
						break;
					} else {
						thePropertyObject = (TypedObject)theData;
					}
				}
				if(theOtherVariableIndex != null) {
					Join theConstraint = new Join(
							theIndexIntheLeftToken, theIndexInTheRightToken, 
							myVariableIndex, theOtherVariableIndex, Operator.EQL); 
					
					theNodeInWhichToAdd.addNewJoin(theConstraint);
					
					return true;				
				}
			}
		}
		return false;		
	}

	
	public void addProperty(Property property){
//		Variable propertyVariable = property.getVariable();
//		if(variables_.contains(propertyVariable)){
//			variables_.addAll(property.variables());
//			constraints_.addAll(property.variables());
//			boolean notupdated = true;
//			for(Iterator itr=variables_.iterator(); itr.hasNext() && notupdated;){
//				Variable var=(Variable)itr.next();
//				if(var.equals(propertyVariable))
//				{
//					Vector objIndex = (Vector)var.getIndex().clone(); 
//					CrimeObject object = property.getObject();
//					object.setIndex(objIndex);
//					object.updateIndex();
//					attributes_.add(object);
//					notupdated = false;
//				}
//			}
//		}
	}

	// The number of facts this expression will produce, always 1?
	public int tokenSize() { return 1; }
	
	public abstract void accept(GrammarVisitor visitor);
}