package edu.vub.crime.grammar;

import java.util.Vector;

import edu.vub.crime.CrimeEngine;
import edu.vub.crime.dataflow.CompileAssertionVisitor;
import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.attributes.constant.Symbol;
import edu.vub.crime.grammar.attributes.constant.Type;
import edu.vub.crime.grammar.attributes.datastructures.Pair;
import edu.vub.crime.grammar.attributes.datastructures.TypedObject;

/**
 * Facts consist of at least a label and a list of attributes:  for instance, the text
 * "prefers(`Alice, ?genre)" will be parsed into a genericfact with the type "prefers"
 * as its label, and a list of attributes consisting out of the symbol  `Alice and the 
 * variable ?genre.
 * <p>
 * 
 * In general, a fact can be seen as a pair of a list of attributes together and an 
 * object that encapsulates its meta-data. This object will at least contain the label,
 * but for facts published in a (named) public fact space, it will also contain the 
 * name of that fact space, as well as its publisher.
 *  
 * @author smostinc
 */
public class GenericFact extends AttributedExpression {
	
	public static final String		META_NAME  = "meta";
	public static final String		LABEL_NAME = "label";
	public static final String		SPACE_NAME = "space";
	public static final String		PUB_NAME   = "publisher";
	
	public static final Type		META_TYPE  = ExpressionFactory.type(META_NAME);
	
	public static final String		PRIVATE	   = "private";

	private final		TypedObject	itsMetaData;
	private final 		Vector		itsConstraintsToPropagate;
	private final		boolean		isPublic;
	
	public GenericFact(String theSpace, String theLabel) {
		super(theLabel);
		
		isPublic	= (theSpace != PRIVATE);

		itsMetaData	= new TypedObject(META_TYPE);
		
		itsMetaData.addField(new Pair(LABEL_NAME, ExpressionFactory.type(theLabel)));

		if(isPublic) {
			itsMetaData.addField(new Pair(SPACE_NAME, ExpressionFactory.type(theSpace)));
			itsMetaData.addField(new Pair(PUB_NAME,   ExpressionFactory.symbol("_THIS_")));
		}
		
		itsRepresentation.addField(new Pair(META_NAME, itsMetaData));
		
		// Proper indexing of meta data as this will not be performed by the Indexing visitor
		itsMetaData.setIndex(META_IDX);
		
		itsConstraintsToPropagate = new Vector();
	}
	
	public GenericFact(TypedObject theRepresentation, String theSpace) {
		super(theRepresentation);
		
		isPublic	= (theSpace != PRIVATE);

		itsMetaData	= new TypedObject(META_TYPE);
		
		itsMetaData.addField(new Pair(LABEL_NAME, itsRepresentation.getType()));
		
		if(isPublic) {
			itsMetaData.addField(new Pair(SPACE_NAME, ExpressionFactory.type(theSpace)));
			itsMetaData.addField(new Pair(PUB_NAME,   ExpressionFactory.symbol("_THIS_")));
		}
		
		itsRepresentation.addField(new Pair(META_NAME, itsMetaData));
		
		// Proper indexing of meta data as this will not be performed by the Indexing visitor
		itsMetaData.setIndex(META_IDX);
		
		itsConstraintsToPropagate = new Vector();
	}
	
	/**
	 * Inform the visitor of this node's concrete type, by calling its visit method with the 
	 * correct overloaded signature.
	 */
	public void accept(GrammarVisitor visitor) {
		visitor.visit(this);
	}
		
	public void addMetaData(Pair theMetaData) {
		itsMetaData.addField(theMetaData);
	}
	
//	// HANDLING VARIABLE CONSTRAINTS
//	public void addProperty(Property property){
//		Variable propertyVariable = property.getVariable();
//		if(variables().contains(propertyVariable)) {
//			variables().addAll(property.variables());
//			
//			// Loop of the variables set to find the Variable that .equals() the one that is 
//			// given in the property, since the once stored in this fact is correctly indexed
//			boolean notIndexed = true;
//			
//			for(Iterator variableIt = variables().iterator(); variableIt.hasNext() && notIndexed;) {
//				Variable currentVariable = (Variable) variableIt.next();
//				
//				if(currentVariable.equals(propertyVariable)) {
//					Vector theVariableIndex 		= (Vector) currentVariable.getIndex().clone(); 
//					TypedObject theConstraintObject	= property.getObject();
//					
//					theConstraintObject.setIndex(theVariableIndex);
//					theConstraintObject.updateIndex();
//					itsArguments.addAttribute(theConstraintObject);
//					notIndexed = false;
//				}
//			}
//		} else {
//			itsConstraintsToPropagate.add(property);
//			// System.out.println("A property is being dropped");
//		}
//	}
	
	public Vector getUnresolvedConstraints() {
		return itsConstraintsToPropagate;
	}
	
	// ASSERTION / RETRACTION
    /** 
     * Assert this fact in the working memory of the given engine. Depending on whether the fact
     * is private of public, the assert or the assertShared method of the engine will be invoked.
     */
	public void assert(CrimeEngine engine) {
		impl_flatten(engine);
		
		if(isPublic) {
			engine.assertShared(itsFlattenedRepresentation);
		} else {
			engine.assert(itsFlattenedRepresentation);
		}
	}
	
	/** 
	 * Retract this fact from the working memory of the given engine. Depending on whether the fact
     * is private of public, the retract or the retractShared method of the engine will be invoked. 
	 * </pre> */
	public void retract(CrimeEngine engine) {
		// It is possible that a fact is retracted which has not been flattened for assertion. This
		// is most notably the case when the same fact is parsed twice by some user interface.
		if(itsFlattenedRepresentation == null) impl_flatten(engine);
		
		if(isPublic) {
			engine.retractShared(itsFlattenedRepresentation);
		} else {
			engine.retract(itsFlattenedRepresentation);
		}
	}
	
	// ACCESSORS
	
	public static final Vector	META_IDX;
	public static final Vector	LABEL_IDX;
	public static final Vector	SPACE_IDX;
	public static final Vector	PUB_IDX;

	static {
		// The shared prefix for all of the meta indices
		META_IDX = new Vector(2);
		META_IDX.add(META_NAME);
		
		LABEL_IDX = (Vector)META_IDX.clone();
		LABEL_IDX.add(LABEL_NAME);

		SPACE_IDX = (Vector)META_IDX.clone();
		SPACE_IDX.add(SPACE_NAME);

		PUB_IDX   = (Vector)META_IDX.clone();
		PUB_IDX.add(PUB_NAME);
		
	}
	
	public String getSpace() {
		try {
			if(!isPublic) return PRIVATE;
			
			Type itsSpace = (Type)getAttribute(SPACE_IDX.iterator());
			return itsSpace.toString();
		} catch (IncorrectAttributeIndexException e) { 
			throw new RuntimeException("Index to access the space of a fact should never be invalid");
		}
	}
	
	public String getLabel() {
		try {
			Type itsSpace = (Type)getAttribute(LABEL_IDX.iterator());
			
			return itsSpace.toString();
		} catch (IncorrectAttributeIndexException e) {
			throw new RuntimeException("Index to access the label of a fact should never be invalid");
		}
	}
	
	public boolean isPublic() {
		return isPublic;
	}
	
	public String getPublisher() {
		try {
			if(!isPublic) return null;
			
			Symbol itsPublisher = (Symbol)getAttribute(PUB_IDX.iterator());
			
			return itsPublisher.getValue();
		} catch (IncorrectAttributeIndexException e) {
			throw new RuntimeException("Index to access the publisher of a fact should never be invalid");
		}
	}
	
	public void setPublisher(String publisherId) {
		itsMetaData.setField(PUB_NAME, ExpressionFactory.symbol(publisherId));
	}
	
	// REPRESENTATION
	
	public String toString() {
		StringBuffer theVisualRepresentation = new StringBuffer();
		
		if(isPublic) {
			theVisualRepresentation.append(getSpace());
			theVisualRepresentation.append("->");
		}
		
		theVisualRepresentation.append(itsRepresentation);

		return theVisualRepresentation.toString();
	}
	
	// COMPARISON
	
	public boolean equals(Object theOtherObject) {
		if(theOtherObject == null) return false;
		
		if(theOtherObject instanceof GenericFact) {
			GenericFact theOtherFact = (GenericFact) theOtherObject;
			return itsRepresentation.equals(theOtherFact.itsRepresentation);
		}
		
		return false;
	}
	
	// TODO : HACK (should be tree-recursive and probably use the visitor infrastructure to achieve this)
	
	private GenericFact itsFlattenedRepresentation;
	
	private void impl_flatten(final CrimeEngine theEngineInWhichItWillBeAsserted) {
		itsFlattenedRepresentation = CompileAssertionVisitor.compileFact(this, theEngineInWhichItWillBeAsserted);
	}
}
