package edu.vub.crime.grammar;

import edu.vub.crime.grammar.operators.BagOf;
import edu.vub.crime.grammar.operators.FindAll;
import edu.vub.crime.grammar.operators.Length;


/**
 * The GrammarVisitor class provides a framework to easily extend the set of operations 
 * that can be performed on Grammar elements. Such operations vary from creating a rete
 * network, transforming facts in a consequence to events, serializing facts and so on.
 * <p>
 * An obvious downside to the use of a visitor pattern is that extending the grammar is
 * more complicated since all visitor classes need to be updated. The GrammarVisitor 
 * class attempts to alleviate this issue by providing implementations for all methods 
 * which throw a {@link RuntimeException} with a customizable message. Subclasses can
 * thus contend themselves with providing a correct error message and override only the 
 * relevant methods.
 *  
 * @author smostinc
 *
 */
public class GrammarVisitor {
	
	private final String itsPurpose;
	private final String itsModality;

	public GrammarVisitor(String thePurpose, String theModality) {
		itsPurpose = thePurpose;
		itsModality = theModality;
	}
	
	public void visit(BagOf theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a BagOf node" + itsModality + "."); };
	public void visit(Event theNode) { throw new RuntimeException("Cannot " + itsPurpose + " an Event node" + itsModality + "."); };
	public void visit(ExpressionList theNode) { throw new RuntimeException("Cannot " + itsPurpose + " an ExpressionList node" + itsModality + "."); };
	public void visit(FindAll theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a FindAll node" + itsModality + "."); };
	public void visit(Length theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a Length node" + itsModality + "."); };
	public void visit(Negation theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a Negation node" + itsModality + "."); };
	public void visit(Package theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a Package node" + itsModality + "."); };
	public void visit(Rule theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a Rule node" + itsModality + "."); };
	public void visit(Property theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a Property node" + itsModality + "."); };
	
	public void visit(GenericFact theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a Fact node" + itsModality + "."); };
	public void visit(Constraint theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a Constraint node" + itsModality + "."); };

	
	/* / Extensions for temporal operators 
	public void visit(Between theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a Between node" + itsModality + "."); };
	public void visit(Contra theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a Contra node" + itsModality + "."); };
	public void visit(MostRecent theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a MostRecent node" + itsModality + "."); };
	public void visit(Since theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a Since node" + itsModality + "."); };
	public void visit(SometimePast theNode) { throw new RuntimeException("Cannot " + itsPurpose + " a SometimePast node" + itsModality + "."); };
	*/
}
