/** <pre> 
 * @file Attribute.java
 * 
 * @author Eline Philips
 * @date 2006-09-22
 * </pre> */

package edu.vub.crime.grammar.attributes;

import java.io.Serializable;
import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.utils.AttributePicker;

/** <pre> 
 * Attribute class.
 * @brief This abstract class is the common root for all attributes,
 * attributes are used in facts.
 *  
 * for example someFact(one,1,?one). 
 * In this example 
 * 	 + one is a string attribute
 *   + 1 is a integer attribute
 *   + ?one is a variable attribute
 *   
 * A special attribute is the list attribute which contains a list of attributes
 * 
 * </pre> 
 * 
 * @author cfscholl
 * @author ephilips
 * @author smostinc
 */
public abstract class Attribute implements Serializable {
	private Vector index;

	public Vector getIndex() {
		return index;
	}

	public void setIndex(Vector index) {
		this.index = index;
	}
	
	abstract public boolean equals(Object o);
	
	abstract public Object clone();
	
	/**
	 * This method is called whenever the attribute appears in the consequence of 
	 * a rule. The method is passed an expression list of the rule's antecedents, 
	 * which it can use to determine the position of variables in the activating
	 * token. 
	 * 
	 * Constant attributes simply return a clone of themselves, while variables
	 * return a picker which extracts the correct value from the token. Compound
	 * attributes will create several dependent pickers.
	 */
	abstract public AttributePicker makeAttributePicker(ExpressionList facts);
	
	public abstract void accept(AttributeVisitor theVisitor);
	
	/**
	 * Traverses an abstract pattern (the receiver) to see whether it matches a concrete 
	 * instance (the argument). The parallel traversal of both hierarchies to see whether
	 * they match is required since CRIME supports data structures: for flat patterns, a
	 * filter node can content itself checking whether the correct constants are provided
	 * and to check whether multiple occurrences of the same variable yield the same value.
	 * <p>
	 * When using a data structure, filter nodes have to filter out compound attributes as
	 * well, implying that they need a mechanism to see whether a pattern such as 
	 * <tt>{ hobby: "magic" }</tt> matches with a concrete attribute such as <tt>{ name: 
	 * "Charlotte", hobby: "magic" }</tt>. The filtering of compound attributes is done by
	 * invoking this method.
	 * <p>
	 * Note that it is crucial to invoke the method on the abstract pattern. This ensures
	 * that an underspecified pattern (e.g. one that specifies only the hobby of a person)
	 * can still match with a complete pattern. Should this method be invoked on the 
	 * concrete instance instead, valid matches could be omitted. 
	 */
	public boolean matches(Attribute aConcreteAttribute) {
		return this.equals(aConcreteAttribute);
	}

	/**
	 * Identifies a nested attribute. Due to the introduction of datastructures, the index of 
	 * an attribute in a fact is now a more-dimensional index. The passed iterator allows one
	 * to check each index in turn.
	 * <p>
	 * Atomic attributes (which cannot be indexed further) will raise an error when the index
	 * that is given to them is not empty.
	 * 
	 * @param indexIterator an iterator on the more-dimensional index of an attribute
	 * @return the corresponding Attribute
	 */
	public Attribute getAttribute(Iterator indexIterator) throws IncorrectAttributeIndexException {
		if(indexIterator.hasNext()) {
			throw new IncorrectAttributeIndexException(indexIterator.next(), this);
		} else {
			return this;
		}
	}
}