package edu.vub.crime.grammar.attributes;

import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import edu.vub.crime.grammar.Expression;
import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.attributes.datastructures.Operator;
import edu.vub.crime.grammar.utils.AttributePicker;
import edu.vub.crime.grammar.utils.UtilsFactory;

/** <pre>
 * A Variable is a special kind of Attribute. 
 * The rule, given below, has five variables namely
 * ?genre, ?number, ?person, ?similarTastes, and ?number.
 * A variable is denoted by a question mark, followed by a name.
 * 
 *      likeMyMusic(?genre, ?number) :-
 *              genreILike(?genre),
 *              findall(?person, (
 *                         isPresent(?person, LivingRoom),
 *                         prefers(?person, ?genre)),
 *                      ?similarTastes),
 *              length(?similarTastes, ?number).
 * 
 * A variable can be bound, when a fact of the same type is entered to
 * the Rete network.
 * So, for example, when the fact genreILike(?Rock) is entered to the engine,
 * the token < add, genreILike(?Rock) > is added to the root node of the network.
 * 
 *     FILTER: genreILike           FILTER: isPresent           FILTER: prefers
 *             |                             |                          |
 *              \                    FILTER: LivingRoom                 |
 *               \                           |                          |
 *                \                           \                        /
 *                 \                           \                      /   
 *                  \                          JOIN: ?person = ? person     
 *                   \                                  |                   
 *                    \                                 /                    
 *                     \                               /                     
 *                      \                             /                      
 *                           JOIN: ?genre = ?genre      
 *                                    |                 
 *                                    |
 *                                 FINDALL
 *                                    |
 *                                    |
 *                                  LENGTH
 *                                    |
 *                                    |
 *                                PRODUCTION
 * 
 * This token will be passed through the network, depicted above, and it will pass
 * the filter node which filters on "genreILike".  
 * The variable ?genre will now be bound to the value "Rock".
 * </pre> */
public class Variable extends Attribute {
	
	private static int theGeneratedVariablesCount = 0;
	
	public static Variable generate() {
		return new Variable("_" + theGeneratedVariablesCount++);
	}
	
	/** <pre>
	 * TYPE : 2
	 * name_ : the name of the variable
	 * value_ : the value of the variable
	 * bound_ : boolean, true when variable is bound, false otherwise
	 * </pre> */
	public static final int TYPE = 2;
	protected String name_;
	private Attribute value_;
	private boolean bound_;
	
	private Operator itsConstraint;

	/** <pre>
	 * Constructor.
	 * </pre> */
	public Variable(String name) {
		name_ = name;
		value_ = null;
		bound_ = false;
		itsConstraint = Operator.EQL;
	}

	/** <pre>
	 * Accessor for value_ datamember.
	 * </pre> */
	public Attribute getValue() {
		return value_;
	}

	/** <pre>
	 * Accessor for datamember bound_.
	 * </pre> */
	public boolean isBound() {
		return bound_;
	}


	/** <pre>
	 * Accessor for datamember name_.
	 * @return
	 * </pre> */
	public String getName() {
		return name_;
	}

	/** <pre>
	 * Mutator for datamember name_.
	 * @param name
	 * </pre> */
	public void setName(String name) {
		name_ = name;
	}
	
	public void setIndex(Vector index) {
		if(getValue() != null)
			getValue().setIndex(index);
		
		super.setIndex(index);
	}
	
	/** <pre>
	 * Mutator for value_ datamember.
	 * </pre> */
	public void setValue(Attribute value) {
		if(value != null) {
			bound_ = true;
		}
		else {
			bound_ = false;
		}			
		value_ = value;
		
		if(value_.getIndex() == null)
			value_.setIndex(getIndex());
	}
	
	/** <pre>
	 * CompareTo nethod returns an integer.
	 * </pre> */
	public int compareTo(Object variable) {
		return name_.compareTo(((Variable) variable).getName());
	}
	
	/** <pre>
	 * Checks if a given object is equal to this variable.
	 * A boolean is returned, true when they 're equal, false otherwise.
	 * </pre> */
	public boolean equals(Object possibleEqual) {
		try {
			Variable asVariable = (Variable)possibleEqual;
			return 
				name_.equals(asVariable.getName()) /* &&
				itsConstraint.equals(asVariable.getConstraint()) */ ;
		}
		catch(ClassCastException exception){
			return false;
		}
	}
	
	/** <pre>
	 * Get the hash code.
	 * </pre> */
	public int hashCode(){
		return name_.hashCode();
	}
	
	/** <pre>
	 * Return string (question mark followed by the name).
	 * </pre> */
	public String toString(){
		return "?" + name_;
	}
	
	/** <pre>
	 * Clone the variable.
	 * Returns an object.
	 * </pre> */
	public Object clone() {		
		 Attribute clone = ExpressionFactory.variable(name_);
		 clone.setIndex(getIndex());
		 return clone;
	}

	/** <pre>
	 * This method makes a new var picker (special kind of attribute picker)
	 * given some expressions.
	 * When the variable of one of these expressions equals this variable,
	 * the var picker is instantiated and returned.
	 * When no such variable is present, null is returned.
	 * </pre> */
	public AttributePicker makeAttributePicker(ExpressionList expressions) {
		int factIndex = 0;
		for(Iterator expItr=expressions.iterator(); expItr.hasNext(); ) {
			Expression expression = (Expression) expItr.next();	
			Set factVars = expression.variables();			
			for(Iterator varItr=factVars.iterator(); varItr.hasNext(); ) {
				Variable variable = (Variable) varItr.next();	
				if(variable.equals(this)) {
					return UtilsFactory.varPicker(this, factIndex, variable);
				}
			}
			factIndex += expression.tokenSize();
		}
		return null;
	}

	/** <pre>
	 * Serialize.
	 * </pre> */
	public String serialize() {
		return Integer.toString(Variable.TYPE) + "@" + name_ + " ";
	}
	
	/**
	 * Inform the visitor of this object's concrete type
	 */
	public void accept(AttributeVisitor theVisitor) {
		theVisitor.visit(this);
	}
	
	/**
	 * In pattern matching, where the matching occurs against a concrete instance
	 * which can therefore not contain any variables, a variable in the pattern 
	 * can always match.
	 * <p>
	 * Note that checks to see whether a variable which occurs multiple times in a
	 * single pattern (in this case in a single fact) always matches with the same
	 * value are done independently of this match function.
	 */
	public boolean matches(Attribute concreteAttribute) {
		return true;
	}
	
	public Operator getConstraint() {
		return itsConstraint;
	}

	public void setConstraint(Operator theConstraint) {
		itsConstraint = theConstraint;
	}
}