package edu.vub.crime.grammar.attributes.datastructures;

import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.AttributeVisitor;
import edu.vub.crime.grammar.attributes.CompoundAttribute;
import edu.vub.crime.grammar.attributes.Variable;
import edu.vub.crime.grammar.utils.AttributePicker;
import edu.vub.crime.grammar.utils.PrintingVisitor;
import edu.vub.crime.grammar.utils.UtilsFactory;
import edu.vub.crime.util.Logging;

/** 
 * List class.
 * @brief This class is a CompoundAttribute. 
 * Each list has a head and a tail. Head and tail are attributes, since
 * the tail can be a variable or a list.
 * 
 * The list can take prolog syntax.
 *  
 * for example someFact([1,2,[3|[4,5]]]). 
 * In this example 
 * 	+List head 1
 * 			tail    List	head 2
 * 							tail    List	head	List	head	3
 *   											tail	[]		tail	List	head 4
 *   																			tail	List	head	5
 *   																							tail	[]
 * In this example the index is defined as follows:
 * [1,2,[3,4,5]] 	-> 0 
 * 1					-> 0-0
 * [2,[3,4,5]]		-> 0-1
 * 2					-> 0-1-0
 * [[3,4,5]]			-> 0-1-1
 * [3,4,5]			-> 0-1-1-0
 * []					-> 0-1-1-1
 * 3					-> 0-1-1-0-0
 * [4,5]				-> 0-1-1-0-1
 * 
 * 
 */			
public class List extends CompoundAttribute {
	public static final int HEAD = 0;
	public static final int TAIL = 1;
	
	private Attribute itsHead = null;
	private Attribute itsTail = null;
	
	/**
	 * Default Constructor. Creates an empty list.
	 */
	public List() {
		isNil	= true;
	}

	List(Attribute head){
		itsHead	= (Attribute)head.clone();
		itsTail	= new List();
		isNil	= false;
	}

	/**
	 * This method is called when the parser encounters a pipe (<tt>|</tt>) while reading 
	 * a list expression. According to the grammar specifications, the subsequent token
	 * must be a single attribute, which is passed to this function.
	 * <p>
	 * We distinguish three cases: <ul>
	 * <li> The attribute is a List attribute (e.g. <tt>[1 | [2]]</tt>) which implies that
	 * 		the list is simply extended with more elements.
	 * <li> The attribute is a Variable attribute (e.g. <tt>[1 | ?T]</tt>) which can match
	 * 		with an arbitrary amount of subsequent elements.
	 * <li>	The attribute is a Constant attribute (e.g <tt>[1 | 2]</tt>) in which case a
	 * 		precise match will be required.
	 * </ul>
	 * @param theRestAttribute of the list expression being parsed
	 */
	public void setTail(Attribute theRestAttribute) {
		// If the parser enforces correct semantics, the list on which setTail is called should 
		// never be nil. Moreover, when propagating the setTail invocation we always ensure that
		// the process stops is the tail is nil. Nevertheless, we provide a sanity check.
		if(this.isNil) {
			Logging.Engine_LOG.fatal("Invoked setTail on an empty list");
			throw new RuntimeException("Invoked setTail on an empty list");
		}
		
		// The list is constructed as a sequence of cons-cells, if this List instance is not
		// the last one, propagate the setTail call down the cdr-chain.
		List cdrChain = (List) itsTail;
		if(! cdrChain.isNil) {
			 cdrChain.setTail(theRestAttribute);
		} else {
			if(theRestAttribute instanceof List) {
				List asList = (List)theRestAttribute;
				
				for (Iterator elementIt = asList.iterator(); elementIt.hasNext();) {
					Attribute currentAttribute = (Attribute) elementIt.next();
					
					add(currentAttribute);
				}
				
				return;
			}
			
			if(theRestAttribute instanceof Variable) {
				RestVariable asRestVariable = new RestVariable((Variable)theRestAttribute);
				itsTail = asRestVariable;
				addAttribute(asRestVariable);
				return;
			};

			
			/* else */
			cdrChain.add(theRestAttribute);
		}
		
	}
	
	public Attribute getTail(){
		return itsTail;
	}
	
	public Attribute getHead(){
		return itsHead;
	}
	
	public void add(Attribute theAttribute) {
		if(isNil){
			itsHead = theAttribute;
			itsTail = new List();
			isNil = false;
		}else {
			List list = (List)itsTail;
			if(list.isNil)
				itsTail = new List(theAttribute);
			else{
				list.add(theAttribute);	
			}
		}	
		addAttribute(theAttribute);
	}
	
	public void remove(Attribute theAttribute) {
		if(itsHead.equals(theAttribute)) {
			if(itsTail instanceof List) {
				List theRemainder = (List)itsTail;
				
				if(theRemainder.isNil) {
					itsHead = null;
					itsTail = null;
					isNil	= true;
				} else {
					itsHead = theRemainder.itsHead;
					itsTail = theRemainder.itsTail;
				}
			} else {
				itsHead = itsTail;
				itsTail = new List();
			}
		}
			
		super.remove(theAttribute);
	}
	
	public int length() {
		return itsAttributes.size();
	}

	public boolean equals(Object o) {
		if(o == null) return false;
		try{
			List list = (List)o;
			if(isNil && (list.isNil)){
				return true;
			}else 	
				if(!isNil && !(list.isNil))
					return (itsHead.equals(list.getHead()) &&
						          itsTail.equals(list.getTail()));
			return false;		
		}
		catch(ClassCastException exception){
			return false;
		}
	}

	public Object clone() {
		List theClone;
		
		if(!isEmpty()) {
			theClone = new List(itsHead);
			theClone.setTail((Attribute)itsTail.clone());
			theClone.itsAttributes = (Vector)itsAttributes.clone();
		} else {
			theClone = new List();
		}
		
		return theClone;
	}
	
	public AttributePicker makeAttributePicker(ExpressionList facts) {
		return UtilsFactory.listPicker(facts, this);
	}
	
	public Attribute getAttribute(Iterator indexIterator) throws IncorrectAttributeIndexException {
		if(!indexIterator.hasNext()) 	return this;
		
		Attribute theResultingAttribute = null;
		
		Integer theListIndex = (Integer)indexIterator.next();
		
		switch(theListIndex.intValue()) {
			case HEAD:
				theResultingAttribute = itsHead;
				break;
			case TAIL:
				theResultingAttribute = itsTail;
				break;
			default:
				Logging.Engine_LOG.fatal("Invalid index in List:" + theListIndex);
				throw new IncorrectAttributeIndexException(theListIndex, this);
		};
		
		return theResultingAttribute.getAttribute(indexIterator);
	}
	
	public boolean matches(Attribute l){
		if(l==null) return false;
		try{
			boolean unify = true;
			List list = (List)l;
			if(isNil && (list.isNil))
				return unify;
			else 
			{	
				if(!isNil && !(list.isNil)){
					unify &= itsHead.matches(list.getHead());
					unify &= itsTail.matches(list.getTail());
				}
				else
					unify &= false;
			}
			return unify;
		}
		catch(ClassCastException exception){
			return false;
		}
	}
	
	/**
	 * Inform the visitor of this object's concrete type
	 */
	public void accept(AttributeVisitor theVisitor) {
		theVisitor.visit(this);
	}
	
	public String toString() {
		return PrintingVisitor.print(this);
	}
	
	public boolean isEmpty() {
		return isNil;
	}
	
	public Iterator iterator() {
		return new ListIterator(this);
	}
}

class ListIterator implements Iterator {
	
	private List itsListPosition;
	
	public ListIterator(List theStartingPoint) {
		itsListPosition = theStartingPoint;
	}
	
	public boolean hasNext() {
		return ! itsListPosition.isEmpty();
	}
	
	public Object next() {
		Object theResult = itsListPosition.getHead();
		
		Attribute theNextValue = itsListPosition.getTail();
		
		if (theNextValue instanceof List) {
			itsListPosition = (List)theNextValue;
		} else {
			itsListPosition = new List(theNextValue);
		}
		
		return theResult;
	}
	
	public void remove() {
		throw new RuntimeException("Cannot remove from a CRIME list.");
	}
}