package edu.vub.crime.grammar.attributes.datastructures;

import java.io.Serializable;

import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.constant.Fraction;
import edu.vub.crime.grammar.attributes.constant.Number;

/**
 * The operator class allows testing for the (in)equality of two attributes. The class provides an
 * abstract evaluate method which performs the test. The class uses the flyweight design pattern by
 * instantiating different concrete instances which may be of use in the system. 
 * 
 * @author smostinc
 */
public abstract class Operator implements Serializable {
		
	/** 
	 * This method is used by the parser to translate the parsed relational operator into
	 * a concrete instance of the Operator flyweight.
	 */
	public static Operator getOperator(String theParsedOperator) {
		if(theParsedOperator.equals(EQL.itsRepresentation)) return EQL;
		if(theParsedOperator.equals(NEQ.itsRepresentation)) return NEQ;
		if(theParsedOperator.equals(LTH.itsRepresentation)) return LTH;
		if(theParsedOperator.equals(GTH.itsRepresentation)) return GTH;
		if(theParsedOperator.equals(LEQ.itsRepresentation)) return LEQ;
		if(theParsedOperator.equals(GEQ.itsRepresentation)) return GEQ;
		
		throw new RuntimeException("Invalid operator specification");
	}
	
	public static final Operator EQL = new Operator("=") {
		public boolean evaluate(Attribute theLeftAttribute, Attribute theRightAttribute) {
			return theLeftAttribute.equals(theRightAttribute);
		};
		
		public Operator invert() { return EQL; }
		public Operator negate() { return NEQ; }
	};
	
	public static final Operator NEQ = new Operator("!=") {
		public boolean evaluate(Attribute theLeftAttribute, Attribute theRightAttribute) {
			return ! theLeftAttribute.equals(theRightAttribute);
		};
		
		public Operator invert() { return NEQ; }
		public Operator negate() { return EQL; }
	};
	
	public static final Operator LTH = new Operator("<") {
		public boolean evaluate(Attribute theLeftAttribute, Attribute theRightAttribute) {
			return getValue(theLeftAttribute) < getValue(theRightAttribute);
		};
		
		public Operator invert() { return GTH; }
		public Operator negate() { return GEQ; }
	};
	
	public static final Operator GTH = new Operator(">") {
		public boolean evaluate(Attribute theLeftAttribute, Attribute theRightAttribute) {
			return getValue(theLeftAttribute) > getValue(theRightAttribute);
		};
		
		public Operator invert() { return LTH; }
		public Operator negate() { return LEQ; }
	};
	
	public static final Operator LEQ = new Operator("<=") {
		public boolean evaluate(Attribute theLeftAttribute, Attribute theRightAttribute) {
			return 
				EQL.evaluate(theLeftAttribute, theRightAttribute) ||
				LTH.evaluate(theLeftAttribute, theRightAttribute);
		};
		
		public Operator invert() { return GEQ; }
		public Operator negate() { return GTH; }
	};
	
	public static final Operator GEQ = new Operator(">=") {
		public boolean evaluate(Attribute theLeftAttribute, Attribute theRightAttribute) {
			return 
				EQL.evaluate(theLeftAttribute, theRightAttribute) ||
				GTH.evaluate(theLeftAttribute, theRightAttribute);
		};
		
		public Operator invert() { return LEQ; }
		public Operator negate() { return LTH; }
	};
	
	private final String itsRepresentation;
	
	public Operator(String theRepresentation) {
		itsRepresentation = theRepresentation;
	}
	
	public String toString() {
		return itsRepresentation;
	}
	
	public abstract boolean evaluate(Attribute theLeftAttribute, Attribute theRightAttribute);
	
	public abstract Operator invert();
	public abstract Operator negate();
	
	protected double getValue(Attribute theNumericalAttribute) {
		if(theNumericalAttribute instanceof Number) {
			return ((Number)theNumericalAttribute).getValue();
		}
		if(theNumericalAttribute instanceof Fraction) {
			return ((Fraction)theNumericalAttribute).getValue();
		}
		throw new RuntimeException(theNumericalAttribute + "is not a number");
	}

	public Object readResolve() {
		return Operator.getOperator(itsRepresentation);
	}
}
