package edu.vub.crime.grammar.attributes.datastructures;


import java.util.Iterator;
import java.util.Vector;


import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.AttributeVisitor;
import edu.vub.crime.grammar.attributes.CompoundAttribute;
import edu.vub.crime.grammar.attributes.Variable;
import edu.vub.crime.grammar.utils.AttributePicker;

import edu.vub.crime.grammar.utils.UtilsFactory;

/**
 * A pair is formed by a symbol that represents the name,
 * an operator and an attribute that it is the value.
 * 
 * 	person({name:Bob}
 * 
 * In this example:
 * 	+ name is the symbol
 * 	+ :		is the operator
 * 	+ Bob   is the value
 * @author mhernand
 *
 */
public class Pair extends CompoundAttribute {
	private String name;
	private Attribute value;
	private Operator operator;
	private Attribute constraint;
	
	public Pair(String theName, Attribute theValue) {
		this(theName, Operator.EQL, theValue);
	}
	
	public Pair(String name, Operator operator, Attribute value){
		this.name = name;
		this.operator = operator;
		this.value = value;
		if(value instanceof Variable && !(operator.equals(Operator.EQL))){
			Variable variable = (Variable)value;
			variable.setConstraint(operator);
		}
		if(value instanceof Variable)
		{
			constraint = value;
		}
		addAttribute(this.value);
	}
	
	public String toString(){ 
		return name+operator+value;
	}
	
	public Object clone() {
		Pair theClonedPair = new Pair(name, operator, value);
		theClonedPair.setIndex(getIndex());

		return theClonedPair;
	}

	public boolean equals(Object o) {
		if(o == null) return false;
		try{
				boolean equal = true;
				Pair p = (Pair)o;
				equal &= this.name.equals(p.getName());
				equal &= this.value.equals(p.getValue());
				return equal;
		}
		catch(ClassCastException exception){
			return false;
		}
	}
	
	public boolean matches(Attribute o){
		if(o == null) return false;
		try{
				boolean unify = true;
				Pair p = (Pair)o;
				unify &= this.name.equals(p.getName()); 
				if(value instanceof CompoundAttribute)
					unify &= value.matches(p.getValue());
				else
					if(!(value instanceof Variable))
						unify &= operator.evaluate(p.getValue(), value);
				return unify;
		}
		catch(ClassCastException exception){
			return false;
		}
	}

	public AttributePicker makeAttributePicker(ExpressionList facts) {
		return UtilsFactory.pairPicker(facts, this);
	}

	public String getName() {
		return name;
	}

	public Operator getOperator() {
		return operator;
	}

	public Attribute getValue() {
		return value;
	}

	public Attribute getAttribute(Iterator itr) {
		try{
			if (! itr.hasNext()){
				return value;
			}
			if( value instanceof CompoundAttribute)
				return ((CompoundAttribute)value).getAttribute(itr);
		}catch(Exception e){return null;}
		return null;
	}
	
	//public Iterator iterator(){return null;};

	public void setOperator(Operator operator) {
		this.operator = operator;
		if(value instanceof Variable)
			((Variable)value).setConstraint(operator);
	}

	public void setValue(Attribute value) {
		this.value = value;
		this.itsAttributes.clear();
		addAttribute(this.value);
	}
	
	public void setIndex(Vector index) {
		super.setIndex(index);
		
		if(value != null) 
			value.setIndex(index);
	}

	public Attribute getConstraint() {
		return constraint;
	}
	
	public boolean isConstrained() {
		return (constraint != null);
	}
		
	/**
	 * Inform the visitor of this object's concrete type
	 */
	public void accept(AttributeVisitor theVisitor) {
		theVisitor.visit(this);
	}
}
