/**
 * 
 */
package edu.vub.crime.grammar.attributes.datastructures;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.AttributeVisitor;
import edu.vub.crime.grammar.attributes.CompoundAttribute;
import edu.vub.crime.grammar.attributes.constant.Type;
import edu.vub.crime.grammar.utils.AttributePicker;
import edu.vub.crime.grammar.utils.UtilsFactory;

/**
 * A typed object attribute is of the form &lt;type&gt;(&lt;a list of pairs&gt;). It will match any typed object
 * whose type is a subtype of the given type and who has matching attributes for all the listed pairs.
 * 
 * @author smostinc
 */
public class TypedObject extends CompoundAttribute {

	public static final String	SELF_NAME  = "self";
	
	private final	Attribute	itsType;
	private final	HashMap		itsFields;
	
	protected		Vector		itsConstraints;
	
	public TypedObject(Attribute theType) {
		itsType		= theType;
		itsFields	= new HashMap();
	}

	public int size() {
		return itsFields.size();
	}
	
	public void addField(Pair theField) {
		String theFieldName = theField.getName();
		
		itsFields.put(theFieldName, theField);
		itsVariables.addAll(theField.getVariables());
		
		if(theField.isConstrained())
			addConstraint(theField);
	}
	
	public void setField(String theFieldName, Attribute theValue) {
		itsFields.remove(theFieldName);
		addField(new Pair(theFieldName, theValue));
	}
	
	public void addConstraint(Pair theConstraint) {
		// Not sure whether this is actually necessary information
	}

	public Iterator iterator() {
		return itsFields.values().iterator();
	}
	
	public String toString(){
		StringBuffer theBuffer = new StringBuffer(64);
		theBuffer.append(itsType);
		
		theBuffer.append('(');
		
		Iterator fieldIt = iterator();
		for(int fieldIndex = itsFields.size(); fieldIndex != 0; fieldIndex--) {
			Pair currentField = (Pair) fieldIt.next();
			
			theBuffer.append(currentField);
			
			if(fieldIndex != 1)
				theBuffer.append(", ");
		}
		
		theBuffer.append(")");

		return theBuffer.toString();
	}
	
	public Attribute getAttribute(Iterator indexIterator) throws IncorrectAttributeIndexException {
		if(indexIterator.hasNext()) {
			Object	theIndex	= indexIterator.next();
			
			if(theIndex.equals(SELF_NAME))
				return this.getAttribute(indexIterator);
			
			Pair 	theField	= getField(theIndex);
			
			if(theField == null) {
				throw new IncorrectAttributeIndexException(theIndex, this);
			} else {
				return theField.getValue().getAttribute(indexIterator);
			}
		}
		
		return this;
	}

	public void accept(AttributeVisitor theVisitor) {
		theVisitor.visit(this);
	}

	public Object clone() {
		TypedObject theResult = new TypedObject((Type) itsType.clone());
		
		for (Iterator fieldIt = iterator(); fieldIt.hasNext();) {
			Pair currentField = (Pair) fieldIt.next();
			
			theResult.addField((Pair) currentField.clone());
		}

		return theResult;
	}

	public boolean equals(Object anotherObject) {
		if(anotherObject instanceof Type) return itsType.equals(anotherObject);
		
		if(! (anotherObject instanceof TypedObject))
			return false;
		
		TypedObject anotherTypedObject = (TypedObject) anotherObject;
		
		if(! itsType.equals(anotherTypedObject.getType()))
			return false;
		
		if(size() != anotherTypedObject.size())
			return false;
		
		for (Iterator fieldIt = iterator(); fieldIt.hasNext();) {
			Pair currentField = (Pair) fieldIt.next();
			
			Pair hisCorrespondingField = anotherTypedObject.getField(currentField.getName());
			
			if(	hisCorrespondingField == null ||
				!hisCorrespondingField.equals(currentField))
				return false;
		}

		return true;
	}

	/* (non-Javadoc)
	 * @see edu.vub.crime.grammar.attributes.Attribute#makeAttributePicker(edu.vub.crime.grammar.ExpressionList)
	 */
	public AttributePicker makeAttributePicker(ExpressionList facts) {
		return UtilsFactory.objectPicker(facts, this);
	}
	
	public Attribute getType() {
		return itsType;
	}
	
	private Pair getField(Object theName) {
		return (Pair) itsFields.get(theName);
	}
	
	public void setIndex(Vector index) {
		super.setIndex(index);
		
		for (Iterator attributeIterator = itsFields.entrySet().iterator(); attributeIterator.hasNext();) {
			Map.Entry 	currentField	= (Map.Entry) attributeIterator.next();
			
			String		currentName		= (String)	  currentField.getKey();
			Attribute	currentValue	= (Attribute) currentField.getValue();
			
			Vector 		currentIndex	= (Vector)	  index.clone();

			currentIndex.add(currentName);
			currentValue.setIndex(currentIndex);
		}
	}

}
