package edu.vub.crime.grammar.operators;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import edu.vub.crime.grammar.AccumulatingPrimitive;
import edu.vub.crime.grammar.Expression;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.GrammarVisitor;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.Variable;
import edu.vub.crime.rete.nodes.Node;


/** <pre>
 * The bagof class implements the building of the network for it's statement.
 * The general form of bagof is:
 * 
 *  expression one,
 *  ...,
 *  expression x-1,
 * 	bagof(?variable, list of facts, ?accumulator).
 *  expression x+1,
 *  ...
 *  expression n.
 * 
 *  For the list of facts there will be built a rete network up until the second layer
 *  but without the production node.
 *           .
 *           :
 *         FILTER FN-1   FILTER FN
 *               \       /
 *               JOIN LIST
 *                  |
 *                  :
 *                  .
 *                  
 *  This network then is joined with the previous network 
 *  that is built from the expressions one up until x-1.
 *           .
 *           :
 *           |   
 *      Filter EXP1 X-2  FILTER EXP X-1   JOIN LIST
 *           \              /              / 
 *                JOIN EXP                /
 *                       \               /
 *                         JOIN EXP_LIST
 *                              |
 *                              :
 *                              .
 *                          
 *  The output of that node is then connected with a bagofNode.
 *  
 *   JOIN EXP_LIST
 *        |
 *     BAGOFNODE
 *        | 
 * </pre> */
public class BagOf extends AccumulatingPrimitive {
	
	private Set itsVariables;
	
	public boolean compileJoinConstraint(Variable theVariable, int theIndexIntheLeftToken, int theIndexInTheRightToken, Node theNodeInWhichToAdd) {
		boolean addedConstraint = false;
		
		for (Iterator expressionIt = itsQuery.iterator(); expressionIt.hasNext() && ! addedConstraint;) {
			Expression currentExpression = (Expression) expressionIt.next();
			
			addedConstraint = currentExpression.compileJoinConstraint(
					theVariable, theIndexIntheLeftToken++, 
					theIndexInTheRightToken, theNodeInWhichToAdd);
		}
		
		return addedConstraint;
	}

	public int tokenSize() {
		return 1;
		// return itsQuery.size(); 
	}
	
	/**
	 * Inform the visitor of this node's concrete type, by calling its visit method with the 
	 * correct overloaded signature.
	 */
	public void accept(GrammarVisitor visitor) {
		visitor.visit(this);
	}
	
	public void initialize(Attribute theAccumulatedVariable, ExpressionList theQuery, Variable theAccumulator) {
		super.initialize(theAccumulatedVariable, theQuery, theAccumulator);

		itsVariables = new HashSet(itsQuery.variables());
		itsVariables.remove(itsAccumulatedVariable);
		itsVariables.add(itsAccumulator);
	}
	
	public Set variables() {
		return itsVariables;
	}
}