package edu.vub.crime.grammar.operators;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import edu.vub.crime.grammar.AccumulatingPrimitive;
import edu.vub.crime.grammar.Expression;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.GrammarVisitor;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.Variable;
import edu.vub.crime.rete.nodes.Node;

/** <pre>
 * The rule needed for explanation is given below.
 * 
 *         likeMyMusic(?genre, ?number) :-
 *              genreILike(?genre),
 *              findall(?person, (
 *                         isPresent(?person, LivingRoom),
 *                         prefers(?person, ?genre)),
 *                      ?similarTastes),
 *              length(?similarTastes, ?number).
 *
 * This rule calculates the number of people who like the same genre as I am in favour of.
 * This will be done for each genre I prefer.
 * 
 * The second condition in the right-hand-side of this rule is a findall statement.
 * This statement consists of an attribute, the variable ?person,
 * the queries isPresent(?person, LivingRoom) and prefers(?person, ?genre),
 * and an accumulator ?similarTastes.
 * </pre> */
public class FindAll extends AccumulatingPrimitive {
	
	private HashSet itsVariables = new HashSet(1);
	
	public boolean compileJoinConstraint(Variable theVariable, int theIndexIntheLeftToken, int theIndexInTheRightToken, Node theNodeInWhichToAdd) {
		boolean addedConstraint = false;
		
		for (Iterator expressionIt = itsQuery.iterator(); expressionIt.hasNext() && ! addedConstraint;) {
			Expression currentExpression = (Expression) expressionIt.next();
			
			addedConstraint = currentExpression.compileJoinConstraint(
					theVariable, theIndexIntheLeftToken++, 
					theIndexInTheRightToken, theNodeInWhichToAdd);
		}
		
		return addedConstraint;
	}
	
	/** <pre>
	 * Get the number of facts the query contains.
	 * In the example, this number will equal two. 
	 * </pre> */
	public int tokenSize() {
		return itsQuery.size(); 
	}
	
	/**
	 * Inform the visitor of this node's concrete type, by calling its visit method with the 
	 * correct overloaded signature.
	 */
	public void accept(GrammarVisitor visitor) {
		visitor.visit(this);
	}
	
	public void initialize(Attribute theAccumulatedVariable, ExpressionList theQuery, Variable theAccumulator) {
		super.initialize(theAccumulatedVariable, theQuery, theAccumulator);
		
		itsVariables.add(theAccumulator);
	}
	
	public Set variables() {
		return itsVariables;
	}

}
