package edu.vub.crime.grammar.operators;

import java.util.HashSet;
import java.util.Set;
import java.util.Stack;

import edu.vub.crime.grammar.Expression;
import edu.vub.crime.grammar.GrammarVisitor;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.Variable;
import edu.vub.crime.rete.CompileNetworkVisitor;
import edu.vub.crime.rete.nodes.LengthNode;

/**
 * Length is a grammar node which corresponds to "length(?list, ?length)", it can be used to bind a
 * variable with the length of a list or to verify whether a list has the expected length. However, 
 * it cannot be used to produce lists of a given length.  In other words, the first argument should
 * either be a literal list, or a variable that was previously bound.
 * <p>
 * This primitive operator is interpreted by a custom RETE node: {@link LengthNode}. This node is 
 * constructed by the {@link CompileNetworkVisitor}.
 * 
 * @see LengthNode
 * @see CompileNetworkVisitor#visit(Length);
 * 
 * @author cfscholl
 * @author ephilips
 * @author smostinc
 *
 */
public class Length extends Expression {

	private final Attribute itsGivenList;
	private final Attribute itsGivenLength;

	public Length(Attribute theGivenList, Attribute theGivenLength){
		itsGivenList 	= theGivenList;
		itsGivenLength	= theGivenLength;
		
		// Since the length is not visited by the indexing visitor and it may be a new variable 
		// bound by the length construct, set its index manually 
		Stack lengthIndex = new Stack();
		lengthIndex.push("length");
		
		itsGivenLength.setIndex(lengthIndex);
	}
	
	public Attribute getList()		{ return itsGivenList;  }
	public Attribute getLength()	{ return itsGivenLength; }
	
	/**
	 * Inform the visitor of this node's concrete type, by calling its visit method with the 
	 * correct overloaded signature.
	 */
	public void accept(GrammarVisitor visitor) {
		visitor.visit(this);
	}
	
	public Set variables() {
		if (itsGivenLength instanceof Variable) {
			HashSet theVariables = new HashSet(1);
			theVariables.add(itsGivenLength);
			
			return theVariables;
		} else return super.variables();
	}

}