package edu.vub.crime.grammar.utils;

import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.datastructures.List;
import edu.vub.crime.rete.data.Token;

/**
 * The ListPicker attribute picker provides a means to pass on a literal list of 
 * values on to an event. This ListPicker thus constructs a new list based on the
 * values of the attribute pickers it contains. Consider the following rule: 
 * <pre>
 * :log(["Critical mass to go", ?somewhere]) :-
 *   public -> letsGoTo(?a, ?somewhere),
 *   public -> letsGoTo(?b, ?somewhere),
 *   public -> letsGoTo(?c, ?somewhere),
 *   different([?a, ?b, ?c]).
 * </pre>
 * The <tt>:log</tt> event will then have a list picker for its first argument which
 * contains a constant picker for the passed string and a variable picker to find out
 * the preferred destination.
 * 
 * @author mhernand
 * @author smostinc
 */
public class ListPicker extends AttributePicker {

	private final Vector itsConstituentPickers = new Vector();
	
	/**
	 * Constructor which implicitly creates all constituent pickers for this List.
	 * 
	 * @param theFacts in the order in which they will be present in the tokens that will be passed to the picker
	 * @param theList for which the constructor will create an aggregated picker 
	 */
	public ListPicker(ExpressionList theFacts, List theList) {
		for(Iterator attributeIt = theList.iterator(); attributeIt.hasNext();) {
        	Attribute currentAttribute = (Attribute) attributeIt.next();
        	
        	itsConstituentPickers.add(currentAttribute.makeAttributePicker(theFacts));
        }
    }
	
	/**
	 * Create a list attribute containing values constructed from all aggregated 
	 * attribute pickers.
	 */
	public Attribute pickAttribute(Token theToken) throws IncorrectAttributeIndexException {
		List theResultingList = new List();
		
		for (Iterator constituentPickerIt = itsConstituentPickers.iterator(); constituentPickerIt.hasNext();) {
			AttributePicker currentAttributePicker = (AttributePicker) constituentPickerIt.next();
			
			// When the attribute picker picks a variable, it is crucial to know whether the
			// variable was an ordinary variable or a rest variable. Consider that so far a
			// list [ 1,2,3 ] was constructed and the variable picker returns [ 4 ]. If the
			// variable was an ordinary variable, the result is [ 1,2,3, [ 4 ]], whereas if
			// if was a rest variable, the result is [ 1,2,3,4 ].
			if(currentAttributePicker instanceof RestVariablePicker) {
				theResultingList.setTail(currentAttributePicker.pickAttribute(theToken));
			} else {
				theResultingList.add(currentAttributePicker.pickAttribute(theToken));
			}
		}
		
		return theResultingList;
	}
}
