package edu.vub.crime.grammar.utils;

import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.datastructures.Pair;
import edu.vub.crime.grammar.attributes.datastructures.TypedObject;
import edu.vub.crime.rete.data.Token;

/**
 * The ObjectPicker attribute picker provides a means to pass on a literal object 
 * expression on to an event. This ObjectPicker thus constructs a new object whose
 * fields are initialized with the values returned by the attribute pickers it 
 * contains. Consider the following rule: 
 * <pre>
 * :grade({ student: ?s; course: "Structure of Computer Programs"}) :-
 *   public -> enrolled(?s, ?program),
 *   requires(?program, "Structure of Computer Programs").
 * </pre>
 * The <tt>:grade</tt> event will then have an object picker for its first argument 
 * which contains a variable picker for the passed student and a constant picker to 
 * return the passed course title.
 * 
 * @author mhernand
 * @author smostinc
 */
public class ObjectPicker extends AttributePicker {

	private final AttributePicker	itsTypePicker;
	private final Vector 			itsConstituentPickers = new Vector();
	
	/**
	 * Constructor which implicitly creates all constituent pickers for this Object.
	 * 
	 * @param theFacts in the order in which they will be present in the tokens that will be passed to the picker
	 * @param theObject for which the constructor will create an aggregated picker 
	 */
	public ObjectPicker(ExpressionList theFacts, TypedObject theObject) {
		itsTypePicker = theObject.getType().makeAttributePicker(theFacts);
		
        for(Iterator attributeIt = theObject.iterator(); attributeIt.hasNext();) {
        	Attribute currentAttribute =(Attribute)attributeIt.next();
        	
        	itsConstituentPickers.add(currentAttribute.makeAttributePicker(theFacts));
        }
    }
	
	/**
	 * Create an object attribute containing pairs constructed from all aggregated 
	 * attribute pickers.
	 */
	public Attribute pickAttribute(Token theToken) throws IncorrectAttributeIndexException {
		TypedObject theResultingObject = new TypedObject(itsTypePicker.pickAttribute(theToken));
		
		for(Iterator pickersIt = itsConstituentPickers.iterator(); pickersIt.hasNext();) {
        	PairPicker currentPicker = (PairPicker) pickersIt.next();
        	
        	// Pair pickers return a newly constructed pair containing the name 
        	// as well as the picked attribute value.
        	theResultingObject.addField((Pair)currentPicker.pickAttribute(theToken));
        }
		return theResultingObject;
	}

}
