package edu.vub.crime.grammar.utils;

import java.util.Vector;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.rete.data.Token;

/** 
 * The VariablePicker is an attribute picker which returns the value associated to
 * a variable by looking at the first place (a fact & attribute index) in the token 
 * where this variable is expected to occur. This allows for rules such as :
 * <pre>
 * :showBuddy(?user) :-
 *   public -> colocated(me, ?user),
 *   buddy(?user).
 * </pre>
 * A possible token which could be fed to the attribute picker is then <tt>
 * &lt add, public -> colocated(me, Eline), buddy(Eline) &gt </tt> from which the
 * variable picker could then select the second attribute of the first fact.
 * <p>
 * Due to the introduction of datastructures in the language, the attribute index 
 * is not necessarily a number but it may be a more complex index. For instance,
 * consider that the user is in fact an object which describes the colocated user,
 * but we only want to show the name of our buddies when they come into range.
 * This can be achieved through a rule of the form :
 * 
 * <pre>
 * :showBuddy(?user) :-
 *   public -> colocated(me, { name: ?user }),
 *   buddy(?user).
 * </pre>
 * Here the attribute index is <tt>[ 2, "name" ]</tt>, since the variable is found 
 * by taking the second attribute (of the first fact) and then look for its <tt>
 * "name"</tt> field.
 * 
 * @author cfscholl
 * @author ephilips
 * @author mhernand
 */
public class VariablePicker extends AttributePicker {

    private final int 		itsFactIndex;
    private final Vector 	itsAttributeIndex;

    /** 
     * Constructor taking the indices of both the fact and the attribute where 
     * this picker may find the value of the variable it is supposed to extract
     * from a token. 
     */
    public VariablePicker(int thefactIndex, Vector theAttributeIndex) {
        itsFactIndex 		= thefactIndex;
        itsAttributeIndex 	= theAttributeIndex;
    }

    /** 
     * Extracts the value from a token at the given indices (which corresponds
     * to a variable in the original program).
     */
    public Attribute pickAttribute(Token token) throws IncorrectAttributeIndexException {
        GenericFact fact = token.getFact(itsFactIndex);
        return (Attribute) fact.getAttribute(itsAttributeIndex.iterator()).clone();
    }
}