package edu.vub.crime.memory;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.grammar.GenericFact;

/** 
 * LocalMemory implements the Memory interface for a single engine. It can be used as the
 * default memory when working in an unconnected setting. It is also used internally by 
 * the {@link NetworkMemory} which uses one LocalMemory per connected client.
 * 
 * @author cfscholl
 * @author ephilips
 * @author smostinc
 */
public class LocalMemory implements Memory {

	/**
	 * Maps names to the corresponding FactSpaces 
	 */
	private final HashMap itsMemory = new HashMap();
			
	/** 
	 * @return is there a fact space with the given name?
	 */
	public boolean isFactSpace(String space) {
		return itsMemory.containsKey(space);
	}
	
	/** 
	 * Removes a fact space from the memory. Note that the facts from that
	 * fact space are not automatically retracted.
	 */
	public void removeFactSpace(String space) {
		itsMemory.remove(space);
	}
	
	public void insert(GenericFact fact) {
		String space = fact.getSpace();
		if(isFactSpace(space)) {
			FactSpace factSpace = (FactSpace) itsMemory.get(space);
			factSpace.addFact(fact);
		}
		else {
			FactSpace factSpace = new FactSpace(space);
			factSpace.addFact(fact);
			itsMemory.put(factSpace.space(), factSpace);
		}
	}
	
	public void remove(GenericFact fact) {
		String space = fact.getSpace();
		if(isFactSpace(space)) {
			FactSpace factSpace = (FactSpace) itsMemory.get(space);
			factSpace.removeFact(fact);
			if(factSpace.size() == 0) {
				removeFactSpace(space);
			}
		}
	}

	/** 
	 * @return whether the given fact was previously committed to this memory.
	 */
	public boolean check(GenericFact fact) {
		FactSpace factSpace = (FactSpace) itsMemory.get(fact.getSpace());
		if((factSpace != null) && (factSpace.check(fact))) {
			return true;
		}
		
		return false;
	}
	
	/** 
	 * Forgetfulness: removes all facts from the memory. Note that these facts are
	 * not automatically retracted by the CRIME engine.
	 */
	public void clear() {
		itsMemory.clear();	
	}
	
	/** 
	 * Return all facts committed to this memory. Gathers all facts that are spread into
	 * the different fact spaces that make up this local memory.
	 */
	public Vector facts() {
		Vector result = new Vector();
		for(Iterator it=itsMemory.keySet().iterator(); it.hasNext(); ) {
			String space = (String) it.next();
			FactSpace factSpace = (FactSpace) itsMemory.get(space);
			Vector facts = factSpace.facts();
			result.addAll(facts);
		}
		return result;
	}
}
