package edu.vub.crime.memory;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.CrimeEngine;
import edu.vub.crime.grammar.GenericFact;

/** 
 * The NetworkMemory class implements the Memory interface and groups all facts
 * not only according to the fact spaces, but also according to the publishing
 * engine, which makes it possible to retract all facts published by an engine 
 * if it is no longer reachable.
 */
public class NetworkMemory implements Memory {

	/** 
	 * Maps a serializable id of the publishing engine to the associated local memory.
	 */
	private final HashMap 		itsConnectedEngines = new HashMap();
	
	/**
	 * Contains the facts published privately by this engine (which do not have a publisher)
	 */
	private final LocalMemory 	itsPrivateMemory = new LocalMemory();

	/** 
	 * @param thePublisher a serializable object that identifies the engine.
	 * @return a vector of all facts published by a certain client.
	 */
	public Vector facts(Serializable thePublisher) {
		LocalMemory theMemory = (LocalMemory) itsConnectedEngines.get(thePublisher);
		return theMemory!= null?theMemory.facts():new Vector();
	}
	
	/** 
	 * Retracts all facts that were known to be published by the publisher that
	 * is now being removed from the memory. 
	 */
	public void remove(Serializable thePublisher, CrimeEngine theEngine) {
	   	 Vector facts = this.facts(thePublisher);
         for(Iterator factIt = facts.iterator(); factIt.hasNext(); ) {
        	 GenericFact currentFact = (GenericFact) factIt.next();
        	 
        	 // Retract the fact and remove it from the underlying collection. 
             theEngine.retract(currentFact);
             factIt.remove();
         }
	}
	
	/**
	 * Commits the fact to the memory, and associated it with the proper publisher such
	 * that should this publisher disconnect, the fact can be properly retracted.
	 */
	public void insert(GenericFact fact) {
		LocalMemory theMemory;
		
		// If the fact is published in a public fact space
		if(fact.isPublic()) {
			// Try to identify whether we already have a memory dedicated to its publisher
			theMemory = (LocalMemory) itsConnectedEngines.get(fact.getPublisher());
			
			// If not, create and install one.
			if(theMemory == null) {
				theMemory = new LocalMemory();
				itsConnectedEngines.put(fact.getPublisher(), theMemory);
			}

		// Private facts on the other hand are stored separately
		} else {
			theMemory = itsPrivateMemory;
		}
		
		if(! theMemory.check(fact)) {
			theMemory.insert(fact);			
		}
	}
	
	public void remove(GenericFact fact) {
		LocalMemory theMemory;
		
		// If the fact is published in a public fact space
		if(fact.isPublic()) {
			// Try to identify which memory is dedicated to its publisher
			theMemory = (LocalMemory) itsConnectedEngines.get(fact.getPublisher());
			
		// Private facts on the other hand are stored separately
		} else {
			theMemory = itsPrivateMemory;
		}
		
		if(theMemory != null) {
			theMemory.remove(fact);			
		}
	}
}
