package edu.vub.crime.network;

import java.util.Iterator;

import edu.vub.crime.CrimeEngine;
import edu.vub.crime.CrimeFactory;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.rete.agenda.Activation;
import edu.vub.crime.rete.agenda.ExternalActivation;
import edu.vub.crime.rete.data.DataFactory;
import edu.vub.crime.rete.data.MemoryTable;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.util.Logging;

/**
 * The ConnectedCrimeEngine extends the default CrimeEngine and overrides methods which allow 
 * it to trap the assertion and retraction of qualified facts. When a rule activation derived
 * these assertions and retractions, they are to be propagated over the NetworkLayer to all
 * connected engines.
 * 
 * @author smostinc
 */
public class ConnectedCrimeEngine extends CrimeEngine {

	protected 	NetworkLayer 	layer_;
	protected	MemoryTable		toBePropagated_ = new MemoryTable();
	
	public ConnectedCrimeEngine() {
		layer_ 			= CrimeFactory.createNetworkLayer(this);
		//toBePropagated_	= new MemoryTable();
	}
	
	/**
	 * A ConnectedCrimeEngine resets its record of all public facts published by
	 * an activation whenever a new activation is started.
	 */
	public void beginActivation(Activation theActivation) {
		super.beginActivation(theActivation);
		
		toBePropagated_.clear();
	}
	
	/**
	 * A ConnectedCrimeEngine keeps a record of all assertions of shared facts 
	 * since these should be published to connected engines as soon as the current
	 * activation terminates.
	 */
	public void assertShared(GenericFact theFactToBePropagated) {
		super.assertShared(theFactToBePropagated);
		
		Token assertion = DataFactory.addToken(theFactToBePropagated);
		toBePropagated_.insert(assertion);
	}
		
	/**
	 * A ConnectedCrimeEngine keeps a record of all retractions of shared facts 
	 * since these should be published to connected engines as soon as the current
	 * activation terminates.
	 */
	public void retractShared(GenericFact theFactToBePropagated) {
		super.retract(theFactToBePropagated);

		Token retraction = DataFactory.removeToken(theFactToBePropagated);
		toBePropagated_.insert(retraction);
	}
	
	/**
	 * When the current activation ends, a ConnectedCrimeEngine should evaluate whether
	 * it needs to propagate information to the connected engines.
	 */
	public void endActivation(Activation theActivation) {
		super.endActivation(theActivation);
		
		// If the activation was the result of a local rule being triggered, all qualified facts
		// that were produced should be published to all connected CRIME engines.
		if (!(theActivation instanceof ExternalActivation) && (toBePropagated_.size() > 0)) {
			Logging.Network_LOG.info("ConnectedCrimeEngine is propagating the following operations to all connected engines:" + toBePropagated_);
			
			//ExternalActivation publishedActivation = new ExternalActivation(toBePropagated_, layer_.toString());
			//layer_.publishActivation(publishedActivation);
			
			for (Iterator tokenit = toBePropagated_.iterator(); tokenit.hasNext();) {
				Token currentToken = (Token) tokenit.next();
				
				GenericFact theFactToBePropagated = currentToken.getFact(0);
								
				layer_.publishActivation(
						new ExternalActivation(currentToken.getTag(), theFactToBePropagated, layer_.toString()));
			}
			
			toBePropagated_.clear();
		}
	}
}
