package edu.vub.crime.network.m2mi;

import java.io.Serializable;
import java.util.HashMap;

import edu.rit.m2mi.Eoid;
import edu.rit.m2mi.M2MI;
import edu.rit.util.Timer;
import edu.rit.util.TimerTask;
import edu.rit.util.TimerThread;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.network.DisconnectionObserver;
import edu.vub.crime.rete.agenda.Activation;

/**
 * The ConnectionManager contains a partial function, mapping all discovered objects (denoted by their
 * {@link Eoid}) to a {@link Timer} object which will trigger when the lease for that discovered object 
 * expires (and it is thus disconnected). The timers are reset every time a heartbeat message is received
 * from the discovered object.
 * 
 * @author smostinc
 *
 */
public class ConnectionManager implements HeartbeatRef, NetworkInterface {

	/**
	 * beatBoxMap_ serves as a partial function mapping {@link Eoid} objects to a corresponding 
	 * {@link Timer} object.
	 */
	private final HashMap beatBoxMap_ = new HashMap();
	
	/**
	 * observer_ is notified whenever an object has not sent any {@link HeartbeatRef#beat(Eoid)} messages
	 * for a period of {@link HeartbeatRef#LEASE_TIME} milliseconds.
	 */
	private DisconnectionObserver observer_ = null;
	
	private class RevokeLease implements TimerTask {
		private Eoid whichLease_;
		
		public RevokeLease(Eoid whichLease) {
			whichLease_ = whichLease;
		}
		
		public void action(Timer theTimer) {
			leaseTimedOut(theTimer, whichLease_);
		}
	}
	
	public ConnectionManager(DisconnectionObserver observer) {
		observer_ = observer;
		M2MI.export(this, HeartbeatRef.class);
		M2MI.export(this, NetworkInterface.class);
	}
	
	public synchronized void beat(Eoid aBox) {
		Timer lease = (Timer)beatBoxMap_.get(aBox);
		
		if(lease == null) {
			lease = TimerThread.getDefault().createTimer(
					new RevokeLease(aBox));
			
			beatBoxMap_.put(aBox, lease);
			
			if(observer_ != null)
				observer_.connectedSource(aBox.toString());
		}
		
		lease.start(LEASE_TIME);
	}
	
	/**
	 * When a CRIME engine publishes an activation, the connection manager will acknowledge this and
	 * will allow this message to prevent the lease of that engine to expire.
	 */
	public synchronized void receiveActivation(Eoid source, Activation tokens) {
		beat(source);
	}
	
	public void assert(Serializable theSender, GenericFact theFact, int totalRetractsSoFar, int theAssertionsID) {
		beat((Eoid) theSender);
	}

	public void retract(Serializable theSender, GenericFact theFact, int theRetractionsID, int totalAssertionsSoFar) {
		beat((Eoid) theSender);
	}
	/**
	 * This method is invoked by the {@link RevokeLease} auxiliary class when a particular Timer is 
	 * triggered. It will remove the lease on the {@link Eoid} object and notify its observer.
	 * 
	 * @param theTimer the timer which has triggered, thereby causing this method to be called
	 * @param theLease the eoid to which the timer was bound
	 */
	private synchronized void leaseTimedOut(Timer theTimer, Eoid theLease) {
		// parallel to the timer's invocation of the RevokeLease#action(Timer) method,
		// the DisconnectionManager may have received a beat(Eoid) message which has
		// reset the timer. In this case, the lease did not expire and the timer will
		// respond false when asked whether it was triggered.
		if(theTimer.isTriggered()) {
			theTimer = (Timer) beatBoxMap_.remove(theLease);
			
			// If the lease was really removed (i.e. this is no double remove)
			if (theTimer != null && observer_ != null) {
				observer_.disconnectedSource(theLease.toString());
			}
		}
	}
	
	
}
