package edu.vub.crime.network.m2mi;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;

import edu.rit.m2mi.Eoid;
import edu.rit.m2mi.M2MI;
import edu.rit.m2mi.Unihandle;
import edu.rit.util.Timer;
import edu.rit.util.TimerTask;
import edu.rit.util.TimerThread;
import edu.vub.crime.CrimeEngine;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.memory.NetworkMemory;
import edu.vub.crime.network.DisconnectionObserver;
import edu.vub.crime.network.NetworkLayer;
import edu.vub.crime.rete.agenda.Activation;
import edu.vub.crime.rete.agenda.ExternalActivation;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.rete.data.TokenTag;
import edu.vub.crime.util.Logging;


/**
 * The M2MIEngine class provides an implementation of the NetworkLayer which allows publishing (and receiving)
 * activations over the M2M protocol. This protocol allows creating handles which refer to a dynamic collection
 * of nearby CRIME engines.
 * 
 * @author smostinc
 */
public class M2MINetworkLayer extends NetworkLayer implements NetworkInterface, DisconnectionObserver {
	
	private final NetworkMemory theMemory;
	
	static {
		M2MI.initialize();
	}
	
	/**
	 * This omnihandle binds to all reachable Crime engines, such that they can be notified of
	 * new activations which are being published. 
	 */
	NetworkInterface nearbyEngines_ = (NetworkInterface) M2MI.getOmnihandle(NetworkInterface.class);
	
	/**
	 * This omnihandle binds to all reachable implementors of the HeartBeatRef interface: all 
	 * disconnection managers which are interested to detect the disconnection of previously 
	 * connected CRIME engines.  
	 */
	HeartbeatRef nearbyConnectionManagers_ = (HeartbeatRef) M2MI.getOmnihandle(HeartbeatRef.class);
		
	/**
	 * This handle corresponds to this CrimeEngine and its ID is used to identify the engine 
	 * while sending out heartbeat messages.
	 */
	Unihandle me_ = M2MI.getUnihandle(this, NetworkInterface.class);
	
	
	private static int REQUEST_LEASE_RENEWAL_TIME = HeartbeatRef.LEASE_TIME / 5;
	
	private Random randomGenerator_;
	private Timer leaseRenewalTimer_;
	
	/**
	 * Auxiliary class used to periodically report the presence of this M2MI CRIME engine
	 * 
	 * @author smostinc
	 */
	private class ReportPresence implements TimerTask {
		public void action(Timer theTimer) {
			beat(theTimer);
		}
	}
	
	public M2MINetworkLayer(CrimeEngine theEngine) {
		super(theEngine);
		
		theMemory = (NetworkMemory)theEngine.getMemory();
		
		randomGenerator_ = new Random();
		leaseRenewalTimer_ = TimerThread.getDefault()
			.createTimer(new ReportPresence());
		
		scheduleBeat();
	}
	
	/**
	 * Provides a mapping from fact space names to multihandles which determine the scope of distribution
	 * for qualified facts published in this fact space. If no NetworkInterface is found in this mapping,
	 * the default nearbyEngines will be used.
	 */
	private Map itsSpecializedHandles = new HashMap();
	
	public void publishActivation(Activation tokens) {
		for (Iterator tokenIt = tokens.iterator(); tokenIt.hasNext();) {
			Token currentToken = (Token) tokenIt.next();
			GenericFact currentFact = currentToken.getFact(0); 
			
			NetworkInterface theHandle = (NetworkInterface)itsSpecializedHandles.get(currentFact.getSpace());
			if(theHandle == null)
				theHandle = nearbyEngines_;
			
			switch (currentToken.getTag()) {
			case TokenTag.add:
				theHandle.assert(me_.getEoid(), currentFact, 0, 0);
				break;

			case TokenTag.remove:
				theHandle.retract(me_.getEoid(), currentFact, 0, 0);
				break;
			}
		}
	}
	
	public void assert(Serializable theSender, GenericFact theFact, int totalRetractsSoFar, int theAssertionsID) {
		Logging.M2MI_LOG.info("Receiving assertion from " + theSender);
		
		ExternalActivation theActivation = new ExternalActivation(TokenTag.add, theFact, theSender.toString());
		
		if(theSender != me_.getEoid())
			itsEngine.insertActivation(theActivation);
	}
	
	public void retract(Serializable theSender, GenericFact theFact, int theRetractionsID, int totalAssertionsSoFar) {
		Logging.M2MI_LOG.info("Receiving retraction from " + theSender);
		
		ExternalActivation theActivation = new ExternalActivation(TokenTag.remove, theFact, theSender.toString());
		
		if(theSender != me_.getEoid())
			itsEngine.insertActivation(theActivation);
	}
	
	public void disconnectedSource(Serializable source) {
		Logging.M2MI_LOG.info("A possible source of context information disconnected : " + source);
		
		theMemory.remove(source, itsEngine);
	}
	
	
	public void connectedSource(Serializable source) {
		Logging.M2MI_LOG.info("A possible source of context information was discovered : " + source);
		
		// TODO
	}
	

	/**
	 * Send out a heartbeat message to all connected ConnectionManagers.
	 * 
	 * @param theTimer
	 */
	private synchronized void beat(Timer theTimer) {
		nearbyConnectionManagers_.beat(me_.getEoid());

		scheduleBeat();
	}
	
	/**
	 * Schedules sending out a heartbeat message for some random time between 1 and 2 times the lease renewal time
	 */
	private void scheduleBeat() {
		// timeout to request a new lease is at least 0.2 * leaseTime and at most 0.4 * leaseTime
		long timeout = randomGenerator_.nextInt(REQUEST_LEASE_RENEWAL_TIME) + REQUEST_LEASE_RENEWAL_TIME;
		
		// restart the timer to renew the lease
		leaseRenewalTimer_.start(timeout);
	}
	
	public Eoid getPublishersId() {
		return me_.getEoid();
	}
}
