package edu.vub.crime.parser;

import java.util.Iterator;
import java.util.Stack;
import java.util.Vector;

import edu.vub.crime.dataflow.DataFlowAttribute;
import edu.vub.crime.grammar.AttributedExpression;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.AttributeVisitor;
import edu.vub.crime.grammar.attributes.ConstantAttribute;
import edu.vub.crime.grammar.attributes.Variable;
import edu.vub.crime.grammar.attributes.datastructures.List;
import edu.vub.crime.grammar.attributes.datastructures.Pair;
import edu.vub.crime.grammar.attributes.datastructures.TypedObject;

/**
 * The IndexingVisitor computes for each {@link Attribute} what their precise index in the 
 * fact is. This index can be a simple number, if the fact contains no data structure, but
 * it can be nested and/or contain names.
 * 
 * @author smostinc
 */
public class IndexingVisitor extends AttributeVisitor {

	private static final Integer LIST_HEAD = new Integer(0);
	private static final Integer LIST_TAIL = new Integer(1);
	
	private AttributedExpression itsParentExpression;
	private Stack 		 		 itsCurrentIndex = new Stack();
	
	public IndexingVisitor(AttributedExpression theParentExpression) {
		super("index", "");
		
		// this expression will accumulate all attributes
		itsParentExpression = theParentExpression;
		
		itsCurrentIndex.push("dummy");
	}
	
	/**
	 * Interface method invoked by the parser while it is walking the Attributes of 
	 * a particular expression. By specifying an explicit entry point we can do some 
	 * additional checks, automatically augment the base index and add the fact to its 
	 * parent expression.
	 * 
	 * @param theAttribute the new attribute to be indexed by the visitor.
	 */
	public void nextAttribute(String theName, Attribute theAttribute) {
		itsCurrentIndex.pop();
		itsCurrentIndex.push(theName);
		
		theAttribute.accept(this);
		
		itsParentExpression.addNamedAttribute(theName, theAttribute);
	}
	
	private void setCurrentIndex(Attribute theAttribute, Vector theIndex) {
		theAttribute.setIndex(theIndex);
	}
	
	public void visit(TypedObject theTypedObject) {
		setCurrentIndex(theTypedObject, (Vector)itsCurrentIndex.clone());
		
		for(Iterator fieldIt = theTypedObject.iterator(); fieldIt.hasNext(); ) {
			Pair currentField = (Pair)fieldIt.next();
			
			currentField.accept(this);
		}
	}
	
	public void visit(List theList) {
		setCurrentIndex(theList, (Vector)itsCurrentIndex.clone());
		
		if(theList.isEmpty())
			return;
		
		// Traverse the head, and push an appropriate index to the index stack 
		itsCurrentIndex.push(LIST_HEAD);
		theList.getHead().accept(this);
		itsCurrentIndex.pop();
		
		itsCurrentIndex.push(LIST_TAIL);
		theList.getTail().accept(this);
		itsCurrentIndex.pop();
	}

	public void visit(Pair thePair) {
		itsCurrentIndex.push(thePair.getName());
		setCurrentIndex(thePair, (Vector)itsCurrentIndex.clone());
		
		// visit the Pair's value which contains additional attributes
		thePair.getValue().accept(this);
		itsCurrentIndex.pop();
	}
	
	public void visit(ConstantAttribute theConstant) {
		setCurrentIndex(theConstant, (Vector)itsCurrentIndex.clone());
	}
	
	public void visit(Variable theVariable) {
		setCurrentIndex(theVariable, (Vector)itsCurrentIndex.clone());
	}
	
	public void visit(DataFlowAttribute theFlow) {
		setCurrentIndex(theFlow, (Vector)itsCurrentIndex.clone());
	}	
}
