package edu.vub.crime.parser;

import java.io.ByteArrayInputStream;
import java.io.InputStream;

import junit.framework.TestCase;

import org.apache.regexp.RE;

import antlr.CommonAST;

public class ParserTest extends TestCase {
	
	/*
	 * Check whether the parse produces the correct ANTLR AST tree.
	 */
	private static void testParse(String parserInput, String expectedOutput) throws Exception {
		InputStream input = new ByteArrayInputStream(parserInput.getBytes());
		LexerImpl lexer = new LexerImpl(input);
		ParserImpl parser = new ParserImpl(lexer);
		parser.statement();
		CommonAST parseTree = (CommonAST)parser.getAST();
		System.out.println(parseTree.toStringList());
		
        assertEquals(new RE("\\s").subst(expectedOutput, ""), new RE("\\s").subst(parseTree.toStringList(), ""));
	}
	
	/*
	 * Check whether the parse correctly reports the input to be faulty
	 */
	private static void testParseError(String parserInput, String expectedErrorMessage) {
		try {
			InputStream input = new ByteArrayInputStream(parserInput.getBytes());
			LexerImpl lexer = new LexerImpl(input);
			ParserImpl parser = new ParserImpl(lexer);
			parser.program();
			CommonAST parseTree = (CommonAST)parser.getAST();
			System.out.println(parseTree.toStringList());
			fail("Expected to throw exception: " + expectedErrorMessage);
		} catch(Exception e) {
			assertEquals(expectedErrorMessage, e.getMessage());
		}
	}
	
	/**
	 * Tests the parsing of facts, events and rules which use simple attributes.
	 */
	public void testFactParsing() throws Exception {
		// Zero-arg unqualified fact
		testParse("alice().", "(assert (private fact (name alice)))");
		
		// Single-arg unqualified fact
		testParse("female(name = `alice).", "(assert (private fact (name female) (pair (name name) (operator =) (symbol alice))))");
		
		// Multi-arg unqualified fact
		testParse("parent(parent = `alice, child = `bob).", "(assert (private fact (name parent) (pair (name parent) (operator =) (symbol alice)) (pair (name child) (operator =) (symbol bob))))");
		
		// Multi-arg qualified fact
		testParse("location -> isPresent(person = `alice, place = `living_room).", "(assert (public fact (space location) (name isPresent) (pair (name person) (operator =) (symbol alice)) (pair (name place) (operator =) (symbol living_room))))");
				
		// A simple rule
		testParse("mother(mother = ?X, child = ?Y) :- female(person = ?X), parent(parent = ?X, child = ?Y).", "(rule (private fact (name mother) (pair (name mother) (operator =) (variable X)) (pair (name child) (operator =) (variable Y))) (condition (private fact (name female) (pair (name person) (operator =) (variable X))) (private fact (name parent) (pair (name parent) (operator =) (variable X)) (pair (name child) (operator =) (variable Y )))))");
	}
	
	/**
	 * Tests whether rules with more than one effect are parsed correctly.
	 */
	public void testComplexRuleParsing() throws Exception {
		testParse("PointOfInterest(x=?x, y=?y), :AddToMap(x=?x, y=?y), :AddDetail(x=?x, y=?y, detail=?annotation) :- UserAnnotation(x=?x, y=?y, detail=?annotation)", 
				  "(rule (private fact (name PointOfInterest) (pair (name x) (operator =) (variable x)) (pair (name y)(operator =) (variable y))) (event (name AddToMap) (pair (name x) (operator =) (variable x)) (pair (name y)(operator =) (variable y))) (event (name AddDetail) (pair (name x) (operator =) (variable x)) (pair (name y)(operator =) (variable y)) (pair (name detail) (operator =) (variable annotation))) (condition (private fact (name UserAnnotation)(pair (name x) (operator =) (variable x)) (pair (name y)(operator =) (variable y)) (pair (name detail) (operator =) (variable annotation)))))");
		testParse(":AddToMap(x=?x, y=?y), PointOfInterest(x=?x, y=?y), :AddDetail(x=?x, y=?y, detail=?annotation) :- UserAnnotation(x=?x, y=?y, detail=?annotation)", 
		  "(rule (event (name AddToMap) (pair (name x) (operator =) (variable x)) (pair (name y)(operator =) (variable y))) (private fact (name PointOfInterest) (pair (name x) (operator =) (variable x)) (pair (name y)(operator =) (variable y))) (event (name AddDetail) (pair (name x) (operator =) (variable x)) (pair (name y)(operator =) (variable y)) (pair (name detail) (operator =) (variable annotation))) (condition (private fact (name UserAnnotation)(pair (name x) (operator =) (variable x)) (pair (name y)(operator =) (variable y)) (pair (name detail) (operator =) (variable annotation)))))");
	}
	
	/**
	 * Tests the parsing of facts, events and rules which use compound attributes.
	 */
	public void testCompoundFactParsing() throws Exception {
		// Test empty list parsing
		testParse("attendants(list = []).", "( assert ( private fact (name attendants) (pair (name list) (operator =) (null list))))");
		
		// Test literal list parsing
		testParse("attendants(list = [Tom, Wolf, Jessie]).", " (assert (private fact (name attendants) (pair (name list) (operator =) (list (type Tom) (type Wolf) (type Jessie)))))");
		
		// Test header | tail notation
		testParse("attendants(list = [Stijn | ?registred]) :- findall(?name, [ registred(person = ?name) ], ?registred).", "(rule (private fact (name attendants) (pair (name list) (operator =) (list (type Stijn) (tail (variable registred))))) (condition (findall (variable name) (private fact (name registred) (pair (name person) (operator =) (variable name))) (variable registred))))");
		
		// Test literal | tail notation
		testParse("attendants(list = [Tom, Wolf, Jessie | ?registred]) :- findall(?name, [ registred(person = ?name) ], ?registred).", "(rule (private fact (name attendants) (pair (name list) (operator =) (list (type Tom) (type Wolf) (type Jessie) (tail (variable registred))))) (condition (findall (variable name) (private fact (name registred) (pair (name person) (operator =) (variable name))) (variable registred))))");
		
		// Test cons-style list notation
		testParse("attendants(list = [Tom | [Jessie | [Wolf]]]).", "(assert (private fact (name attendants) (pair (name list) (operator =) (list (type Tom) (tail (list (type Jessie) (tail (list (type Wolf)))))))))");
		
		// Test erroneous header | tail | tail notation
		testParseError("attendants(list = [Tom | Jessie | Wolf]).", "expecting a right bracket, found '|'");
		
		// Test empty object parsing
		testParse("scheduled(slot = presentation).", "(assert (private fact (name scheduled) (pair (name slot) (operator =) (type presentation))))");
		
		// Test single-layered object parsing
		testParse("scheduled(slot = presentation(title = \"An introduction to CRIME\", presenter = Stijn)).", "(assert (private fact (name scheduled) (pair (name slot) (operator =) (object (type presentation) (pair (name title) (operator =) (text \"An introduction to CRIME\")) (pair (name presenter) (operator =) (type Stijn))))))");
		
		// Test erroneous double pair notation
		testParseError("scheduled(slot = presentation(title = \"An introduction to CRIME\", presenter = name = Stijn)).", "unexpected token: name");
		
		// Test nested object expression
		testParse("scheduled(slot = presentation(title = \"An introduction to CRIME\", presenter = person(name = Stijn, affiliation = VUB), attendants = [Tom, Wolf, Jessie])).", "(assert (private fact (name scheduled) (pair (name slot) (operator =) (object (type presentation) (pair (name title) (operator =) (text \"An introduction to CRIME\")) (pair (name presenter) (operator =) (object (type person) (pair (name name) (operator =) (type Stijn)) (pair (name affiliation) (operator =) (type VUB)))) (pair (name attendants) (operator =) (list (type Tom) (type Wolf) (type Jessie)))))))");				
	}
	
	/**
	 * Tests the parsing of constraints over free variables in a precondition.
	 */
	public void testConstraintsParsing() throws Exception {
		// Test the erroneous use of a constraint outside of a precondition
		testParseError("?attendants > 10", "expecting EOF, found '?'");
		
		// Test the erroneous use of a constraint as an attribute
		testParseError("worthwhile(popular = ?attendants > 10).", "unexpected token: >");
		
		// Parse a constraint expression over an externally bound variable
		testParse("worthwhile(event = ?e) :- event(event = ?e, numberOfAttendants = ?x), ?x > 10.", "(rule (private fact (name worthwhile) (pair (name event) (operator =) (variable e))) (condition (private fact (name event) (pair (name event) (operator =) (variable e)) (pair (name numberOfAttendants) (operator =) (variable x))) (constraint (variable x) (operator >) (number 10))))");
		
		// Parse a constraint expression during the binding of a variable
		testParse("worthwhile(event = ?e) :- event(event = workshop(self = ?e, numberOfAttendants > 10)).", "(rule (private fact (name worthwhile) (pair (name event) (operator =) (variable e))) (condition (private fact (name event) (pair (name event) (operator =) (object (type workshop) (pair (name self) (operator =) (variable e)) (pair (name numberOfAttendants) (operator >) (number 10)))))))");
		
		// Parse a constraint expression over a field of a bound variable
		testParse("worthwhile(event = ?e) :- event(event = ?e), ?e.numberOfAttendants > 10.", "(rule (private fact (name worthwhile) (pair (name event) (operator =) (variable e))) (condition (private fact (name event) (pair (name event) (operator =) (variable e))) (constraint (variable e numberOfAttendants) (operator >) (number 10))))");
		
		// Parse an implicit constraint using a nested variable
		testParse("preferred(event = ?e2) :- event(event = workshop(self = ?e, topic = ?t)), event(event = workshop(self = ?e2, topic = ?t, numberOfAttendants > ?e.numberOfAttendants))", "(rule (private fact (name preferred) (pair (name event)(operator =) (variable e2))) (condition (private fact (name event) (pair (name event) (operator =) (object (type workshop) (pair (name self) (operator =) (variable e)) (pair (name topic) (operator =) (variable t))))) (private fact (name event) (pair (name event) (operator =) (object (type workshop) (pair (name self) (operator =) (variable e2)) (pair (name topic) (operator =) (variable t)) (pair (name numberOfAttendants) (operator >) (variable e numberOfAttendants)))))))");
	}
	
	/**
	 * Tests the parsing of events of all allowed forms
	 */
	public void testEventParsing() throws Exception {
		// Test the parsing of an ordinary event
		testParseError(":Print(text = \"Hello World\").", "unexpected token: .");
		testParse(":Print(text = \"Hello World\") :- true().", "(rule (event (name Print) (pair (name text) (operator =) (text \"Hello World\"))) (condition (private fact (name true))))");
		
		// Test the parsing of a quantified event
		testParseError(":edu.vub.crime.test.Print().", "unexpected token: .");
		testParse(":edu.vub.crime.test.Print(text = \"Hello World\") :- true().", "(rule (event (name edu vub crime test Print) (pair (name text) (operator =) (text \"Hello World\"))) (condition (private fact (name true))))");
		
		// Test the erroneous use of an event in a precondition
		testParseError("error() :- :SimpleEvent(test).", "unexpected token: :");
		testParseError("error() :- :edu.vub.crime.test.SimpleEvent(test).", "unexpected token: :");
		
		// Test the erroneous use of a event in a qualified (public)fact
		testParseError("cannotTrigger -> :remoteEvent.", "expecting a symbol, found ':'");
	}
	
	public void testTypeDefinition() throws Exception {
		testParse("type ColorPrinter extends Printer", "(extends (type ColorPrinter) (type Printer))");
		testParse("type Copier extends Printer, Scanner", "(extends (type Copier) (type Printer) (type Scanner))");
	}
	
	public void testActionAbbreviations() throws Exception {
		testParse("action :start is :edu.vub.crime.example.iTunes.StartPlayer", "(action (event start) (event edu vub crime example iTunes StartPlayer))");
	}
}