package edu.vub.crime.parser;

import java.io.ByteArrayInputStream;

import org.apache.regexp.RE;

import antlr.CommonAST;
import edu.vub.crime.CrimeEngine;
import edu.vub.crime.CrimeTestCase;
import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.attributes.constant.Type;

public class WalkerTest extends CrimeTestCase {

	public static CrimeEngine engine_ = null;
	
	private static void testWalker(String input, String expectedOutput) throws Exception {
        LexerImpl lexer = new LexerImpl(new ByteArrayInputStream(input.getBytes()));
        ParserImpl parser = new ParserImpl(lexer);
        // Parse the input expression
        parser.program();
        CommonAST t = (CommonAST)parser.getAST();
        // Print the resulting tree out in LISP notation
        System.out.println(t.toStringList());
        TreeWalkerImpl walker = new TreeWalkerImpl();
        ExpressionList program = walker.program(t);
        
        assertEquals(new RE("\\s").subst(expectedOutput, ""), new RE("\\s").subst(program.toString(), ""));
	}
	
	public void testFactParsing() throws Exception {
		testWalker("alice().", "[alice(meta=meta(label = alice))]");

		// Single-arg unqualified fact
		testWalker("female(name = `alice).", "[female(name = `alice, meta=meta(label = female))]");
		
		// Multi-arg unqualified fact
		testWalker("parent(parent = `alice, child = `bob).", "[parent(child = `bob, parent = `alice, meta = meta(label = parent))]");
		
		// Multi-arg qualified fact
		testWalker("location -> isPresent(person = `alice, room = `living_room).", "[location -> isPresent(room = `living_room, person = `alice, meta = meta(label = isPresent, publisher = `_THIS_, space = location))]");
		
		// Event
		// testParse(":deploy_application(MediaHub).", "(assert (event (name  deploy_application) parameters (symbol MediaHub)))");
		
		// Action list consisting of events and facts
		// testParse(":switchProfile(`loud), profile(`loud).", "(assert (event (name switchProfile) parameters (symbol loud)) (private fact (name profile) (symbol loud )))");
		
		// A simple rule
		testWalker("mother(mother = ?X, child = ?Y) :- female(person = ?X), parent(parent = ?X, child = ?Y).", 
				   "[[mother(child=?Y,meta=meta(label=mother),mother=?X)] :- [female(person = ?X, meta = meta(label = female)), parent(child = ?Y, parent = ?X, meta = meta(label = parent))]]");
		testWalker("public -> mother(mother = ?X, child = ?Y) :- female(person = ?X), parent(parent = ?X, child = ?Y).", 
				   "[[mother(child=?Y,meta=meta(label=mother,publisher=`_THIS_,space=public),mother=?X)] :- [female(person = ?X, meta = meta(label = female)), parent(child = ?Y, parent = ?X, meta = meta(label = parent))]]");
	
	}
	
	public void testTypeDefinitions() throws Exception {
		testWalker("type Seminar extends Course", "[]");
		
		Type	theCourseType  = ExpressionFactory.type("Course");
		Type 	theSeminarType = ExpressionFactory.type("Seminar");
		Type[]	theSeminarParentTypes = theSeminarType.getParentTypes();
		
		assertEquals(1, theSeminarParentTypes.length);
		assertEquals(theCourseType, theSeminarParentTypes[0]);
		
		testWalker("type Copier extends Scanner, Printer", "[]");
		
		Type	theScannerType = ExpressionFactory.type("Scanner");
		Type	thePrinterType = ExpressionFactory.type("Printer");
		Type 	theCopierType  = ExpressionFactory.type("Copier");
		Type[]	theCopierParentTypes = theCopierType.getParentTypes();
		
		assertEquals(2, theCopierParentTypes.length);
		assertEquals(theScannerType, theCopierParentTypes[0]);
		assertEquals(thePrinterType, theCopierParentTypes[1]);
		
	}
	
	/**
	 * Tests the correct triggering of a rule with a single condition.
	 /
	public void testSingleConditionRule() {
		ExpressionList theProgram = CrimeParser.parse(
				"package edu.vub.crime.test \n" +
				":trigger(?name) :- colocated(?name).");
		
		testPrintedForm(theProgram, "[package edu.vub.crime.test, [edu.vub.crime.Assert(trigger, ?name)] :- [colocated(?name)]]");
	} */
	
	/*	public void testWalker() {		
			testWalker("%package edu.vub.crime :Assert(likes,?genre,?count) :- bagof(?person,(location(?person,living),prefers(?person,?genre).),?persons),length(?persons,?count).", "");
			
			testWalker("location(hans,living).", "");
			testWalker("location(bert,living).", "");
			testWalker("location(ernie,living).", "");
			
			testWalker("prefers(bert,house).", "");
			testWalker("prefers(hans,metal).", "");
			testWalker("prefers(ernie,metal).", "");
			engine_ = null;
			CrimeEngine.PACKAGE_DECL = null;
		} */
		
	/*	public void testNot() {
			testWalker("justsomewhere -> likes(hallo,?genre) :- cool -> prefers(?person,?genre),not iets -> hallo(x).", "");				
			testWalker("cool -> prefers(bert,house).", "");
			testWalker("cool -> prefers(hans,metal).", "");
			testWalker("hallo(x).", "");
			testWalker("iets -> hallo(x).","");
			engine_ = null;
		}*/
		
	/*	public void testBoard() {
			testWalker("test(genre, ?absolute, ?total) :- "+
			"category(?genre, ?absolute),"+
			"not total(0),"+
			"total(?total).", "");

			testWalker("total(2)." , "");	
			testWalker("category(Rock,asdf).","");	
			testWalker("category(Rock, absolute).","");
			engine_ = null;
		}*/
		
	/*	public void testFindAll() {		
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////FINDALL TEST//////////////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("inLiving(?count) :- findall(?person,(location(?person,living).),?persons),length(?persons,?count).", "");		
			testWalker("location(hans,living).", "");
			testWalker("location(bert,living).", "");
			testWalker("location(ernie,living).", "");
			engine_ = null;
		}*/
		
	/*	public void testSometimePast() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////SOMETIME-PAST TEST////////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("wereHere(?count) :- findall(?person,(sometimePast((location(hans,living).), (location(?person,living).)).),?persons),length(?persons,?count).", "");

	        engine_.beginActivation();
			testWalker("location(bert,living).", "");
	        engine_.beginActivation();
			testWalker("location(ernie,living).", "");
	        engine_.beginActivation();
			testWalker("location(hans,living).", "");
			engine_ = null;
		}*/
		
	/*	public void testSometimePast2() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////SOMETIME-PAST TEST2///////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("wereHere(?count) :- findall(?person,(location(bert,kitchen), sometimePast((location(hans,living).), (location(?person,living).)).),?persons),length(?persons,?count).", "");
			testWalker("location(bert,living).", "");
	        engine_.beginActivation();
			testWalker("location(ernie,living).", "");
	        engine_.beginActivation();
			testWalker("location(hans,living).", "");
			engine_.beginActivation();
			testWalker("location(bert,kitchen).", "");
			engine_ = null;
		}*/
		
	/*	public void testSometimePast3() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////SOMETIME-PAST TEST3///////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			//testWalker("test(?available, ?used, ?name, ?file) :- bagof(?p, (job(?name, ?file), sometimePast(printerUsed(?name, ?p)).), ?used), findall(?pr, (public -> printer(?pr).), ?available).", ""); 
			testWalker("test(?used, ?name, ?file) :- bagof(?p, (job(?name, ?file), sometimePast(printerUsed(?name, ?p)).), ?used).", ""); 
			testWalker("public -> printer(Lexmark).", "");
	        engine_.beginActivation();
	        testWalker("public -> printer(HP).", "");
	        engine_.beginActivation();
	        testWalker("public -> printer(Brother).", "");
	        engine_.beginActivation();
	        testWalker("printerUsed(Eline,Lexmark).", "");
	        engine_.beginActivation();
	        testWalker("printerUsed(Eline,Brother).", "");
	        engine_.beginActivation();
	        testWalker("printerUsed(Eline,Lexmark).", "");
	        engine_.beginActivation();
	        testWalker("job(Eline,pdfFile).", "");
	        engine_.beginActivation();
			engine_ = null;
		} */
		
	/*	public void testSometimePast3a() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////SOMETIME-PAST TEST3a//////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("test(?p, ?name, ?file) :- job(?name, ?file), sometimePast(printerUsed(?name, ?p).).", ""); 
			testWalker("public -> printer(Lexmark).", "");
	        engine_.beginActivation();
	        testWalker("public -> printer(HP).", "");
	        engine_.beginActivation();
	        testWalker("public -> printer(Brother).", "");
	        engine_.beginActivation();
	        testWalker("printerUsed(Eline,Lexmark).", "");
	        engine_.beginActivation();
	        testWalker("printerUsed(Eline,Brother).", "");
	        engine_.beginActivation();
	        testWalker("printerUsed(Eline,Lexmark).", "");
	        engine_.beginActivation();
	        testWalker("job(Eline,pdfFile).", "");
	        engine_.beginActivation();
	        testWalker("printerUsed(Eline,Brother).", "");
	        engine_.beginActivation();
	        testWalker("printerUsed(Eline,Brother).", "");
	        engine_.beginActivation();
	        testWalker("job(Eline,pdfFile2).", "");
	        engine_.beginActivation();
	        engine_ = null;
		}*/
		
	/*	public void testSometimePast4() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////SOMETIME PAST TEST4///////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("test(?p, ?s) :- login(?p), mostRecent(jukeboxOf(done),mostRecent(login(?p).), since(jukeboxOf(done),sometimePast(song(?p, ?s).).).).", "");	
			engine_.beginActivation();
			testWalker("login(Eline).", "");
	        engine_.beginActivation();
	        testWalker("song(Eline, song1).", "");
	        engine_.beginActivation();
	        Main.retract("login(Eline).");
	        testWalker("jukeboxOf(done).", "");
	        engine_.beginActivation();
	        Main.retract("jukeboxOf(done).");
	        testWalker("login(Eline).", "");
	        engine_.beginActivation();
	        testWalker("song(Eline, song2).", "");
	        engine_.beginActivation();
	        testWalker("song(Eline, song3).", "");
	        engine_.beginActivation();
	        Main.retract("login(Eline).");
	        testWalker("jukeboxOf(done).", "");
	        engine_.beginActivation();
	        Main.retract("jukeboxOf(done).");
	        testWalker("login(Eline).", "");
	        engine_.beginActivation();
			engine_.printMemory();
	        engine_ = null;
		}*/ 
		
	/*	public void testMostRecent() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////MOST RECENT TEST//////////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("test(?name, ?file, ?p) :- job(?name, ?file), mostRecent(printer(?name, ?p)).", ""); 
			testWalker("printer(Eline,Lexmark).", "");
	        engine_.beginActivation();
	        testWalker("printer(Eline, HP).", "");
	        engine_.beginActivation();
	        testWalker("printer(Eline,Lexmark).", "");
	        engine_.beginActivation();
	        testWalker("printer(Eline, HP).", "");
	        engine_.beginActivation();
	        testWalker("job(Eline, pdfFile).", "");
	        engine_.beginActivation();
	        testWalker("job(Eline, pdfFile).", "");
	        engine_.beginActivation();
	        engine_.printMemory();
	        engine_ = null;
		}  */
		
		
	/*	public void testMostRecent2() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////MOST RECENT TEST2/////////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("test(?p, ?s) :- login(?p), mostRecent(logout(?p),mostRecent(song(?p, ?s).).).", "");	
			engine_.beginActivation(null);
			testWalker("login(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("login(Christophe).", "");
	        engine_.beginActivation(null);
	        testWalker("song(Eline, song1).", "");
	        engine_.beginActivation(null);
	        testWalker("song(Christophe, song1).", "");
	        engine_.beginActivation(null);
	        Main.retract("login(Eline).");
	        testWalker("logout(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("login(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("song(Eline, song2).", "");
	        engine_.beginActivation(null);
	        testWalker("song(Eline, song3).", "");
	        engine_.beginActivation(null);
	        Main.retract("login(Christophe).");
	        testWalker("logout(Christophe).", "");
	        engine_.beginActivation(null);
	        testWalker("login(Christophe).", "");
	        engine_.beginActivation(null);
	        testWalker("song(Christophe, song2).", "");
	        engine_.beginActivation(null);
	        Main.retract("login(Eline).");
	        testWalker("logout(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("song(Christophe, song3).", "");
	        engine_.beginActivation(null);
	        testWalker("song(Christophe, song4).", "");
	        engine_.beginActivation(null);
	        testWalker("login(Eline).", "");
	        engine_.beginActivation(null);
	        Main.retract("login(Christophe).");
	        testWalker("logout(Christophe).", "");
	        engine_.beginActivation(null);
	        testWalker("login(Christophe).", "");
	        engine_.beginActivation(null);
			engine_.printMemory(System.out);
	        engine_ = null;
		} 
		
		public void testSince() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////SINCE TEST////////////////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("test(?p, ?s) :- login(?p), mostRecent(not login(?p),mostRecent(login(?p).), since(not login(?p),sometimePast(song(?s).).).).", "");	
			engine_.beginActivation();
			testWalker("login(Eline).", "");
	        engine_.beginActivation();
	        testWalker("song(song1).", "");
	        engine_.beginActivation();
	        Main.retract("login(Eline).");
	        engine_.beginActivation();
	        testWalker("login(Eline).", "");
	        engine_.beginActivation();
	        testWalker("song(song2).", "");
	        engine_.beginActivation();
	        Main.retract("login(Eline).");
	        engine_.beginActivation();
	        testWalker("login(Eline).", "");
			engine_.printMemory();
	        engine_ = null;
		} */
		
	/*	public void testSince2() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////SINCE TEST2///////////////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("test(?p, ?s) :- login(?p), mostRecent(logout(?p),mostRecent(login(?p).), since(logout(?p),sometimePast(song(?s).).).).", "");	
			engine_.beginActivation(null);
			testWalker("login(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("song(song1).", "");
	        engine_.beginActivation(null);
	        Main.retract("login(Eline).");
	        testWalker("logout(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("song(song99).", "");
	        engine_.beginActivation(null);
	        Main.retract("logout(Eline).");
	        testWalker("login(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("song(song2).", "");
	        engine_.beginActivation(null);
	        testWalker("song(song3).", "");
	        engine_.beginActivation(null);
	        Main.retract("login(Eline).");
	        testWalker("logout(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("song(song100).", "");
	        engine_.beginActivation(null);
	        Main.retract("logout(Eline).");
	        testWalker("login(Eline).", "");
			engine_.printMemory(System.out);
	        engine_ = null;
		} 
		
		
		public void testSince3() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////SINCE TEST3///////////////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("test(?p, ?s) :- login(?p), mostRecent(logout(?p),mostRecent(login(?p).), since(logout(?p),sometimePast(song(?s).).).).", "");	
			engine_.beginActivation(null);
			testWalker("login(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("song(song1).", "");
	        engine_.beginActivation(null);
	        testWalker("login(Christophe).", "");
	        engine_.beginActivation(null);
	        testWalker("song(songA).", "");
	        engine_.beginActivation(null);
	        Main.retract("login(Eline).");
	        testWalker("logout(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("song(song99).", "");
	        engine_.beginActivation(null);
	        testWalker("song(songB).", "");
	        engine_.beginActivation(null);
	        Main.retract("logout(Eline).");
	        testWalker("login(Eline).", "");
	        engine_.beginActivation(null);
	        Main.retract("login(Christophe).");
	        testWalker("logout(Christophe).", "");
	        engine_.beginActivation(null);
	        Main.retract("logout(Christophe).");
	        testWalker("login(Christophe).", "");
	        engine_.beginActivation(null);
	        testWalker("song(song2).", "");
	        engine_.beginActivation(null);
	        testWalker("song(song3).", "");
	        engine_.beginActivation(null);
	        testWalker("song(songC).", "");
	        engine_.beginActivation(null);
	        Main.retract("login(Eline).");
	        testWalker("logout(Eline).", "");
	        engine_.beginActivation(null);
	        testWalker("song(song100).", "");
	        engine_.beginActivation(null);
	        Main.retract("logout(Eline).");
	        testWalker("login(Eline).", "");
	        Main.retract("login(Christophe).");
	        testWalker("logout(Christophe).", "");
	        engine_.beginActivation(null);
	        Main.retract("logout(Christophe).");
	        testWalker("login(Christophe).", "");
	        engine_.beginActivation(null);
			engine_.printMemory(System.out);
	        engine_ = null;
		} */
		
		
	/*	public void testBetween() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////BETWEEN TEST//////////////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("hallo(?name, ?song) :- since((login(?name).), (jukeboxOf(?x).), (song(?song).)).", "");
			testWalker("login(Eline).", "");
	        engine_.beginActivation();
	        testWalker("song(song1).", "");
	        engine_.beginActivation();
	        testWalker("song(song2).", "");
	        engine_.beginActivation();
	        testWalker("song(song3).", "");
	        engine_.beginActivation();
	        testWalker("jukeboxOf(done).", "");
	        engine_.beginActivation();
	        testWalker("login(Eline).", "");
	        engine_.beginActivation();
	        testWalker("song(song4).", "");
	        engine_.beginActivation();
	        testWalker("song(song5).", "");
	        engine_.beginActivation();
	        testWalker("song(song6).", "");
	        engine_.beginActivation();
	        testWalker("jukeboxOf(done).", "");
	        engine_.beginActivation();
			engine_ = null;
		} */
		
	/*	public void testSince2() {
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////BETWEEN TEST2/////////////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("hallo(?name, ?song) :- login(?name), since((mostRecent login(?name).), (sometimePast((contra login(?name).), (mostRecent login(?name).)).), (song(?song).)).", "");
			testWalker("login(Eline).", "");
	        engine_.beginActivation();
	        testWalker("song(song1).", "");
	        engine_.beginActivation();
	        testWalker("song(song2).", "");
	        engine_.beginActivation();
	        testWalker("song(song3).", "");
	        engine_.beginActivation();
	        Main.retract("login(Eline).");
	        engine_.beginActivation();
	        testWalker("song(song99).", "");
	        engine_.beginActivation();
	        testWalker("login(Eline).", "");
	        engine_.beginActivation();
	        testWalker("song(song4).", "");
	        engine_.beginActivation();
	        testWalker("song(song5).", "");
	        engine_.beginActivation();
	        testWalker("song(song6).", "");
	        engine_.beginActivation();
	        Main.retract("login(Eline).");
	        engine_.beginActivation();
	        testWalker("song(song100).", "");
	        engine_.beginActivation();
	        testWalker("login(Eline).", "");
	        engine_.beginActivation();
	        engine_.printMemory();
			engine_ = null;
		}
		*/
		
	/*	public void testBagof() {		
			System.out.println("///////////////////////////////////////////////////////");
			System.out.println("/////////BAGOF TEST////////////////////////////////////");
			System.out.println("///////////////////////////////////////////////////////");
			testWalker("UpdateRating(?genre, ?absolute, ?total) :-"+
				    "category(?genre, ?absolute),"+
				    "total(?total),"+
				  "not total(0)." ,"");
			
			testWalker("total(?quantity) :-"+
			"findall(?person, ("+
				"location(?person, LivingRoom).),"+
				"?persons),"+
			"length(?persons, ?quantity).","");
				
			testWalker("category(?genre, ?quantity) :-"+
				"bagof(?person, ("+
			"location(?person, LivingRoom),"+ 
			"prefers(?person, ?genre).), "+
			"?persons),"+
			"length(?persons, ?quantity).","");
			
			testWalker("prefers(76539,Rock)." , "");
			testWalker("location(76539, LivingRoom).","");
		
			StringReader reader = new StringReader("location(76539, LivingRoom).");
			engine_.remove(((Fact)Main.parse(reader).getExpression(0)));
			
			engine_ = null;
		}*/
		
	/*	public void testNotify() {
			testWalker("nolify(?name, ?room, ?number) :-"+
					"location(?id, ?room),"+
					"person(?id, ?name, ?type),"+
					"phone(?room, ?number),"+
					"nolify(?name, ?room).","");


			testWalker("room(LivingRoom).","");
			testWalker("room(Kitchen).","");

			testWalker("public -> person(75773, Alice, student).","");
			testWalker("public -> person(76539, Bob, student).","");

			//testWalker("public -> application("Select_Favorites", "http://wilma.vub.ac.be/~cfscholl/Select_Favorites.zip").", "");
			//testWalker("public -> application("Messenger", "http://wilma.vub.ac.be/~cfscholl/Messenger.zip").,"");

			testWalker("public -> phone(LivingRoom, INTERN_0042).","");
			testWalker("public -> phone(Kitchen, INTERN_0043).","");
			testWalker("public -> location(76539, LivingRoom).","");
			
			engine_ = null;
		}*/	
}
