package edu.vub.crime.rete;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import edu.vub.crime.CrimeEngine;
import edu.vub.crime.dataflow.DataFlowAttribute;
import edu.vub.crime.grammar.Constraint;
import edu.vub.crime.grammar.Event;
import edu.vub.crime.grammar.Expression;
import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.ExpressionList;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.GrammarVisitor;
import edu.vub.crime.grammar.Negation;
import edu.vub.crime.grammar.Property;
import edu.vub.crime.grammar.Rule;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.AttributeVisitor;
import edu.vub.crime.grammar.attributes.CompoundAttribute;
import edu.vub.crime.grammar.attributes.ConstantAttribute;
import edu.vub.crime.grammar.attributes.Variable;
import edu.vub.crime.grammar.attributes.constant.Fraction;
import edu.vub.crime.grammar.attributes.constant.Number;
import edu.vub.crime.grammar.attributes.datastructures.List;
import edu.vub.crime.grammar.attributes.datastructures.Operator;
import edu.vub.crime.grammar.attributes.datastructures.Pair;
import edu.vub.crime.grammar.attributes.datastructures.TypedObject;
import edu.vub.crime.grammar.operators.BagOf;
import edu.vub.crime.grammar.operators.FindAll;
import edu.vub.crime.grammar.operators.Length;
import edu.vub.crime.grammar.utils.AttributePicker;
import edu.vub.crime.rete.combiner.Combiner;
import edu.vub.crime.rete.combiner.JoinCombiner;
import edu.vub.crime.rete.combiner.NegationCombiner;
import edu.vub.crime.rete.filter.FilterAttribute;
import edu.vub.crime.rete.filter.FilterFactory;
import edu.vub.crime.rete.filter.Join;
import edu.vub.crime.rete.filter.UnresolvedConstraint;
import edu.vub.crime.rete.nodes.BagofNode;
import edu.vub.crime.rete.nodes.FilterNode;
import edu.vub.crime.rete.nodes.FindallNode;
import edu.vub.crime.rete.nodes.JoinNode;
import edu.vub.crime.rete.nodes.LengthNode;
import edu.vub.crime.rete.nodes.NegationFilterNode;
import edu.vub.crime.rete.nodes.Node;
import edu.vub.crime.rete.nodes.NodesFactory;
import edu.vub.crime.rete.nodes.ProductionNode;
import edu.vub.crime.rete.nodes.RootNode;

public class CompileNetworkVisitor extends GrammarVisitor {

	/**
	 * When compiling the prerequisites of a rule, the first step is to build a filter node
	 * for each Fact. The {@link FilterNode} has two goals: it checks whether the provided
	 * token has appropriate values for the different attributes. This implies having the 
	 * correct values for all constants appearing in the pattern as well as having identical
	 * values for variables which occur more than once in a pattern.
	 * <p>
	 * Due to the introduction of data structures in the CRIME language, the construction of
	 * a filter node is slightly more complicated that iterating over all attributes, since
	 * they can now appear in a nested form. For instance, given a prerequisite of the form:
	 * <pre>
	 * fact(?var, [?var])
	 * </pre>
	 * The filter node is required to verify whether the second argument is a list with the
	 * first argument as its sole element. Therefore, we provide the compilation of the filter
	 * node through an {@link AttributeVisitor}.
	 * 
	 * @author smostinc
	 *
	 */
	public final class CompileFilterNode extends AttributeVisitor {

		private final FilterNode 		itsFilterNode;
		
		private final Map				itsSetOfEncounteredVariables;
		private final LinkedList		itsListOfConstrainedVariables;
		
		private CompileFilterNode(GenericFact theFact) {
			super("compile a filter for", "");
			
			itsFilterNode = NodesFactory.filterNode();
			itsSetOfEncounteredVariables = new HashMap();
			itsListOfConstrainedVariables = new LinkedList();
		}
		
		private void compileConstantFilter(Attribute theAttribute) {
			itsFilterNode.addNewFilter(FilterFactory.filterAttribute(theAttribute));
		}
		
		private void compileFlowFilter(DataFlowAttribute theDataFlowAttribute, Operator filterComparison) {
			itsFilterNode.addNewFilter(FilterFactory.filterFlow(theDataFlowAttribute, filterComparison, itsFilterNode, itsEngine));
		}
		
		private void compileDuplicateVariableFilter(Variable firstOccurrence, Variable newOccurrence) {
			itsFilterNode.addNewJoin(
					new Join(
							firstOccurrence.getIndex(), 
							newOccurrence.getIndex(),
							Operator.EQL));
		}
		
		// For each of the constrained variables, determine whether there is an unconstrained occurrence
		// (these are listed in the set itsSetOfEncounteredVariables). If there are, compare the constrained
		// variant with the unconstrained variant using the correct constraint.
		void compileConstrainedDuplicateVariableFilters() {
			for (Iterator constrainedVarIt = itsListOfConstrainedVariables.iterator(); constrainedVarIt.hasNext();) {
				
				Variable currentConstrainedVar = 
					(Variable)constrainedVarIt.next();
				Variable currentUnconstrainedVar = 
					(Variable)itsSetOfEncounteredVariables.get(currentConstrainedVar.getName());
				
				if(currentUnconstrainedVar != null) {
					itsFilterNode.addNewJoin(
							new Join(
									currentConstrainedVar.getIndex(), 
									currentUnconstrainedVar.getIndex(),
									currentConstrainedVar.getConstraint()));
					
					// The constraint is checked internally, hence it can be disregarded by future join nodes
					constrainedVarIt.remove();
				}
			}
		}
		
		public void visit(TypedObject theTypedObject) {
			// TODO filter out the type
			// itsFilterNode.addNewFilter(FilterFactory.filterType(theTypedObject));
			
			// Propagate over all fields
			for (Iterator attributesIt = theTypedObject.iterator(); attributesIt.hasNext();) {
				Pair currentField = (Pair) attributesIt.next();
				
				currentField.accept(this);
			}
		}
		
		public void visit(List theList) {
			// Propagate over all elements
			for (Iterator attributesIt = theList.iterator(); attributesIt.hasNext();) {
				Attribute currentAttribute = (Attribute) attributesIt.next();
				
				currentAttribute.accept(this);
			}
		}
		
		public void visit(Pair thePair) {
			final Operator theOperator = thePair.getOperator();
			
			// If the pair uses an '=' sign to relate the pair's name and value (the standard case)
			if(theOperator.equals(Operator.EQL)) {
				// Propagate to the pair's content
				thePair.getValue().accept(this);
			} else {
				Attribute	thePairValue = thePair.getValue();
				
				// Other relational operators are only supported for non-compound attributes
				if (thePairValue instanceof CompoundAttribute) {
					throw new RuntimeException("Cannot use relational operators in combination with compound attributes");
				} else {
					thePairValue.accept(new AttributeVisitor("compile", " into an attribute filter") {
						
						public void visit(Fraction theFraction) {
							itsFilterNode.addNewFilter(new FilterAttribute(theFraction, theOperator));
						}
						
						public void visit(Number theNumber) {
							itsFilterNode.addNewFilter(new FilterAttribute(theNumber, theOperator));
						}
						
						public void visit(Variable theVariable) {
							theVariable.setConstraint(theOperator);
							itsListOfConstrainedVariables.add(theVariable);
						}
						
						public void visit(DataFlowAttribute theFlow) {
							compileFlowFilter(theFlow, theOperator); 
						}
					});
				}
			}
		}
		
		public void visit(ConstantAttribute theConstant) {
			compileConstantFilter(theConstant);
		}
				
		public void visit(Variable theVariable) {
			Variable previouslyEncounteredInstance = 
				(Variable)itsSetOfEncounteredVariables.get(theVariable.getName());
			
			if(previouslyEncounteredInstance != null) {
				// If the same variable occurs multiple times, compile a filter
				compileDuplicateVariableFilter(previouslyEncounteredInstance, theVariable);
			} else {
				// Store that we have encountered this variable before
				itsSetOfEncounteredVariables.put(theVariable.getName(), theVariable);
			}
		}
		
		public void visit(DataFlowAttribute theFlow) {
			compileFlowFilter(theFlow, Operator.EQL); 
		}
	}
		
	private final CrimeEngine		itsEngine;
	private final RootNode			itsRootNode;
	private final ExpressionList	itsListOfCompiledExpressions;

	private Node					itsPreviouslyCompiledNode;
	
	private java.util.List			itsMapOfConstraints = new LinkedList();
	
	public CompileNetworkVisitor(CrimeEngine theEngine, RootNode theRootNode) {
		super("compile", " to a node in the RETE network");
		
		itsEngine 	= theEngine;
		itsRootNode = theRootNode;
		
		itsListOfCompiledExpressions = new ExpressionList();
	}
	
	public Node getCompiledNode() {
		evaluateConstraints();
		
		return itsPreviouslyCompiledNode;
	}
	
	/**
	 * This auxiliary method is the entry point for compiling a fact into a filter node. The visitor is
	 * created implicitly and is allowed to visit the attributes of the fact, yielding a full-fledged
	 * filter node. Note that this visitor does not add any filters to filter out the correct kind of
	 * fact, nor does it filter out whether the fact originated from the correct space. These filters
	 * are added in the {@link CompileNetworkVisitor}.
	 * 
	 * @param theFact to be compiled
	 * @return a {@link FilterNode} corresponding to the fact's pattern
	 */
	private FilterNode compileToFilterNode(GenericFact theFact) {
		CompileFilterNode theCompilingVisitor = new CompileFilterNode(theFact);
		
		for (Iterator attributesIt = theFact.attributeIterator(); attributesIt.hasNext();) {
			Attribute currentAttribute = (Attribute) attributesIt.next();
			
			currentAttribute.accept(theCompilingVisitor);
		}
		
		theCompilingVisitor.compileConstrainedDuplicateVariableFilters();
		
		for (Iterator constraintIt = theCompilingVisitor.itsListOfConstrainedVariables.iterator(); constraintIt.hasNext();) {
			Variable currentConstraint = (Variable) constraintIt.next();
			
			UnresolvedConstraint theNewConstraint = new UnresolvedConstraint(Variable.generate(), currentConstraint.getConstraint(), currentConstraint);
			
			theNewConstraint.setFactIndex(itsListOfCompiledExpressions.size());
			theNewConstraint.getConstrainedVariable().setIndex(currentConstraint.getIndex());
			currentConstraint.setIndex(null);
			
			itsMapOfConstraints.add(theNewConstraint);
		}
		
		return theCompilingVisitor.itsFilterNode;
	}
	
	private FilterNode compileAlphaNetwork(GenericFact theCurrentlyCompilingFact) {
		// Nodes are first filtered based on type, try to find an existing child node of the root
		// which already filters for this type of facts, creating one if necessary.
		FilterNode theTypeNode  = itsRootNode.getChild(theCurrentlyCompilingFact.getLabel());
		
		// Create a new filter node which checks whether published facts have matching constants etc.
		FilterNode theNewFilter = compileToFilterNode(theCurrentlyCompilingFact);
		
		// Hook the new filter into the network
		theTypeNode.addChild(theNewFilter);
		
		if(theCurrentlyCompilingFact.isPublic())
	 		// Then add an additional constraint so that the filter node only accepts fact from the correct space
			theTypeNode.addNewFilter(FilterFactory.filterSpace(theCurrentlyCompilingFact.getSpace()));

		
		return theNewFilter;

	}
		
 	private NegationFilterNode compileAlphaNetwork(Negation theCurrentlyCompilingFact) {
		// Nodes are first filtered based on type, try to find an existing child node of the root
		// which already filters for this type of facts, creating one if necessary.
		FilterNode theFilterNode  = compileAlphaNetwork(theCurrentlyCompilingFact.getNegatedFact());
		
		NegationFilterNode theLeftmostNegationNode = NodesFactory.negationFilterNode();       
		theFilterNode.addChild(theLeftmostNegationNode);
		
		// Determine whether a dummy token is needed to trigger the negation in absence of the filtered fact.
		theLeftmostNegationNode.triggerIfAbsent();
		
		return theLeftmostNegationNode;
 	}
 	
	private void compileBetaJoins(Expression theCurrentlyCompilingFact, Node theCurrentlyCompilingBetaNode) {
		Set theNewVariables = theCurrentlyCompilingFact.variables();

		for (Iterator variableIt = theNewVariables.iterator(); variableIt.hasNext();) {
			Variable currentVariable = (Variable) variableIt.next();
			
			// For the compilation of the Join constraints, we need to be able to extract
			// the correct fact from the token, hence this index.
			int indexInTheLeftToken = 0;
			
			boolean alreadyEncountered = false;
			
			for(Iterator expressionIt = itsListOfCompiledExpressions.iterator(); expressionIt.hasNext() && ! alreadyEncountered; ) {			
				Expression currentExpression = (Expression) expressionIt.next();
				
				alreadyEncountered = currentExpression.compileJoinConstraint(
						/* Compile a Join Constraint for the current Variable: */ 
						currentVariable, 
						/* Between this fact in the left token */ 
						indexInTheLeftToken,
						/* And the first fact in the right token */ 
						0,
						/* Which is to be added to this joinNode */
						theCurrentlyCompilingBetaNode);		
				
				
				indexInTheLeftToken += currentExpression.tokenSize();
			}
		}
	}
 	
	private void compileBetaProperties(Expression theCurrentlyCompilingFact, Node theCurrentlyCompilingBetaNode) {
		if (theCurrentlyCompilingFact instanceof GenericFact) {
			GenericFact theGenericFact = (GenericFact) theCurrentlyCompilingFact;
			
			Vector theConstraints = theGenericFact.getUnresolvedConstraints();
			
			for (Iterator constraintIt = theConstraints.iterator(); constraintIt.hasNext();) {
				Property currentConstraint = (Property) constraintIt.next();
				
				Variable theConstrainedVariable = currentConstraint.getVariable();
				Set theCompiledVariables = itsListOfCompiledExpressions.variables();
				
				if(!theCompiledVariables.contains(theConstrainedVariable)) {
					throw new RuntimeException("Constraint on an unbound variable"); 
				} else {
					
					// the correct fact from the token, hence this index.
					int indexInTheLeftToken = 0;
					
					boolean alreadyEncountered = false;
					
					for(Iterator expressionIt = itsListOfCompiledExpressions.iterator(); expressionIt.hasNext() && ! alreadyEncountered; ) {			
						Expression currentExpression = (Expression) expressionIt.next();
						
						alreadyEncountered = currentExpression.compileJoinConstraint(
								/* Compile a Join Constraint for the current Variable: */ 
								currentConstraint, 
								/* Between this fact in the left token */ 
								indexInTheLeftToken,
								/* And the first fact in the right token */ 
								0,
								/* Which is to be added to this joinNode */
								theCurrentlyCompilingBetaNode);		
						
						
						indexInTheLeftToken += currentExpression.tokenSize();
					}
					
				}
			}
		}
	}
	
	private Node compileBetaNetwork(Expression theCurrentlyCompilingFact, Node theCurrentlyCompiledAlphaNode, Combiner theCombiner) {
		// create a JoinNode with a prescribed combiner which determines which tokens can be passed
		// i.e. those for which there does (not) exist a counterpart with compatible binding for all
		// shared variables.
		JoinNode theJoinNode = NodesFactory.joinNode(theCombiner);

		// compile the Join constraints (which compare the bindings for all shared variables) for the 
		// JoinNode using the currently compiling Fact (i.e. the filter for the JoinNode's right input) 
		// and the list of previously compiled expressions (representing the left input of the JoinNode). 
		// Both inputs will pass tokens to the JoinNode, which will use the constraints to determine 
		// whether the tokens can be combined.
		compileBetaJoins(theCurrentlyCompilingFact, theJoinNode);
		
		compileBetaProperties(theCurrentlyCompilingFact, theJoinNode);
		
		// Wire up the join node as follows:
		// +--------------------------+            +-----------------+
		// | Already Compiled Portion |            | New Filter Node |
		// +--------------------------+            +-----------------+
		//                         \                  /
		//                          \                /
		//                          +---------------+
		//                          |   Join Node   +
		//                          +---------------+
		itsPreviouslyCompiledNode
			.addChild(theJoinNode.itsLeftInput());
		theCurrentlyCompiledAlphaNode
			.addChild(theJoinNode.itsRightInput());
		
		// Update the internal state of the visitor now that another node was compiled
		itsListOfCompiledExpressions.add(theCurrentlyCompilingFact);
		itsPreviouslyCompiledNode = theJoinNode;
		
		return theJoinNode;
	}
	
	private Node compileNestedQuery_BetaNetwork(Expression theCurrentlyCompilingFact, Node theCurrentlyCompiledAlphaNode, Combiner theCombiner) {
		// create a JoinNode with a prescribed combiner which determines which tokens can be passed
		// i.e. those for which there does (not) exist a counterpart with compatible binding for all
		// shared variables.
		JoinNode theJoinNode = NodesFactory.joinNode(theCombiner);

		// compile the Join constraints (which compare the bindings for all shared variables) for the 
		// JoinNode using the currently compiling Fact (i.e. the filter for the JoinNode's right input) 
		// and the list of previously compiled expressions (representing the left input of the JoinNode). 
		// Both inputs will pass tokens to the JoinNode, which will use the constraints to determine 
		// whether the tokens can be combined.
		compileBetaJoins(theCurrentlyCompilingFact, theJoinNode);
		
		compileBetaProperties(theCurrentlyCompilingFact, theJoinNode);
		
		// Wire up the join node as follows:
		// +--------------------------+            +-----------------+
		// | Already Compiled Portion |            | New Filter Node |
		// +--------------------------+            +-----------------+
		//                         \                  /
		//                          \                /
		//                          +---------------+
		//                          |   Join Node   +
		//                          +---------------+
		itsPreviouslyCompiledNode
			.addChild(theJoinNode.itsLeftInput());
		theCurrentlyCompiledAlphaNode
			.addChild(theJoinNode.itsRightInput());
		
		// Update the internal state of the visitor now that another node was compiled
		itsPreviouslyCompiledNode = theJoinNode;
		
		return theJoinNode;
	}
	
	private ProductionNode compileProductionNode(Rule theRule) {
		for(Iterator eventItr = theRule.getConsequence().iterator(); eventItr.hasNext();){
			Event event = (Event) eventItr.next();
			event.initialize(itsEngine, itsListOfCompiledExpressions);
		}
		
		return NodesFactory.productionNode(theRule, itsEngine);
	}
	
	private FindallNode compileFindAllNode(FindAll theCurrentlyCompilingPrimitive, ExpressionList theCompiledQueryExpressions) {
		ExpressionList allCompiledExpressions = new ExpressionList();
		
		allCompiledExpressions.addAll(itsListOfCompiledExpressions);
		allCompiledExpressions.addAll(theCompiledQueryExpressions);
		
		GenericFact theAccumulationFact = ExpressionFactory.unqualifiedFact("findall");
		Variable theAccumulatedVariable = theCurrentlyCompilingPrimitive.getAccumulator();
		theAccumulationFact.addNamedAttribute("accumulated", theAccumulatedVariable);
		itsListOfCompiledExpressions.add(theAccumulationFact);
		
		FindallNode theResultNode = NodesFactory.findallNode(
				// create an attribute picker to identify the value of the variable which is being accumulated
				theCurrentlyCompilingPrimitive.getAccumulatedVariable().makeAttributePicker(allCompiledExpressions), 
				// the cutoff point : which part of the token was not produced by the findall's query and should 
				// therefore be retained by the FindallNode
				computeTokenSize() - 1);
		
		return theResultNode;
	}
	
	private int computeTokenSize() {
		int theResultingSize = 0;
		
		for (Iterator expressionIt = itsListOfCompiledExpressions.iterator(); expressionIt.hasNext();) {
			Expression currentExpression = (Expression) expressionIt.next();
			theResultingSize += currentExpression.tokenSize();
		}
		
		return theResultingSize;
	}
	
	private BagofNode compileBagOfNode(BagOf theCurrentlyCompilingPrimitive, ExpressionList theCompiledQueryExpressions) {

		// Initialize the BagOf node so that we can add Join constraints to it which will 
		// verify whether two tokens have the same binding for all free variables in the 
		// query. These Join constraints will be used internally to determine whether two
		// tokens belong to the same bag or not.
		BagofNode theResultNode = NodesFactory.bagofNode();
		
		// Determine the offset within the token of the part produced by the query.
		// This offset is used as a starting position to compare variables in two tokens.
		int theOffset = itsListOfCompiledExpressions.size();
					
		// Iterate over all query expressions
		for(Iterator queryExpressionIt = theCompiledQueryExpressions.iterator(); queryExpressionIt.hasNext(); ) {
			Expression currentQueryExpression = (Expression) queryExpressionIt.next();
			
			Set queryVariables = currentQueryExpression.variables();
			
			for (Iterator queryVariableIt = queryVariables.iterator(); queryVariableIt.hasNext();) {
				Variable currentQueryVariable = (Variable) queryVariableIt.next();
				
				// ignore the accumulated variable
				if(currentQueryVariable.equals(theCurrentlyCompilingPrimitive.getAccumulatedVariable()))
					continue;
				
				// determine the first index of this variable in a token
				Vector freeVariableIndex = currentQueryVariable.getIndex();
				
				// add a join constraint which compares the variables in two tokens
				theResultNode.addNewJoin(
						new Join(theOffset, theOffset, freeVariableIndex, freeVariableIndex, Operator.EQL));
			}
			
			// the query expression is now properly compiled
			itsListOfCompiledExpressions.add(currentQueryExpression);
			
			// update the offset for comparing variables produced by the next expression
			theOffset++;
			
		}
		
		// after having processed the complete query expression (and thus having processed all variables
		// in the query), we can create a variable picker to find a value for the accumulated variable.
		theResultNode.setPicker(
				theCurrentlyCompilingPrimitive.getAccumulatedVariable().makeAttributePicker(itsListOfCompiledExpressions));
		
		return theResultNode;
	}
	
	private void evaluateConstraints() {
		for (Iterator constraintIt = itsMapOfConstraints.iterator(); constraintIt.hasNext();) {
			UnresolvedConstraint currentConstraint = (UnresolvedConstraint) constraintIt.next();
			
			if(currentConstraint.isFullySpecified(itsListOfCompiledExpressions)) {
				constraintIt.remove();
				
				currentConstraint.compile(itsListOfCompiledExpressions, itsPreviouslyCompiledNode);
			}
		}
	}
		
	public void visit(GenericFact theNode) {
		// compile the fact to a filter node
		Node theCurrentlyCompiledAlphaNode = compileAlphaNetwork(theNode);
		
		if(itsPreviouslyCompiledNode != null) {
			// if this fact is not the first one to be compiled, create a join node 
			// connecting it with the rest of the network
			compileBetaNetwork(theNode, theCurrentlyCompiledAlphaNode, new JoinCombiner());
		} else {
			// else install it as the the initial network created by the compiler
			itsPreviouslyCompiledNode = theCurrentlyCompiledAlphaNode;

			itsListOfCompiledExpressions.add(theNode);
		}
	}

	public void visit(Negation theNode) {
		if(itsPreviouslyCompiledNode == null) {
			// If the negation appears in the left-most position, a special kind of
			// FilterNode needs to be created. This is done by the createAlphaNetwork
			// method which is specialized on Negation
			itsPreviouslyCompiledNode = compileAlphaNetwork(theNode);
		} else {
			// otherwise compile the negated fact into an ordinary filter node
			Node theCurrentlyCompiledAlphaNode = compileAlphaNetwork(theNode.getNegatedFact());
			
			// and construct a join node with a negation-aware combiner
			compileBetaNetwork(theNode.getNegatedFact(), theCurrentlyCompiledAlphaNode, new NegationCombiner());
		}
		
		itsListOfCompiledExpressions.add(theNode);
	}
	
	public void visit(ExpressionList theNode) {
		for(Iterator expressionIt = theNode.iterator(); expressionIt.hasNext(); ) {
			Expression currentExpression = (Expression) expressionIt.next();
			currentExpression.accept(this);
			
			// After each expression is processed, determine whether the constraints that
			// were imposed on the expression can now be evaluated.
			evaluateConstraints();
		}
	}
	
	public void visit(Rule theRule) {
		// Compile a RETE network from the prerequisites of the rule
		theRule.getPrerequisite().accept(this);
		
		// Construct a production node for the rule
		ProductionNode theProductionNode = compileProductionNode(theRule);
		
		// Add the production node as a child of the compiled network
		getCompiledNode().addChild(theProductionNode);
		
		// Register that the rule has been deployed
		itsRootNode.deployed(theRule, theProductionNode);
		
		// Reset the compiled network to allow compiling a list of rules properly
		itsListOfCompiledExpressions.clear();
		itsPreviouslyCompiledNode = null;
	}
	
	// CRIME accumulation extensions
	
	public void visit(FindAll theNode) {
		// Create a child compiler which will compile the query inside the FindAll node
		CompileNetworkVisitor theFindAllQueryCompiler = new CompileNetworkVisitor(itsEngine, itsRootNode);
		
		// Pass it to the query so that a network can be created without interfering with 
		// this compiler's internal state.
		theNode.getQuery().accept(theFindAllQueryCompiler);
		
		// Retrieve the result of the compilation (through class-based encapsulation)
		Node theCurrentlyCompiledAlphaNode = theFindAllQueryCompiler.itsPreviouslyCompiledNode;
		
		if(itsPreviouslyCompiledNode != null) {
			// The compiled query should be hooked to the network created by this compiler.
			compileNestedQuery_BetaNetwork(theNode.getQuery(), theCurrentlyCompiledAlphaNode, new JoinCombiner());			
		} else {
			itsPreviouslyCompiledNode = theCurrentlyCompiledAlphaNode;
		}
		
		// The FindAll rete node will need to consider both the variables which are coming 
		// from the previously compiled expressions, but also those from its query. The latter 
		// however are not visible beyond the scope of the FindAll node. 
		FindallNode theFindAllNode = compileFindAllNode(theNode, theFindAllQueryCompiler.itsListOfCompiledExpressions);
		
		itsPreviouslyCompiledNode.addChild(theFindAllNode);
		
		itsPreviouslyCompiledNode = theFindAllNode;
	}
	
	public void visit(BagOf theNode) {
		// Create a child compiler which will compile the query inside the FindAll node
		CompileNetworkVisitor theBagOfQueryCompiler = new CompileNetworkVisitor(itsEngine, itsRootNode);
		
		// Pass it to the query so that a network can be created without interfering with 
		// this compiler's internal state.
		theNode.getQuery().accept(theBagOfQueryCompiler);
		
		// Retrieve the result of the compilation (through class-based encapsulation)
		Node theCurrentlyCompiledAlphaNode = theBagOfQueryCompiler.itsPreviouslyCompiledNode;
		
		if(itsPreviouslyCompiledNode != null) {
			compileNestedQuery_BetaNetwork(theNode.getQuery(), theCurrentlyCompiledAlphaNode, new JoinCombiner());
		} else {
			itsPreviouslyCompiledNode = theCurrentlyCompiledAlphaNode;
		}

		BagofNode theBagOfNode = compileBagOfNode(theNode, theBagOfQueryCompiler.itsListOfCompiledExpressions);
		
		itsPreviouslyCompiledNode.addChild(theBagOfNode);
		itsListOfCompiledExpressions.add(theNode);
		
		itsPreviouslyCompiledNode = theBagOfNode;
	}
	
	public void visit(Length theNode) {
		AttributePicker theListVariablePicker =
			theNode.getList().makeAttributePicker(itsListOfCompiledExpressions);
		
		LengthNode theLengthNode = null;
		
		{	
			Attribute theLength =
				theNode.getLength();
			
			if(theLength instanceof Variable) {
				Variable theLengthVariable = (Variable) theLength;
				
				if(! itsListOfCompiledExpressions.variables().contains(theLength)) {
					// the length is free variable up to this point in the query
					theLengthNode = NodesFactory.lengthNode(false, theListVariablePicker, null);
				}
			}
			
			// In all other cases (it is a constant or a variable that is already bound)
			if(theLengthNode == null) {
				AttributePicker theLengthVariablePicker = 
					theNode.getLength().makeAttributePicker(itsListOfCompiledExpressions);
				
				theLengthNode = NodesFactory.lengthNode(true, theListVariablePicker, theLengthVariablePicker);
			}
		}
		
		itsPreviouslyCompiledNode.addChild(theLengthNode);
		itsListOfCompiledExpressions.add(theNode);
		
		itsPreviouslyCompiledNode = theLengthNode;
	}
	
	public void visit(Constraint theNode) {
		itsMapOfConstraints.add(new UnresolvedConstraint(theNode.getVariable(), theNode.getConstraint(), theNode.getExpectedValue()));
		
		evaluateConstraints();
	}
}
