package edu.vub.crime.rete.agenda;

import java.util.Iterator;

import edu.vub.crime.CrimeEngine;

/** 
 * The Agenda class is the abstract superclass which stores all activations which
 * still need to be processed by the {@link CrimeEngine}. An activation contains
 * actions which may be the consequence of triggering a local rule or which are 
 * broadcasted by a nearby engine.
 * <p>
 * Different variations of the agenda can be conceived to introduce different 
 * scheduling strategies such as first-in-first-out (CRIME's default strategy),
 * prioritized scheduling, or last-in-first-out.
 * <p>
 * The Agenda can also be used to filter out conflicting activation. For instance, 
 * when a new activation is scheduled to retract a fact, while the assertion of the
 * same fact is currently scheduled, both activations cancel each other out.
 * 
 * @author cfscholl
 * @author ephilips
 * @author smostinc
 */
public abstract class Agenda {
	
	protected boolean askedToClose = false; // to determine whether this agenda can issue further activations
	
	/**
	 * Checks whether a new activation conflicts with an activation which is currently
	 * in the schedule. A conflict arises when an activation was produced by a rule/device
	 * based on a token T and the schedule contains an activation which originates from 
	 * the same rule/device and was caused by the Token's complement c(T).
	 * 
	 * @param theActivation a new activation which should be scheduled.
	 */
	protected boolean checkForConflicts(Activation theActivation) {
		for(Iterator scheduleIt = iterator(); scheduleIt.hasNext(); ) {
    		Activation scheduledActivation = (Activation) scheduleIt.next();
    		
    		if(scheduledActivation.isComplementOf(theActivation)) {
    			scheduleIt.remove();
    			return true;
    		}
    	}
    	return false;
	}
	
	/**
	 * Returns an in-order iterator which traverses the agenda.
	 */
	protected abstract Iterator iterator();
    
	/**
	 * Returns and possibly awaits the first applicable activation and removes this from the agenda.
	 */
	public abstract Activation getNextActivation() throws InterruptedException;
	

	/**
	 * A non-blocking variant of getNextActivation which returns null if there is no activation at present
	 */
	public abstract Activation pollForActivation();
	
	/**
	 * Schedules a new activation in the agenda.
	 */
	public abstract void schedule(Activation theActivation);
	
	/**
	 * Reports the number of activations waiting to be executed
	 */
	public abstract int size();
	
	/**
	 * This method is used to determine when the scheduler can be safely stopped. This is not
	 * the case when the agenda is empty, yet when it has received a call to its close method.
	 * Therefore this method always returns true, even when the agenda is currently empty.
	 * @return false only if the agenda has been closed
	 */
    public boolean hasNext() {
    	return ! askedToClose;
    }

    /**
     * Requests this agenda to close, causing the scheduler to terminate normally.
     */
    public void close() {
    	askedToClose = true;
    }
}
