package edu.vub.crime.rete.combiner;

import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.rete.data.DataFactory;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.rete.filter.Join;

/**
 * The JoinCombiner class ensures that when a token is passed to a JoinNode (via either its left
 * or right input) that the tokens in the opposite memory are evaluated to determine which tokens
 * have matching bindings for the variables that both tokens share. For all of these tokens, a new
 * combined token is constructed which contains the newly received token, together with one token
 * from the JoinNode's opposite memory. These combined tokens can then be propagated further down
 * the network.
 * 
 * @author cfscholl
 * @author ephilips
 */
public class JoinCombiner extends Combiner {
	
	/** 
	 * Combines a token received from the JoinNode's right input with all tokens in the 
	 * node's left memory. The resulting Vector contains a combined token (containing the 
	 * content of the token from the memory followed by the content of the token from the 
	 * right input) for every matching token in the left memory.
	 */
	public Vector combineLeft(Token theTokenFromTheRightInput) {
		Vector theCombinedTokens = new Vector();
		Vector theMatchingTokens = checkLeft(theTokenFromTheRightInput);
		
		if(theMatchingTokens.size() > 0) {
			// First add all facts of the tokens of the left memory
			for(Iterator matchingTokenIt = theMatchingTokens.iterator(); matchingTokenIt.hasNext(); ) {
				Token currentMatchingLeftToken = (Token) matchingTokenIt.next();
				
				// Create a new token with the tag of the newly inserted token
				Token theCombinedToken 	= DataFactory.token(theTokenFromTheRightInput.getTag());
				
				// Copy all facts from the token in the left memory (which should come first)
				theCombinedToken 		= currentMatchingLeftToken.copyFacts(theCombinedToken);

				// Copy all facts from the newly inserted token
				theCombinedToken		= theTokenFromTheRightInput.copyFacts(theCombinedToken);
				
				theCombinedTokens.add(theCombinedToken);
			}
		}
		return theCombinedTokens;
	}

	/** 
	 * Combines a token received from the JoinNode's right input with all tokens in 
	 * the node's left memory. The resulting Vector contains a combined token (
	 * containing the content of the left token followed by the content of the token
	 * from the memory) for every token matching token in the right memory.
	 */
	public Vector combineRight(Token theTokenFromTheLeftInput) {
		Vector theCombinedTokens = new Vector();
		Vector theMatchingTokens = checkRight(theTokenFromTheLeftInput);
		
		if(theMatchingTokens.size() > 0) {
			// First add all facts of the tokens of the left memory
			for(Iterator matchingTokenIt = theMatchingTokens.iterator(); matchingTokenIt.hasNext(); ) {
				Token currentMatchingRightToken = (Token) matchingTokenIt.next();
				
				// Create a new token with the tag of the newly inserted token
				Token theCombinedToken 	= DataFactory.token(theTokenFromTheLeftInput.getTag());
				
				// Copy all facts from the newly inserted token (which should come first)
				theCombinedToken		= theTokenFromTheLeftInput.copyFacts(theCombinedToken);
				
				// Copy all facts from the token in the right memory 
				theCombinedToken 		= currentMatchingRightToken.copyFacts(theCombinedToken);

				theCombinedTokens.add(theCombinedToken);
			}
		}
		return theCombinedTokens;
	}	
	
	/** 
	 * Identifies all tokens in the left memory which match with the given token that
	 * was received from the node's right input. These tokens will be returned in a 
	 * Vector class, such that they can be properly combined and returned to the node
	 * which may propagate them.
	 */
	private Vector checkLeft(Token theTokenFromTheRightInput) {
		Vector theTokensToBeCombinedAndPropagated = new Vector();
		
        for(Iterator tokenIt = itsLeftMemory.iterator(); tokenIt.hasNext(); ) {
        	Token currentTokenFromMemory = (Token) tokenIt.next();
        	
        	// Assume that the token will pass all Join constraints
            boolean shouldBeCombinedAndPropagated = true;
            
            for(Iterator joinIt=itsConstraints.iterator(); joinIt.hasNext(); ) {
            	Join currentJoinConstraint = (Join) joinIt.next();
            	
            	// Until we can identify a failing constraint
                if(! currentJoinConstraint.join(currentTokenFromMemory, theTokenFromTheRightInput)) {
                    shouldBeCombinedAndPropagated = false;
                    break;
               }
            }
            
            // If no such failing constraint could be found
            if(shouldBeCombinedAndPropagated) {
            	// Add the token to the list of tokens to be propagated in combination 
            	// with the given token from the right input.
            	theTokensToBeCombinedAndPropagated.add(currentTokenFromMemory);
            }
        }
        
        return theTokensToBeCombinedAndPropagated;
	}
	
	/** 
	 * Identifies all tokens in the right memory which match with the given token that
	 * was received from the node's left input. These tokens will be returned in a 
	 * Vector class, such that they can be properly combined and returned to the node
	 * which may propagate them.
	 */
	private Vector checkRight(Token theTokenFromTheLeftInput) {
		Vector theTokensToBeCombinedAndPropagated = new Vector();
		
        for(Iterator tokenIt = itsRightMemory.iterator(); tokenIt.hasNext(); ) {
        	Token currentTokenFromMemory = (Token) tokenIt.next();
        	
        	// Assume that the token will pass all Join constraints
            boolean shouldBeCombinedAndPropagated = true;
            
            for(Iterator joinIt=itsConstraints.iterator(); joinIt.hasNext(); ) {
            	Join currentJoinConstraint = (Join) joinIt.next();
            	
            	// Until we can identify a failing constraint
            	// NOTE : in checkLeft the order of the arguments is different!!!
                if(! currentJoinConstraint.join(theTokenFromTheLeftInput, currentTokenFromMemory)) {
                    shouldBeCombinedAndPropagated = false;
                    break;
               }
            }
            
            // If no such failing constraint could be found
            if(shouldBeCombinedAndPropagated) {
            	// Add the token to the list of tokens to be propagated in combination 
            	// with the given token from the right input.
            	theTokensToBeCombinedAndPropagated.add(currentTokenFromMemory);
            }
        }
        
        return theTokensToBeCombinedAndPropagated;
	}
}
