package edu.vub.crime.rete.combiner;

import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.rete.data.DataFactory;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.rete.data.TokenTag;
import edu.vub.crime.rete.filter.Join;

/** <pre>
 * NegationCombiner is used for combining the tokens of the 
 * left- and right-memory of the JoinNode, when there's
 * a negation (right part of the two conditions).
 * The rule which is used for explanation is : 
 *
 *         differentTaste(?person, ?genre) :-
 *              isPresent(?person, LivingRoom),
 *              prefers(?person, ?genre),
 *              not prefers(Alice, ?genre).
 *    
 * This rule determines the persons who are in the living room and have a different
 * taste than Alice.
 * 
 * Suppose: Rete network given below (until last join node).
 * 
 *     FILTER: isPresent             FILTER: prefers            FILTER: prefers
 *             |                             |                          |
 *     FILTER: LivingRoom                   /                    FILTER: Alice
 *               \                         /                           |
 *                \                       /                            |
 *                 \                     /                             |   
 *                 JOIN: ?person = ?person                             |
 *                           |                                         |
 *                            \                                       /
 *                             \                                     /
 *                              \                                   /
 *                               \                                 /
 *                                       JOIN: ?genre = ?genre      
 *                                                |
 *                                                
 *
 * Suppose Alice and Bob both like Rock. So, the memory contains two facts: 
 * prefers(Alice, Rock) and prefers(Bob, Rock).
 *
 * When Bob is entering the living room, the token < add, isPresent(Bob, LivingRoom) > 
 * is inserted to the Rete network.
 * The updated network is depicted below.
 *       
 *     FILTER: isPresent             FILTER: prefers            FILTER: prefers
 *             |                             |                          |
 *     FILTER: LivingRoom                   /                    FILTER: Alice
 *               \                         /                           |
 *                \                       /                            |
 *                 \                     /                             |   
 *    LEFT MEMORY   JOIN: ?person = ?person     RIGHT MEMORY           |
 *    ===========            |                  ============           |
 *    Bob, LR.               |                  Alice, Rock            |
 *                           |                  Bob, Rock              |
 *                            \                                       /
 *                             \                                     /
 *                              \                                   /
 *                               \                                 /
 *                   LEFT MEMORY         JOIN: ?genre = ?genre     RIGHT MEMORY
 *                   ===========                  |                ============
 *                   Bob, LR. ; Bob, Rock         |                Alice, Rock
 *                                                
 * This last join node (?genre = ?genre) has a NegationCombiner for combining the newly inserted token,
 * < add, isPresent(Bob, LivingRoom), prefers(Bob, Rock) > with the tokens of the right memory.
 * 
 * There will only be sent a token to the children of this join node when there's no match
 * of the token with the tokens of the right memory.
 * So, in this case no token is sent.
 *  
 * A negation combiner works by checking the matches of the left memory with those of the right memory.
 * When there are no matches, the token from the left is passed to the children.
 * Because it's possible for a token form the left to have multiple matches with tokens in the right memory,
 * every token keeps a number indicating how many matches it has.
 * When a token from the left is inserted, we check the right memory and increase this number for every match.
 * 
 * insert => isPresent(Bob,LivingRoom)
 * insert => prefers(Bob,Rock)
 * 
 *         LEFT MEMORY         	    	  	RIGHT MEMORY
 *         ===========          	  		============
 *         Bob,LR; Bob Rock 0   No Matches
 *         				
 *               Pass Token : < add ; Bob, LR ; Bob, Rock >              
 * 
 * 
 * When a token from the right is inserted we check it with the left memory and increase all matches from the left memory by one.
 * Removing a token from the right decreases every match with the left.
 * 
 * insert => prefers(Alice,Rock).
 * 
 *         LEFT MEMORY         	    	  	RIGHT MEMORY
 *         ===========          	  		============
 *         Bob LivingRoom 1   One Match	     Alice, Rock
 *         
 *           Pass Token : < remove ; Bob, LR ; Bob, Rock >     
 *           
 * REMARK: Note that indeed there will be a remove and an add token passed to the children but because these two tokens are 
 * two counterpoles they will be removed by the agenda.
 * </pre> */
public class NegationCombiner extends Combiner {
	
	/** <pre>
	 * Combine methods implement the behaviour in which the memories are combined.
	 * There are two methods implemented, namely combineLeft and combineRight.
	 * This is necessary, because there's a difference in the way the new tokens
	 * (that will be passed to the children of the join node) are formed.
	 * It needs to be ensured that the tokens of the left memory are occuring
	 * in front of that newly formed token, followed by the tokens coming from the
	 * right memory.
	 *
	 * This method is called when a new token is inserted in the right memory
	 * of the join node.
	 * First, all the tokens which matched are calculated (method check is called).
	 * This method, check, will search for every token in the left memory that
	 * has the same variable bindings as the one which is just newly inserted in the
	 * right memory (= variable token which is passed as an argument to combineLeft).
	 * </pre> */
	public Vector combineLeft(Token token) {
		return checkLeft(token); 
	}

	/** <pre>
	 * This method is called when a new token is inserted in the left memory
	 * of the join node.
	 * It implements the same behaviour as the one above (combineLeft),
	 * only the check method is different.
	 * </pre> */
	public Vector combineRight(Token token) {
		return  checkRight(token); 
	}
		
	/** <pre>
	 * Check method is implemented to find all tokens of a certain memory 
	 * (which is passed as an argument) which match with a given token.
	 * A token of the memory matches with the given one, whenever their
	 * variable bindings are the same. 
	 * </pre> */
	private Vector checkLeft(Token token) {
		Vector tokens = new Vector();
        for(Iterator tkn=itsLeftMemory.iterator(); tkn.hasNext(); ) {
        	Token memToken = (Token) tkn.next();
            boolean pass = true;        
            for(Iterator ftr=itsConstraints.iterator(); ftr.hasNext(); ) {
                Join filter = (Join) ftr.next();
                if(! filter.join(memToken, token)) {
                    pass = false;
                    break;
               }
            }
            if(pass) {
            	switch(token.getTag()) {
                case TokenTag.add:
                    if(memToken.matches() == 0) {
                        //making new token because we don't want to change the original token
                        Token removeToken = DataFactory.token(TokenTag.remove);
                        memToken.copyFacts(removeToken);
                        tokens.add(removeToken);
                    }
                    memToken.increment();
                    break;
                case TokenTag.remove:
                    memToken.decrement();
                    if(memToken.matches() == 0) {
                        tokens.add(memToken);
                    }
                    break;
                default: 
            	}
            }
        }
        return tokens;
	}
	
	/** <pre>
	 * Check method is implemented to find all tokens of a certain memory 
	 * (which is passed as an argument) which match with a given token.
	 * A token of the memory matches with the given one, whenever their
	 * variable bindings are the same.
	 * </pre> */
	private Vector checkRight(Token token) {
        Vector tokens = new Vector();
        // Calculate the number of matches for the token (inserted in the left memory),
        // with those of the right memory.
        for(Iterator tkn=itsRightMemory.iterator(); tkn.hasNext(); ) {
        	Token memToken = (Token) tkn.next();
            boolean pass = true;
            for(Iterator ftr=itsConstraints.iterator(); ftr.hasNext(); ) {
                Join filter = (Join) ftr.next();
                if(! filter.join(token, memToken)) {
                    pass = false;
                    break;
               }
            }
            if(pass) {
            	switch(token.getTag()) {
                case TokenTag.add:
                    token.increment();
                    break;
                case TokenTag.remove:
                    token.decrement();
                    break;
                default:
            	}
            }
        }    
        // If there are no matches found, the token can be passed to the children
        // of the JoinNode (where this is the combiner of).
        if(token.matches() == 0) {
           	tokens.add(token);
        }       
        return tokens;
	}
}
