package edu.vub.crime.rete.data;

import java.io.Serializable;
import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.rete.combiner.NegationCombiner;
import edu.vub.crime.rete.nodes.FindallNode;

/** 
 * Tokens are passed between the different nodes in the RETE network. When given a token,
 * any node will be able to compare whether the token fits either their filter, or whether
 * it can be joined safely with other tokens in a compatible memory table. When a token is
 * passed by a node, that node will also store it in its memory.
 * <p>
 * Since facts can be both asserted and removed, and the RETE network reuses the same 
 * mechanism to deal with both operations, tokens are tagged with a token.
 * 
 * @author cfscholl
 * @author ephilips
 * @author smostinc
 */
public class Token implements Serializable {
	
	/**
	 * This tag can take on the different values defined in the TokenTag class, its signals
	 * whether the token deals with an assertion or a retraction.
	 */
	private final int itsTag;

	/**
	 * The facts propagated through this token. This vector contains one element initially 
	 * with more facts being added as matching facts are found in Join nodes in the network.
	 */
	private final Vector itsFacts;
	
	/**
	 * The amount of times this token has produced a match. This counter is used in the 
	 * {@link NegationCombiner} to ensure that a token may only be propagated if it did
	 * not match any elements of the memory.
	 */
	private int itsMatchCount;
	
	/** 
	 * Initializes a token with an empty list of facts.
	 */
	public Token(int tag) {
		itsFacts = new Vector();
		itsTag = tag;
		itsMatchCount = 0;
	}
	
	/** 
	 * Compares two tokens and verifies whether the tokens have the same length 
	 * and whether the facts at the corresponding positions are equal. Note that 
	 * the tag of the token is not taken into account during this comparison.
	 */
	public boolean compare(Token token) {
		if(itsFacts.size() != token.size()) {
			return false;
		}
		Iterator hisFactIt = token.itsFacts.iterator();
	    Iterator myFactIt  = itsFacts.iterator();
	    
	    while(hisFactIt.hasNext()) {
	        GenericFact hisFact = (GenericFact) hisFactIt.next();
	        GenericFact myFact  = (GenericFact) myFactIt.next();
	        
	        if(!myFact.equals(hisFact))
	            return false;
	    }
	    
	    return true;
	}
	
	/** 
	 * This method compares two objects and determines whether they are equal.
	 * To do so it uses {@link Token#compare(Token)}.
	 */
	public boolean equals(Object theOtherObject) {
		if(theOtherObject == null) return false;
		
		if(theOtherObject instanceof Token) {
			Token theOtherToken = (Token)theOtherObject;
			
			return this.compare(theOtherToken);
		}
		
		return false;
	}

	/**
	 * @return a tag indicating whether this token represents an assertion or a retraction.
	 */
	public int getTag() {
		return itsTag;
	}

	/** 
	 * @return the fact stored in this token at the given index
	 */
	public GenericFact getFact(int index) {		
		return (GenericFact) itsFacts.get(index);
	}
	
	/**
	 * Add a fact to this token.
	 */
	public void addFact(GenericFact fact){
		itsFacts.add(fact);
	}
	
	/**
	 * Removes the final element of the token. Used when stripping the token that is internally
	 * produced by a findall construct of all facts that relate to concrete matches and not to
	 * the collective.
	 * 
	 * @see FindallNode#stripToken(Token, int)
	 */
	public void removeLast() {
		itsFacts.removeElementAt(itsFacts.size() - 1);
	}

	/** 
	 * Adds all facts contained in this token to those of the specified token.
	 */
	public Token copyFacts(Token theToken) {
		// NOTE: setting the facts of the token to a clone of our vector is not safe.
		// if the token already has facts in it these would be lost by this implementation.
		
		for(Iterator factIt = itsFacts.iterator(); factIt.hasNext(); ) {
			GenericFact currentFact = (GenericFact) factIt.next();
			theToken.addFact(currentFact);
		}
		return theToken;
	}
	
	/** 
	 * @return the number of facts encapsulated by this token.
	 */
	public int size() {
		return itsFacts.size();
	}
	
	/** 
	 * @return the number of matches there have been between this token and the memory.
	 */
	public int matches() {
		return itsMatchCount;
	}
	
	/** 
	 * Indicates that this token has produced a match.
	 */
	public void increment() {
		itsMatchCount++;
	}

    /** 
     * Indicates that a matching token was removed from the memory.
     */
    public void decrement() {
    	itsMatchCount--;
    }
	
    /**
     * Determines whether two tokens are one another's complement.
     */
	public boolean isComplementOf(Token possibleComplement) {
		if(TokenTag.areComplementary(itsTag, possibleComplement.itsTag)) {
			return this.compare(possibleComplement);
		} else {
			return false;
		}
	}
	
	public String toString() {
		String prefix = "";
		switch (itsTag) {
			case TokenTag.add:
				prefix = "+";
				break;
			case TokenTag.remove:
				prefix = "-";
				break;
		}
		
		return prefix + itsFacts;
	}
	
	public int hashCode() {
		return itsFacts.toString().hashCode();
	}
}