package edu.vub.crime.rete.filter;

import edu.vub.crime.rete.data.Token;

/** 
 * Filters are used by the RETE network to enforce constraints over the tokens that are 
 * passed in filter nodes and join nodes respectively. The former are presented with a 
 * single token which they may or may not pass on, whereas the latter are used to join
 * tokens from the left and right-hand input. Before passing on a token to the lower
 * nodes in the network, these filters verify whether the attributes of all facts in a
 * token compare correctly to the said constants.
 * 
 */
public abstract class Filter {

	/** Identifies the fact that this filter will consider. By default, this id is 0. */ 
	protected final int itsFactId;
	
	/**
	 * Filters are used primarily in the alpha network, in so-called filter nodes. These
	 * nodes are used to filter out all newly asserted facts which do not have a proper 
	 * name, origin or which have incorrect attributes. When filtering out newly asserted
	 * facts, the tokens only 1 fact, which is stored at position 0. Therefore, the default
	 * fact index is 0.
	 */
	public Filter() { 
		this(0); 
	}
	
	/**
	 * This constructor allows explicitly passing a fact id, which allows filtering out facts
	 * in tokens which contain more than a single fact. This is used in the compilation of
	 * for instance constraint expressions.
	 * 
	 * @param theFactId identifies the position of the fact under scrutiny in the token.
	 */
	public Filter(int theFactId) {
		itsFactId = theFactId;
	}
	
    /** 
     * Can this token be passed on to lower networks in the node?
     */
    public abstract boolean pass(Token token);
}