package edu.vub.crime.rete.filter;

import java.util.Vector;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.datastructures.Operator;
import edu.vub.crime.rete.data.Token;

/** 
 * Joins are used by the RETE network to enforce that bindings for a variable in one token
 * are equal to the bindings for the same variable in another token. Joins are used in both
 * filter nodes and join nodes respectively. The former are presented with a single token 
 * and will use joins to determine whether variables that occur twice in a single fact have
 * the same binding. The latter take two separate tokens from the left and right-hand input
 * and verify whether the bindings for the variables are compatible.
 */
public class FilterVariableConstraint extends Filter {

	/** Identifies the fact in the left token that this filter will consider. By default, this id is 0. */ 
	protected final int itsLeftFactId;

	/** Identifies the fact in the right token that this filter will consider. By default, this id is 0. */ 
	protected final int itsRightFactId;
	
	/** Identifies the position in the left Fact where the variable its value can be found */
	protected final Vector itsLeftAttributeIndex;
	
	/** Identifies the position in the right Fact where the variable its value can be found */
	protected final Vector itsRightAttributeIndex;

	/** Details the relation between both selected attributes. The default value is EQ */
	protected final Operator itsComparisonOperator;
	
	public FilterVariableConstraint(Vector theLeftAttributeIndex, Vector theRightAttributeIndex, Operator theComparisonOperator) {
		this(0, 0, theLeftAttributeIndex, theRightAttributeIndex, theComparisonOperator);
	}
	
	public FilterVariableConstraint(int theLeftFactId, int theRightFactId, 
			Vector theLeftAttributeIndex, Vector theRightAttributeIndex, 
			Operator theComparisonOperator) {
		itsLeftFactId 			= theLeftFactId;
		itsRightFactId			= theRightFactId;
		
		itsLeftAttributeIndex 	= theLeftAttributeIndex;
		itsRightAttributeIndex 	= theRightAttributeIndex;
		
		itsComparisonOperator 	= theComparisonOperator;
	}
	
	/**
	 * Verifies whether the value for a variable is compatible in the left and right tokens. If so, both
	 * tokens can be joined and passed on to lower nodes in the network.
	 */
	public boolean pass(Token theToken) {
		try {
			GenericFact theLeftFact 	= theToken.getFact(itsLeftFactId);
			GenericFact theRightFact 	= theToken.getFact(itsRightFactId);
			
			Attribute theLeftAttribute	= theLeftFact.getAttribute(itsLeftAttributeIndex.iterator());
			Attribute theRightAttribute = theRightFact.getAttribute(itsRightAttributeIndex.iterator());
			
			return itsComparisonOperator.evaluate(theLeftAttribute, theRightAttribute);
		} catch (IncorrectAttributeIndexException e) {
			return false;
		}
	}

}
