package edu.vub.crime.rete.nodes;

import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.datastructures.List;
import edu.vub.crime.grammar.utils.AttributePicker;
import edu.vub.crime.rete.data.DataFactory;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.rete.data.TokenTag;

public class BagofNode extends Node {	
	
	/** <pre>
	 * InnerClass keeps track of the tokens that are passed through the BagofNode.
	 * These tokens are kept together with other tokens that have the same values for there free variables.
	 * The first token that is passed will be used as a reference to match other tokens with the same free variables. 
	 * For example:
	 *  
	 * someFact(A,1).
	 * someFact(B,1).
	 * 
	 * bagOf(?x,someFact(?x,?y),?xs)
	 * 
	 * Then there will be one < someFact(1,1) > with attributes [A,B].
	 * </pre> */
	class TokenValues {	
		Vector attributes_;
		
		/**
		 * Reference token which is the first token that passed through the BagOf node with a particular
		 * combination of free variables. This token is used as a template to see whether a new token
		 * has the same bindings for its free variables.
		 */
		Token token_;		
		
		/** <pre>
		 * Constructs it with the reference token and the first attribute.
		 * </pre> */
		TokenValues(Token token, Attribute attr) {
			Token newToken = DataFactory.addToken();
			token.copyFacts(newToken);
			this.token_ = newToken;
			this.attributes_  = new Vector();
			this.attributes_.add(attr);
		}
		
		/** <pre>
		 * Builds a fact with the given attributes.
		 * </pre> */
		private GenericFact accumulate() {
			List acc = ExpressionFactory.list();
			for(Iterator it=attributes_.iterator(); it.hasNext(); ) {
				Attribute attr = (Attribute) it.next();
				acc.add(attr);
			}
			GenericFact fact = ExpressionFactory.unqualifiedFact("bagof");
			fact.addNamedAttribute("accumulated", acc);
			return fact;
		}
		
		/** <pre>
		 * Makes a addToken with the current attributes.
		 * </pre> */
		public Token addToken() {
			Token addToken = DataFactory.addToken();
			
			// Add the reference fact to the token since it contains the bindings for the free variables 
			token_.copyFacts(addToken);
			
			// Add an unqualified "acc" fact which contains the list of accumulated values as its single attribute
			addToken.addFact(accumulate());
			
			return addToken;
		}
		
		/** <pre>
		 * Makes a removeToken with the current attributes.
		 * </pre> */
		public Token removeToken() {
			Token addToken = DataFactory.removeToken();

			// Add the reference fact to the token since it contains the bindings for the free variables 
			token_.copyFacts(addToken);
			
			// Add an unqualified "acc" fact which contains the list of accumulated values as its single attribute
			addToken.addFact(accumulate());
			return addToken;
		}
		
		public boolean hasAttributes() {
			return !this.attributes_.isEmpty();
		}
		
		/** <pre>
		 * Remove an attribute.
		 * </pre> */ 
		public void removeAttr(Attribute attr) {
			this.attributes_.remove(attr);
		}
		
		/** <pre>
		 * Add an attribute.
		 * </pre> */
		public void addAttr(Attribute attr) {
			this.attributes_.add(attr);
		}
		
		/** <pre>
		 * Checks if a token has the same free variables as the reference token.
		 * </pre> */
		public boolean match(Token token) {	
			return join(this.token_, token);
		}
	}
	
	private Vector previous_;
	private AttributePicker picker_;
	
	/** <pre>
	 * Constructor.
	 * </pre> */
	public BagofNode() {		
		previous_ = new Vector();
	}
	
	/** <pre>
	 * Adds a new token to previous_.
	 * </pre> */
	public Token makeNew(Token t) {		
		try {
			if(t.getTag() == TokenTag.add) {
				Attribute attr = picker_.pickAttribute(t);
				TokenValues tokenValues = new TokenValues(t, attr); 
				previous_.add(tokenValues);
				return tokenValues.addToken();
			}
			else
				throw new RuntimeException("Making new with remove tag !!");
		} catch (IncorrectAttributeIndexException e) {
			throw new RuntimeException(e.getMessage(),e);
		}	
	}
	
	/** <pre> 
	 * Inserts a token. 
	 * First update the previous_ memory 
	 * then send it to the children.
	 * </pre> */	
	public void insert(Token newToken) {
		Token t = DataFactory.token(newToken.getTag());
		newToken.copyFacts(t);
		Vector tokens = update(t);
		if(tokens.size() == 0) {
			tokens.add(makeNew(t));			
		}				
		send(tokens);
	}
	
	/** <pre>
	 * Send given tokens to the children of this node.
	 * </pre> */
	public void send(Vector tokens) {
		for(Iterator tkn=tokens.iterator(); tkn.hasNext(); ) {
			Token t = (Token) tkn.next();
			
			record(t);
			
			for(Iterator chd=itsChildNodes.iterator(); chd.hasNext(); ) {
				Node node = (Node) chd.next();
				node.insert(t);
			}
        }
	}
	
	/** <pre>
	 * Update the current memory.
	 * Iterates over the previous_ and if a match is found update it.
	 * </pre> */
	public Vector update(Token t) {
		try {
			Vector tokens = new Vector();
			Attribute attr = picker_.pickAttribute(t);
			for(Iterator itr=previous_.iterator();itr.hasNext();){
				TokenValues tokenValues = (TokenValues) itr.next();
				if(tokenValues.match(t)) {
					switch(t.getTag()) {
						case TokenTag.add:
							tokens.add(tokenValues.removeToken());
							tokenValues.addAttr(attr);
							tokens.add(tokenValues.addToken());
							return tokens;
						case TokenTag.remove:
							tokens.add(tokenValues.removeToken());
							tokenValues.removeAttr(attr);
							// Do not pass empty list token !!
							if (tokenValues.hasAttributes()) {
								tokens.add(tokenValues.addToken());
							}
							else {
								itr.remove();
							}
							return tokens;
						default:	
							throw new RuntimeException("Wrong tagged Token");					
					}
				}
			}						
			return tokens;
		} catch (IncorrectAttributeIndexException e) {
			throw new RuntimeException(e.getMessage(),e);
		}		
	}
	
	public void setPicker(AttributePicker picker) {
		picker_ = picker;
	}
}