package edu.vub.crime.rete.nodes;


import java.util.Vector;

import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.constant.Type;
import edu.vub.crime.grammar.attributes.datastructures.List;
import edu.vub.crime.grammar.attributes.datastructures.Operator;
import edu.vub.crime.grammar.attributes.datastructures.Pair;
import edu.vub.crime.grammar.attributes.datastructures.TypedObject;
import edu.vub.crime.parser.IndexingVisitor;
import edu.vub.crime.rete.data.DataFactory;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.rete.data.TokenTag;
import edu.vub.crime.rete.filter.FilterAttribute;
import edu.vub.crime.rete.nodes.aux.TestNode;

public class FilterNodeTest extends NodeTest {

	/**
	 * Tests the use of a basic filter node which tests the kind of fact that is being passed through as well
	 * as for the presence of constants on specific attribute positions.
	 */
	public void testFlatConstantFilter() {
		// Make a filter for course(?title, ProfileSoftwareEngineering)
		GenericFact theFilterPrescription;
		
		{ 
			theFilterPrescription = ExpressionFactory.unqualifiedFact("course");
		
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theFilterPrescription);
			
			// Attribute 1 is a ?title variable
			Attribute theTitle = ExpressionFactory.variable("title");
			theFilterPrescriptionIndexer.nextAttribute("title", theTitle);
			
			// Attribute 2 is a ProfileSoftwareEngineering symbol
			Attribute theConstant = ExpressionFactory.type("ProfileSoftwareEngineering");
			theFilterPrescriptionIndexer.nextAttribute("profile" ,theConstant);
		}
		
		compileFilterTestNetwork(theFilterPrescription);
		
		// ensure the filter passes on tokens if necessary
		testFilterPasses("course(title = \"Interpretation of Computer Programs 2\", profile = ProfileSoftwareEngineering).");

		// ensure the created filter rejects facts with an incorrect header
		testFilterRejects("seminar(title = \"Techniques of Aspect-Oriented Software Design\", profile = ProfileSoftwareEngineering).");
		
		// ensure the created filter rejects facts with incorrect values for the constrained attributes 
		testFilterRejects("course(title = \"Machine Learning\", profile = ProfileArtificialIntelligence).");	
	}
	
	/**
	 * Tests the construction of a filter node which filters out duplicate variables in a flat fact.
	 */
	public void testFlatDuplicateVariableFilter() {
		// Make a filter for course(?title, ?tutor, ?tutor)
		// i.e. a course where the theory and exercises are taught by the same person.
		GenericFact theFilterPrescription;
		
		{ 
			theFilterPrescription = ExpressionFactory.unqualifiedFact("course");
		
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theFilterPrescription);
			
			// Attribute 1 is a ?title variable
			Attribute theTitle = ExpressionFactory.variable("title");
			theFilterPrescriptionIndexer.nextAttribute("title", theTitle);
			
			// Attribute 2 is a ?tutor variable
			Attribute theProfessor = ExpressionFactory.variable("tutor");
			theFilterPrescriptionIndexer.nextAttribute("tutor", theProfessor);

			// Attribute 3 is a ?tutor variable
			Attribute theTeachingAssistant = ExpressionFactory.variable("tutor");
			theFilterPrescriptionIndexer.nextAttribute("assistant", theTeachingAssistant);
		}
		
		compileFilterTestNetwork(theFilterPrescription);
		
		// ensure the filter passes on tokens if necessary
		testFilterPasses("course(title = \"Concepts of Programming Languages\", tutor = JohanF, assistant = JohanF).");

		// ensure the created filter rejects facts with different values for the variable
		testFilterRejects("course(title = \"Algorithms and Datastructures 1\", tutor = Wolf, assistant = Charlotte).");
	}

	/**
	 * Tests the construction of a filter node for a fact which contains a list attribute. 
	 */
	public void testListFilter() {
		// Make a filter for course(?title, [?title, ProfileSoftwareEngineering, ?tutor, ?tutor])
		// This fact tests the comparison of variables, inside and outside of the list, 
		// the correct values for values of the list, and duplicate variables within the list.
		GenericFact theFilterPrescription;
		
		{ 
			theFilterPrescription = ExpressionFactory.unqualifiedFact("course");
		
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theFilterPrescription);
			
			// Attribute 1 is a ?title variable
			Attribute theTitle = ExpressionFactory.variable("title");
			theFilterPrescriptionIndexer.nextAttribute("title", theTitle);
			
			// Attribute 2 is a list
			List theCourseData = ExpressionFactory.list();
			{	
				// Attribute 2, 1 is a ?title variable
				Attribute theCourseTitle = ExpressionFactory.variable("title");
				theCourseData.add(theCourseTitle);

				// Attribute 2, 2 is a symbol constant
				Attribute theProfile = ExpressionFactory.type("ProfileSoftwareEngineering");
				theCourseData.add(theProfile);
				
				// Attribute 2, 3 is a ?tutor variable
				Attribute theProfessor = ExpressionFactory.variable("tutor");
				theCourseData.add(theProfessor);

				// Attribute 2, 4 is a ?tutor variable
				Attribute theTeachingAssistant = ExpressionFactory.variable("tutor");
				theCourseData.add(theTeachingAssistant);
			}
			theFilterPrescriptionIndexer.nextAttribute("data", theCourseData);
			
		}
		
		compileFilterTestNetwork(theFilterPrescription);
		
		// ensure the filter passes on tokens if necessary
		testFilterPasses("course(title = \"Concepts of Programming Languages\", data = [\"Concepts of Programming Languages\", ProfileSoftwareEngineering, JohanF, JohanF]).");

		// ensure the created filter rejects facts with different values for the tutor variable
		testFilterRejects("course(title = \"Interpretation of Computer Programs 2\", data = [\"Interpretation of Computer Programs 2\", ProfileSoftwareEngineering, Wolf, Andy]).");

		// ensure the created filter rejects facts with different values for the title variable
		testFilterRejects("course(title = \"CPL\", data = [\"Concepts of Programming Languages\", ProfileSoftwareEngineering, JohanF, JohanF]).");
		
		// ensure the created filter rejects facts with an incorrect values for the profile constant
		testFilterRejects("course(title = \"Software Project 1\", data = [\"Software Project 1\", BasicCourse, NN, NN]).");
	}
	
	/**
	 * Tests the construction of a filter node for a fact which contains an object attribute. 
	 */
	public void testObjectFilter() {
		// Make a filter for course(?title, {title = ?title; profile = ProfileSoftwareEngineering; professor = ?tutor; teachingAssistant = ?tutor})
		// This fact tests the comparison of variables, inside and outside of a CRIME object, 
		// the correct values for values of the object, and duplicate variables within the object.
		GenericFact theFilterPrescription;
		
		{ 
			theFilterPrescription = ExpressionFactory.unqualifiedFact("course");
		
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theFilterPrescription);
			
			// Attribute 1 is a ?title variable
			Attribute theTitle = ExpressionFactory.variable("title");
			theFilterPrescriptionIndexer.nextAttribute("title", theTitle);
			
			// Attribute 2 is a typed object
			TypedObject theCourseData = ExpressionFactory.object(ExpressionFactory.type("Course"));
			{	
				// Attribute 2, 1 is a title = ?title pair
				Attribute theCourseTitle = ExpressionFactory.variable("title");
				Pair theCourseTitlePair = ExpressionFactory.pair("title", Operator.EQL, theCourseTitle);
				theCourseData.addField(theCourseTitlePair);

				// Attribute 2, 2 is a profile = symbol pair
				Attribute theProfile = ExpressionFactory.type("ProfileSoftwareEngineering");
				Pair theProfilePair = ExpressionFactory.pair("profile", Operator.EQL, theProfile);
				theCourseData.addField(theProfilePair);
				
				// Attribute 2, 3 is a professor = ?tutor pair
				Attribute theProfessor = ExpressionFactory.variable("tutor");
				Pair theProfessorPair = ExpressionFactory.pair("professor", Operator.EQL, theProfessor);
				theCourseData.addField(theProfessorPair);

				// Attribute 2, 4 is a teachingAssistant = ?tutor pair
				Attribute theTeachingAssistant = ExpressionFactory.variable("tutor");
				Pair theTeachingAssistantPair = ExpressionFactory.pair("teachingAssistant", Operator.EQL, theTeachingAssistant);
				theCourseData.addField(theTeachingAssistantPair);
			}
			theFilterPrescriptionIndexer.nextAttribute("data", theCourseData);
			
		}
		
		compileFilterTestNetwork(theFilterPrescription);
		
		// ensure the filter passes on tokens if necessary
		testFilterPasses("course(title =\"Concepts of Programming Languages\", data = Course( title = \"Concepts of Programming Languages\", profile = ProfileSoftwareEngineering, professor = JohanF, teachingAssistant = JohanF)).");

		// object fields may be in arbitrary positions, the name is used as an index
		testFilterPasses("course(title =\"Compilers\", data = Course( profile = ProfileSoftwareEngineering, title = \"Compilers\", teachingAssistant = Vermeir, professor = Vermeir)).");

		// ensure the created filter rejects facts with different values for the tutor variable
		testFilterRejects("course(title =\"Interpretation of Computer Programs 2\", data = Course( title = \"Interpretation of Computer Programs 2\", profile = ProfileSoftwareEngineering, professor = Wolf, teachingAssistant = Andy)).");

		// ensure the created filter rejects facts with different values for the title variable
		testFilterRejects("course(title =\"CPL\", data = Course( title = \"Concepts of Programming Languages\", profile = ProfileSoftwareEngineering, professor = JohanF, teachingAssistant = JohanF)).");
		
		// ensure the created filter rejects facts with an incorrect values for the profile constant
		testFilterRejects("course(title =\"Software Project 1\", data = Course( title = \"Software Project 1\", profile = BasicCourse, professor = NN, teachingAssistant = NN)).");
	}
	
	/**
	 * Tests the construction of a filter node for a fact which contains an object attribute in which some pairs contain . 
	 */
	public void testConstraintsFilter() {
		// Make a filter for course(?title, {title = ?title; credits > 5; professor = ?tutor; teachingAssistant != ?tutor})
		// This fact tests the comparison of variables, inside and outside of a CRIME object, 
		// the correct values for values of the object, and duplicate variables within the object.
		GenericFact theFilterPrescription;
		
		{ 
			theFilterPrescription = ExpressionFactory.unqualifiedFact("course");
		
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theFilterPrescription);
			
			// Attribute 1 is a ?title variable
			Attribute theTitle = ExpressionFactory.variable("title");
			theFilterPrescriptionIndexer.nextAttribute("title", theTitle);
			
			// Attribute 2 is a list
			TypedObject theCourseData = ExpressionFactory.object(ExpressionFactory.type("Course"));
			{	
				// Attribute 2, 1 is a title = ?title pair
				Attribute theCourseTitle = ExpressionFactory.variable("title");
				Pair theCourseTitlePair = ExpressionFactory.pair("title", Operator.EQL, theCourseTitle);
				theCourseData.addField(theCourseTitlePair);

				// Attribute 2, 2 is a credits > integer pair
				Attribute theCredits = ExpressionFactory.attribute(5);
				Pair theCreditPair = ExpressionFactory.pair("credits", Operator.GTH, theCredits);
				theCourseData.addField(theCreditPair);
				
				// Attribute 2, 3 is a professor = ?tutor pair
				Attribute theProfessor = ExpressionFactory.variable("tutor");
				Pair theProfessorPair = ExpressionFactory.pair("professor", Operator.EQL, theProfessor);
				theCourseData.addField(theProfessorPair);

				// Attribute 2, 4 is a teachingAssistant != ?tutor pair
				Attribute theTeachingAssistant = ExpressionFactory.variable("tutor");
				Pair theTeachingAssistantPair = ExpressionFactory.pair("teachingAssistant", Operator.NEQ, theTeachingAssistant);
				theCourseData.addField(theTeachingAssistantPair);
			}
			theFilterPrescriptionIndexer.nextAttribute("data", theCourseData);
			
		}
		
		compileFilterTestNetwork(theFilterPrescription);
		
		// ensure the filter passes on tokens if necessary
		testFilterPasses("course(title = \"Interpretation of Computer Programs 2\", data = Course( title = \"Interpretation of Computer Programs 2\", credits = 6, professor = Wolf, teachingAssistant = Andy)).");

		// ensure it rejects courses for which the tutor condition does not hold
		testFilterRejects("course(title = \"Compilers\", data = Course( title = \"Compilers\", credits = 8, professor = Vermeir, teachingAssistant = Vermeir)).");

		// ensure the filter rejects courses with a low weight
		testFilterRejects("course(title = \"CPL\", data = Course( title = \"Concepts of Programming Languages\", credits = 4, professor = Isabel, teachingAssistant = Coen)).");
	}
	
	/**
	 * Tests the constructed filter node when using a negated fact as the leftmost expression in a rule
	 */
	public void testNegationFilter() {
		// Construct a filter for "not aFact." To do so we first need to construct a fact which has
		// no variables (to avoid floundering since at the beginning of the program the leftmost
		// negated node automatically triggers).
		GenericFact theFilterPrescription;
		
		{ 
			theFilterPrescription = ExpressionFactory.unqualifiedFact("course");
		
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theFilterPrescription);
			
			// Attribute 1 is a text attribute
			Attribute theTitle = ExpressionFactory.attribute("Compilers");
			theFilterPrescriptionIndexer.nextAttribute("title", theTitle);
			
			// Attribute 2 is an object
			TypedObject theCourseData = ExpressionFactory.object(ExpressionFactory.type("Course"));
			{	
				// Attribute 2, 1 is a title = text pair
				Attribute theCourseTitle = ExpressionFactory.attribute("Compilers");
				Pair theCourseTitlePair = ExpressionFactory.pair("title", Operator.EQL, theCourseTitle);
				theCourseData.addField(theCourseTitlePair);

				// Attribute 2, 2 is a profile = symbol pair
				Attribute theProfile = ExpressionFactory.symbol("ProfileSoftwareEngineering");
				Pair theProfilePair = ExpressionFactory.pair("profile", Operator.EQL, theProfile);
				theCourseData.addField(theProfilePair);
				
				// Attribute 2, 3 is a professor = symbol pair
				Attribute theProfessor = ExpressionFactory.symbol("Vermeir");
				Pair theProfessorPair = ExpressionFactory.pair("professor", Operator.EQL, theProfessor);
				theCourseData.addField(theProfessorPair);

				// Attribute 2, 4 is a teachingAssistant = symbol pair
				Attribute theTeachingAssistant = ExpressionFactory.symbol("Vermeir");
				Pair theTeachingAssistantPair = ExpressionFactory.pair("teachingAssistant", Operator.EQL, theTeachingAssistant);
				theCourseData.addField(theTeachingAssistantPair);
			}
			theFilterPrescriptionIndexer.nextAttribute("data", theCourseData);
			
		}
		
		FilterNode theCompiledFilter = compileFilterTestNetwork(theFilterPrescription);
		
		// Splice in a NegationFilterNode after the compiled filter
		NegationFilterNode theNegatedFilter = NodesFactory.negationFilterNode();
		
		itsTestNode = new TestNode();
		theCompiledFilter.addChild(theNegatedFilter);
		theNegatedFilter.addChild(itsTestNode);
		
		// Insert a dummy token, which will trigger the negated filter into passing on a token 
		// (this is normally done using the NegatingFilterBootstrap agenda activation)
		theNegatedFilter.insert(DataFactory.removeToken());
		itsTestNode.assertTriggered(TokenTag.add);
		itsTestNode.reset();
		
		// Insert a token matching the prescription and ensure that the negated filter retracts 
		// the previously transmitted token
		Token assertionToken = DataFactory.addToken(theFilterPrescription);
		itsRootNode.insert(assertionToken);
		
		itsTestNode.assertTriggered(TokenTag.remove);
		itsTestNode.reset();
		
		// Insert a token to retract a fact matching the prescription and ensure that the negated
		// filter, passes on a token.
		Token retractionToken = DataFactory.removeToken(theFilterPrescription);
		itsRootNode.insert(retractionToken);
		
		itsTestNode.assertTriggered(TokenTag.add);
		itsTestNode.reset();
	}
	
	/**
	 * Tests the correct behavior if a filter node is forced to reevaluate its constraint.
	 */
	public void testReevaluation() {
		// Make a filter for course(?title, ?profile)
		GenericFact theFilterPrescription;
		
		{ 
			theFilterPrescription = ExpressionFactory.unqualifiedFact("course");
		
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theFilterPrescription);
			
			// Attribute 1 is a ?title variable
			Attribute theTitle = ExpressionFactory.variable("title");
			theFilterPrescriptionIndexer.nextAttribute("title", theTitle);
			
			// Attribute 2 is a ?profile variable
			Attribute theProfile = ExpressionFactory.variable("profile");
			theFilterPrescriptionIndexer.nextAttribute("profile", theProfile);
		}
		
		FilterNode theCompiledFilter = compileFilterTestNetwork(theFilterPrescription);

		// ensure the filter passes on tokens if necessary
		testFilterPasses("course(title = \"Interpretation of Computer Programs 2\", profile = SoftwareEngineering).");

		// ensure the created filter rejects facts with an incorrect header
		testFilterRejects("seminar(title = \"Techniques of Aspect-Oriented Software Design\", profile = SoftwareEngineering).");
		
		// ensure the filter passes on tokens if necessary
		testFilterPasses("course(title = \"Machine Learning\", profile = ArtificialIntelligence).");
		
		// Reset attribute 2 to be a ProfileSoftwareEngineering type
		Attribute	theProfile	= ExpressionFactory.type("SoftwareEngineering");
		
		// Properly index the attribute
		{	Vector		theIndex	= new Vector();
			theIndex.add("profile");
		
			theProfile.setIndex(theIndex); }
		
		// Add a filter to the filter node to check the presence of the symbol
		theCompiledFilter.addNewFilter(new FilterAttribute(theProfile));
		
		// Reset the test node and ask the filter node to reevaluate its constraints 
		itsTestNode.reset();
		theCompiledFilter.reevaluate();
		
		// this will retract two facts (both courses that were propagated earlier)
		// and subsequently assert only the course with the proper profile
		itsTestNode.assertTriggered(1, 2);
	}
	
	public void testTypeFilter() {
		// Make a filter for typeTest(#Object)
		GenericFact theFilterPrescription;
		
		{ 
			theFilterPrescription = ExpressionFactory.unqualifiedFact("credit");
		
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theFilterPrescription);
			
			// Attribute 1 is a type filter for the Object type
			Attribute theCourseType = ExpressionFactory.type("Course");
			theFilterPrescriptionIndexer.nextAttribute("kind", theCourseType);

			// Attribute 2 is an ?courseName variable
			Attribute theTitle = ExpressionFactory.variable("title");
			theFilterPrescriptionIndexer.nextAttribute("title", theTitle);

			// Attribute 2 is an ?amount variable
			Attribute theCredits = ExpressionFactory.variable("credits");
			theFilterPrescriptionIndexer.nextAttribute("credits", theCredits);
			
			// Simulate the use of an extends expression saying that a Seminar is a Course
			ExpressionFactory.registerSubtypeRelationship("Seminar", new Type[] { (Type)theCourseType });
		}
		
		FilterNode theCompiledFilter = compileFilterTestNetwork(theFilterPrescription);

		// ensure the filter passes on tokens if necessary
		testFilterPasses("credit(kind = Course, title = \"Interpretation of Computer Programs 2\", credits = 5).");

		// ensure the filter passes facts which use a valid subtype
		testFilterPasses("credit(kind = Seminar, title = \"Techniques of Aspect-Oriented Software Design\", credits = 5).");
		
		// ensure the filter rejects facts if the type is not a valid subtype
		testFilterRejects("credit(kind = GroupProject, title = \"Software Engineering\", credits = 7).");		

	}
	
	public void testTypeForObjectFilter() {
		// Make a filter for typeTest(#Object)
		GenericFact theFilterPrescription;
		
		{ 
			theFilterPrescription = ExpressionFactory.unqualifiedFact("credit");
		
			// To build a filter node, all attributes need to be indexed. This is done by the following visitor
			IndexingVisitor theFilterPrescriptionIndexer = new IndexingVisitor(theFilterPrescription);
			
			// Attribute 1 is a type filter for the Object type
			Attribute theCourseType = ExpressionFactory.type("Course");
			theFilterPrescriptionIndexer.nextAttribute("kind", theCourseType);
			
			// Simulate the use of an extends expression saying that a Seminar is a Course
			ExpressionFactory.registerSubtypeRelationship("Seminar", new Type[] { (Type)theCourseType });
		}
		
		FilterNode theCompiledFilter = compileFilterTestNetwork(theFilterPrescription);

		// ensure the filter passes on tokens if necessary
		testFilterPasses("credit(kind = Course(title = \"Interpretation of Computer Programs 2\", credits = 5)).");

		// ensure the filter passes facts which use a valid subtype
		testFilterPasses("credit(kind = Seminar(title = \"Techniques of Aspect-Oriented Software Design\", credits = 5)).");
		
		// ensure the filter rejects facts if the type is not a valid subtype
		testFilterRejects("credit(kind = GroupProject(title = \"Software Engineering\", credits = 7)).");		

	}
}
