package edu.vub.crime.rete.nodes;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.datastructures.List;
import edu.vub.crime.grammar.utils.AttributePicker;
import edu.vub.crime.rete.data.DataFactory;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.rete.data.TokenTag;
import edu.vub.crime.util.Logging;

/** <pre> 
 * <pre>
 * A FindallNode is a special kind of ReteNode.
 * The rule which is used for explanation is : 
 *
 *         likeMyMusic(?genre, ?number) :-
 *              genreILike(?genre),
 *              findall(?person, (
 *                         isPresent(?person, LivingRoom),
 *                         prefers(?person, ?genre)),
 *                      ?similarTastes),
 *              length(?similarTastes, ?number).
 *
 * This rule calculates the number of people who like the same genre as I am in favour of.
 * This will be done for each genre I prefer.
 * 
 * The Rete network for this rule (until the FindallNode) looks like :
 *
 *     FILTER: genreILike           FILTER: isPresent           FILTER: prefers
 *             |                             |                          |
 *              \                    FILTER: LivingRoom                 |
 *               \                           |                          |
 *                \                           \                        /
 *                 \                           \                      /   
 *                  \                          JOIN: ?person = ? person 
 *                   \                                   |
 *                    \                                 /
 *                     \                               / 
 *                      \                             /
 *                           JOIN: ?genre = ?genre
 *                                    |
 *                                    |
 *                                 FINDALL       
 *                                    |
 *
 * In the example, we consider that I like rock and pop, so the facts 
 * genreILike(Rock) and genreILike(Pop) are part of the Rete memory (Memory class).
 * Alice likes rock and Bob likes rock and pop, so those facts are also part of the memory : 
 * prefers(Alice, Rock), prefers(Bob, Rock) and prefers(Bob, Pop).
 * So, the Rete network is updated and looks like
 * 
 *     FILTER: genreILike           FILTER: isPresent           FILTER: prefers
 *             |                             |                          |
 *              \                    FILTER: LivingRoom                 |
 *               \                           |                          |
 *                \                           \                        /
 *                 \                           \                      /   
 *                  \             LEFT MEMORY   JOIN: ?person = ? person     RIGHT MEMORY 
 *                   \            ===========            |                   ============
 *                    \                                 /                    Alice, Rock
 *                     \                               /                     Bob, Rock
 *                      \                             /                      Bob, Pop
 *         LEFT MEMORY       JOIN: ?genre = ?genre      RIGHT MEMORY
 *         ===========                |                 ============
 *         Rock                       |
 *         Pop                     FINDALL       
 *                                    |
 * 
 * When the fact isPresent(Alice, LivingRoom) is inserted in the network, 
 * the token < add, genreILike(Rock), isPresent(Alice, LivingRoom), prefers(Alice, Rock) >
 * will be inserted in this node.                                     
 * The cache contains for every variable binding (?genre in this example) the persons
 * who are present in the LivingRoom and prefer the same genre as I do.
 * So, the cache will contain a pair: [ ?genre = Rock, {Alice} ].
 * The token that will be sent to this node's children is : 
 * < add, genreILike(Rock), similarTastes(Alice) >.
 * The resulting Rete network is depicted below.
 *     
 *         FILTER: genreILike           FILTER: isPresent           FILTER: prefers
 *             |                             |                          |
 *              \                    FILTER: LivingRoom                 |
 *               \                           |                          |
 *                \                           \                        /
 *                 \                           \                      /   
 *                  \             LEFT MEMORY   JOIN: ?person = ? person     RIGHT MEMORY 
 *                   \            ===========            |                   ============
 *                    \                                 /                    Alice, Rock
 *                     \                               /                     Bob, Rock
 *                      \                             /                      Bob, Pop
 *         LEFT MEMORY       JOIN: ?genre = ?genre      RIGHT MEMORY
 *         ===========                |                 ============
 *         Rock                       |
 *         Pop                     FINDALL       CACHE       
 *                                    |          =====
 *                                    |          Rock, [Alice] 
 *                                    |
 *                                    
 * When < add, genreILike(Pop), isPresent(Bob, LivingRoom), prefers(Bob, Pop) >
 * is also inserted (because the fact isPresent(Bob, LivingRoom) is added), 
 * the cache will be updated and will contain the pairs
 * [ ?genre = Rock, {Alice} ] and [ ?genre = Pop, {Bob} ].
 * The token, which has been previously sent to the children, 
 * < add, genreILike(Rock), similarTastes(Alice) > doesn't need to be retracted, because
 * there's a difference between the variable bindings (the variable ?genre differs).
 * The token that will be sent to the children is
 * < add, genreILike(Pop), isPresent(Bob, LivingRoom), prefers(Bob, Rock) >.
 *         
 * When < add, genreILike(Rock), isPresent(Bob, LivingRoom), prefers(Bob, Rock) > 
 * is inserted as well, the previously sent token, 
 * < add, genreILike(Rock), similarTastes(Alice) >, should be retracted.
 * The cache is updated to contain the pairs 
 * [ ?genre = Rock, {Alice, Bob} ] and [ ?genre = Pop, {Bob} ].
 * The tokens that should be sent to the children is
 * < remove, genreILike(Rock), similarTastes(Alice) > and
 * < add, genreILike(Rock), similarTastes(Alice, Bob) >.
 * The Rete network at this point is represented below:
 * 
 *     FILTER: genreILike           FILTER: isPresent           FILTER: prefers
 *             |                             |                          |
 *              \                    FILTER: LivingRoom                 |
 *               \                           |                          |
 *                \                           \                        /
 *                 \                           \                      /   
 *                  \             LEFT MEMORY   JOIN: ?person = ? person     RIGHT MEMORY 
 *                   \            ===========            |                   ============
 *                    \                                 /                    Alice, Rock
 *                     \                               /                     Bob, Rock
 *                      \                             /                      Bob, Pop
 *         LEFT MEMORY       JOIN: ?genre = ?genre      RIGHT MEMORY
 *         ===========                |                 ============
 *         Rock                       |
 *         Pop                     FINDALL       CACHE       
 *                                    |          =====
 *                                    |          Rock, [Alice, Bob] 
 *                                    |          Pop, [Bob]
 * </pre>
 * </pre> */ 
public class FindallNode extends Node {	

	/** <pre>
	 * cache_ : keeps all the attributes which need to be accumulated in the accumulator
	 *          In the example above, the accumulator is the variable ?similarTastes.
	 *          When the token < add, genreILike(Rock), isPresent(Alice, LivingRoom), prefers(Alice, Rock) >
	 *          is inserted in this node, (and the cache was empty before), the cache will contain 
	 *          an attribute Alice afterwards. 
	 *          These attributes are stored by the previously variable bindings (= variable bindings which
	 *          took not place in the queries of the findall statement).
	 *          So, the cache is a mapping from a token, which contains those variable bindings), to
	 *          the list of attributes.
	 * accumulator_ : variable, needed to form the new fact
	 *                this fact will look like similarTastes(...)  
	 * picker_ : picker which picks the attribute
	 * previous_ : number of facts (conditions) for the findall statement.
	 *             In this example, it will be one.
	 *             This makes it possible to select only those bindings from the fact
	 *             that will be inserted in this node.
	 *             So, when the inserted token is 
	 *             < add, genreILike(Rock), isPresent(Alice, LivingRoom), prefers(Alice, Rock) >,
	 *             the "selected" token looks like < add, genreILike(Rock) >.
	 *             This token will be used in the cache.   
	 * </pre> */
	protected HashMap 			itsCacheMappingTokensToPreviouslyRecordedMatches;
	protected AttributePicker	itsPickerToExtractValuesFromExtendedTokens;
	protected int				noOfFactsProducedByThePrecedingQuery;
	
	
	/** <pre>
	 * Constructor.
	 * </pre> */
	public FindallNode(AttributePicker thePicker, int theLengthOfThePrecedingQuery) {
		super();
		itsCacheMappingTokensToPreviouslyRecordedMatches	= new HashMap();
		itsPickerToExtractValuesFromExtendedTokens			= thePicker;		
		noOfFactsProducedByThePrecedingQuery				= theLengthOfThePrecedingQuery;
	}
	
	/** <pre> 
	 * When a new token is inserted, we can select the needed attribute (= binding for variable ?person in the example)
	 * (because the right picker, picker_, is known) and we can update the cache with this attribute.
	 * 
	 * After the attribute has been picked from the token, the token is shortened (it does not longer contain
	 * the variable bindings which took place in the findall statement).
	 * When the inserted token has a tag equal to add, the new attribute should be inserted in the cache.
	 * When there was already a binding for that shortened token, a previously sent token should be retracted.
	 * So, this token is added to the vector tokens, which contains the tokens that will be sent.
	 * When there wasn't a binding for that shortened token available, a new "pair" is added to the cache
	 * (no retraction needed in this case).
	 * A togen wit tag equal to remove is handled similarly, except that there's an error thrown when an attribute
	 * needs to be removed which was never added to the cache in the first place.
	 * 
	 * Thereafter, the token, updated with the accumulation of the attributes in a fact, is added to the vector tokens.
	 * And afterwards, that vector is sent to the children of this node.
	 * </pre> */
	public void insert(Token newToken) {
		try {
			// First: extract the accumulated variable
			Attribute accumulatedValue = itsPickerToExtractValuesFromExtendedTokens.pickAttribute(newToken);
			
			// Subsequently, reduce the token such that specific elements of the findall query are omitted
			Token theStrippedToken = stripToken(newToken, noOfFactsProducedByThePrecedingQuery);
			
			// Prepare a vector to contain all tokens to be emitted in response
			Vector theTokensToBeEmitted = new Vector(2);
			
			switch(theStrippedToken.getTag()) {
				case TokenTag.add:
					if(itsCacheMappingTokensToPreviouslyRecordedMatches.containsKey(theStrippedToken))
						expandMatches(theStrippedToken, accumulatedValue, theTokensToBeEmitted);
					else
						newMatches(theStrippedToken, accumulatedValue, theTokensToBeEmitted);
					
					break;
					
				case TokenTag.remove:
					// The cache contains only positive tokens, so we need to flip the sign
					Token thePositiveStrippedToken = DataFactory.addToken();
					theStrippedToken.copyFacts(thePositiveStrippedToken);
					
					if(itsCacheMappingTokensToPreviouslyRecordedMatches.containsKey(theStrippedToken))
						reduceMatches(theStrippedToken, accumulatedValue, theTokensToBeEmitted);
					
					break;
			}
			
			send(theTokensToBeEmitted);
		} catch (IncorrectAttributeIndexException e) {
			Logging.Engine_LOG.error("Indexing error when attempting to pick the accumulated variable in a FindallNode", e);
		}
	}
	
	/** <pre>
	 * Remove facts from a given token.
	 * All facts after a given index, should be removed from the token.
	 * A new token must be instantiated, otherwise a reference will be passed,
	 * which will cause problems.
	 * </pre> */
	public Token stripToken(Token tkn, int index) {
		Token result = DataFactory.token(tkn.getTag());
		result = tkn.copyFacts(result);
		for(int i=index; i != tkn.size(); i++) {
			result.removeLast();
		}
		return result;
	}
	
	/** <pre>
	 * Send the given tokens to the children of this node.
	 * </pre> */ 
	public void send(Vector tokens) {
		for(Iterator tkn=tokens.iterator(); tkn.hasNext(); ) {
			Token t = (Token) tkn.next();
			
			record(t);
			for(Iterator chd=itsChildNodes.iterator(); chd.hasNext(); ) {
				Node node = (Node) chd.next();
				node.insert(t);
			}
        }
	}
	
	private final void expandMatches(Token theStrippedToken, Attribute theNewValue, Vector theTokensToBeEmitted) {
		// The token to retract previous matches is creates by accumulating all known matches and removing them
		Token retractPreviousMatches = accumulate(theStrippedToken, TokenTag.remove);
		
		// Subsequently, add the new value to the local cache
		List theMatchingAttributes = (List) itsCacheMappingTokensToPreviouslyRecordedMatches.get(theStrippedToken);
		theMatchingAttributes.add(theNewValue);
		
		// The expanded token is then created by accumulating over all matches (including the new value) and adding them
		Token assertExpandedMatches = accumulate(theStrippedToken, TokenTag.add);
		
		// Finally, make sure that both tokens are emitted
		theTokensToBeEmitted.add(retractPreviousMatches);
		theTokensToBeEmitted.add(assertExpandedMatches);
	}
	
	private final void newMatches(Token theStrippedToken, Attribute theNewValue, Vector theTokensToBeEmitted) {
		// Add an entry to the local cache and put the new value in it.
		List theMatchingAttributes = ExpressionFactory.list();
		theMatchingAttributes.add(theNewValue);
		
		itsCacheMappingTokensToPreviouslyRecordedMatches.put(theStrippedToken, theMatchingAttributes);
		
		// The expanded token is then created by accumulating over all matches (i.e. the new value) and adding them
		Token assertExpandedMatches = accumulate(theStrippedToken, TokenTag.add);
		
		// Finally, make sure that the assertion token is emitted
		theTokensToBeEmitted.add(assertExpandedMatches);
	}
	
	private final void reduceMatches(Token theStrippedToken, Attribute theRemovedValue, Vector theTokensToBeEmitted) {
		// The token to retract previous matches is creates by accumulating all known matches and removing them
		Token retractPreviousMatches = accumulate(theStrippedToken, TokenTag.remove);
		
		// Subsequently, add the new value to the local cache
		List theMatchingAttributes = (List) itsCacheMappingTokensToPreviouslyRecordedMatches.get(theStrippedToken);
		theMatchingAttributes.remove(theRemovedValue);
		
		// Add the retraction token so that it will be emitted
		theTokensToBeEmitted.add(retractPreviousMatches);
		
		if(theMatchingAttributes.isEmpty()) {
			// If this was the last remaining match, all that's let to do is remove the entry from the cache
			itsCacheMappingTokensToPreviouslyRecordedMatches.remove(theStrippedToken);
		} else {
			// Otherwise emit a new token with the remaining matches
			
			// The reduced token is then created by accumulating over all matches (excluding the removed value) and adding them
			Token assertReducedMatches = accumulate(theStrippedToken, TokenTag.add);

			// Ensure that the reduced assertion token is emitted
			theTokensToBeEmitted.add(assertReducedMatches);
		}
	}

	/**
	 * Adds an accumulation fact to a given token, which can then be propagated further.
	 * The fact is of the form findall(accumulated = [non-empty list]).
	 * <p>
	 * The token that is returned is a copy of the input parameter to avoid passing a 
	 * reference to a token (which can for instance be stored in a memory), which can
	 * lead to corruption of the rete network.
	 */
	private Token accumulate(Token theInputToken, int theTagOfTheOutputToken) {
		Token theOutputToken = DataFactory.token(theTagOfTheOutputToken);
		
		// First copy all facts in the input token
		theInputToken.copyFacts(theOutputToken);
		
		// Get all matching attributes from the findall cache
		List theCachedAttributes = (List) itsCacheMappingTokensToPreviouslyRecordedMatches.get(theInputToken);
				
		// Otherwise copy the matching attributes, so we do not give out access to our local cache
		List theOutputAttributes = ExpressionFactory.list();
		
		theCachedAttributes.copyAttributes(theOutputAttributes);
		
		// Instantiate the accumulator fact
		GenericFact theAccumulatorFact = ExpressionFactory.unqualifiedFact("findall");
		theAccumulatorFact.addNamedAttribute("accumulated", theOutputAttributes);
		
		theOutputToken.addFact(theAccumulatorFact);
		
		return theOutputToken;
	}
	
}