/** <pre> 
 * @file JoinNode.java
 * @author Christophe Scholliers
 * @date 2006-09-23
 * </pre> */

package edu.vub.crime.rete.nodes;

import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.rete.combiner.Combiner;
import edu.vub.crime.rete.data.DataFactory;
import edu.vub.crime.rete.data.Token;

/** 
 * A JoinNode is a ReteNode which combines two input tokens and verifies whether a combined token
 * may be passed on. To do this, the JoinNode evaluates whether in both tokens the bindings for
 * the shared variables are compatible.
 * <p>
 * The JoinNode needs a means to be able to distinguish where an input comes from, hence it provides
 * methods to access its left and right input which can be connected to the rest of the Rete network.
 * It is incorrect to directly attach a JoinNode to the network, and doing so will result in the 
 * throwing of a RuntimeException the first time a token is passed to it.
 */
public class JoinNode extends Node {
	
	/** 
	 * The LeftNode inner class represents the left input of the Join node.
	 */
	private class LeftNode extends Node {
				
		/** 
		 * The left parent node has propagated a token to the join node. Notify the
		 * combiner to determine which tokens should be propagated to the Join node's
		 * child nodes. (It is possible that there's no matching token in the right 
		 * memory for the newly inserted token; in that case, no tokens need to be 
		 * sent to the children).
		 */
		public void insert(Token newToken) {
			// First of all, make a copy of the token
			// This copy will be kept in the node's memory, rather than the original
			Token token = DataFactory.token(newToken.getTag());
			newToken.copyFacts(token);
			
			record(token);
			combineRight(token);
		}
	}
	
	/** 
	 * The RightNode inner class represents the right input of the Join node.
	 */
	private class RightNode extends Node {
		
		/** 
		 * The right parent node has propagated a token to the join node. Notify the
		 * combiner to determine which tokens should be propagated to the Join node's
		 * child nodes. (It is possible that there's no matching token in the left 
		 * memory for the newly inserted token; in that case, no tokens need to be 
		 * sent to the children).
		 */
		public void insert(Token newToken) {
			// First of all, make a copy of the token
			// This copy will be kept in the node's memory, rather than the original
			Token token = DataFactory.token(newToken.getTag());
			newToken.copyFacts(token);
			
			record(token);
			combineLeft(token);

		} 
	}
	
	private LeftNode	itsLeftInput;
	private RightNode	itsRightInput;
	private Combiner	itsCombiner;
	
	/** 
	 * Constructs a Join node which produces tokens by combining tokens from both of its inputs 
	 * provided that these tokens are compatible.
	 * <p>
	 * The combiner that is provided in this constructor determines the notion of compatible 
	 * tokens, most often this implies that to be passed the memory of the opposite input must
	 * contain one or more facts that have the same values for all corresponding variables. 
	 * When using negation, however, the inverse is true, when a token exists with the same
	 * bindings for all variables, the token should not be passed.
	 */
	public JoinNode(Combiner theCombiner) {
		super();
		itsCombiner		= theCombiner;
		itsLeftInput 	= new LeftNode();
		itsRightInput	= new RightNode();
		
		itsCombiner.setLeftMemory(	itsLeftInput.itsMemory);
		itsCombiner.setRightMemory(	itsRightInput.itsMemory);
		itsCombiner.setConstraints(	itsJoins);
	}
	
	public Node itsLeftInput() {
		return itsLeftInput;
	}

	public RightNode itsRightInput() {
		return itsRightInput;
	}
	
	/** 
	 * This method raises a {@link RuntimeException} to signal that the Join node has been 
	 * incorrectly wired into the network. Join nodes should not be connected to directly
	 * but rather through their left and right inputs.
	 */
	public void insert(Token theToken) {
		throw new RuntimeException("JoinNode should never be attached directly to another node but only through its left or right inputs");
	}
	
	/** 
	 * This method is called by the left input when a new token was propagated to it. The 
	 * {@link Combiner} is then requested to verify which tokens need to be propagated in
	 * response. The method then will pass on these tokens after having verified whether
	 * they satisfy all filters.
	 * */
	private void combineLeft(Token theLeftToken) {
		Vector theCombinedTokens = itsCombiner.combineLeft(theLeftToken);
		
		for(Iterator combinedTokenIt = theCombinedTokens.iterator(); combinedTokenIt.hasNext(); ) {
			Token currentCombinedToken = (Token) combinedTokenIt.next();
			
			propagate(currentCombinedToken);
		}
	}
	
	/** 
	 * This method is called by the right input when a new token was propagated to it. The 
	 * {@link Combiner} is then requested to verify which tokens need to be propagated in
	 * response. The method then will pass on these tokens after having verified whether
	 * they satisfy all filters.
	 * */
	private void combineRight(Token theRightToken) {
		Vector theCombinedTokens = itsCombiner.combineRight(theRightToken);
		
		for(Iterator combinedTokenIt = theCombinedTokens.iterator(); combinedTokenIt.hasNext(); ) {
			Token currentCombinedToken = (Token) combinedTokenIt.next();
			
			propagate(currentCombinedToken);
		}
	} 	

	/**
	 * This method propagates a token that is the result of either a call to the auxiliary
	 * combineLeft or combineRight methods. Before propagating the token to all child nodes,
	 * this method verifies whether the imposed filters are satisfied.
	 */
	protected void propagate(Token theCombinedToken) {
		if(pass(theCombinedToken)) {
			record(theCombinedToken);
			
			for(Iterator childNodeIt = itsChildNodes.iterator(); childNodeIt.hasNext(); ) {
				Node currentChildNode = (Node) childNodeIt.next();
				
				currentChildNode.insert(theCombinedToken);
			}
		}
	}
	
	public void unhook(Node theChildNode) {
		super.unhook(theChildNode);
		
		itsLeftInput.unhook(this);
		itsRightInput.unhook(this);
	}
}
