package edu.vub.crime.rete.nodes;

/**
 * 
 * @author smostinc
 */
public class JoinNodeTest extends NodeTest {
		
	/**
	 * Tests the joining of flat facts which use no data structures as attributes and which need to be
	 * joined over a single shared variable.
	 */
	public void testFlatJoinNode() {
		compileJoinTestNetwork("attending(student = ?s, course = ?c), weight(course = ?c, weigth = ?w).");
		
		// The test node is not yet triggered when adding a single fact 
		testFilterRejects("attending(student = Stijn, course = SoftwareEngineering).");
		
		// Adding a fact with a non-matching variable does not trigger the join node either
		testFilterRejects("weight(course = Compilers, weight = 8).");

		// Adding a fact on the right-hand side triggers the join node
		testFilterPasses("weight(course = SoftwareEngineering, weight = 7).");

		// Adding a fact on the left also triggers the join node
		testFilterPasses("attending(student = Jona, course = SoftwareEngineering).");
		testFilterPasses("attending(student = Stijn, course = Compilers).");
	}

	/**
	 * Tests the joining of two facts which share one flat variable as well as one variable which is
	 * featured in a list attribute in one branch. 
	 */
	public void testListJoinNode() {
		// make a network to determine which courses are particular student attends which are part of his major
		compileJoinTestNetwork("profiles(student = ?s, profiles = [ ?major | ?other ]), attending(student = ?s, course = ?c, profile = ?major).");
		
		// The test node does not trigger by adding a student on the left-hand side
		testFilterRejects("profiles(student = Stijn, profiles = [ SoftwareEngineering, AdaptiveSystems]).");
		
		// The test node does not trigger by adding an unrelated fact on the right-hand side
		testFilterRejects("attending(student = Jona, course = \"Machine Learning\", profile = AdaptiveSystems).");
		
		// The test node does not trigger if only the ?student fact matches (i.e. the ?major variable in the
		// list from the right-hand side of the node is properly joined).
		testFilterRejects("attending(student = Stijn, course = \"Adaptive Systems 2\", profile = AdaptiveSystems).");
		
		// The test node is however triggered when a matching fact is added from the right
		testFilterPasses("attending(student = Stijn, course = \"Compilers\", profile = SoftwareEngineering).");
		
		// or from the left
		testFilterPasses("profiles(student = Jona, profiles = [ AdaptiveSystems, SoftwareEngineering]).");
	}
	
	/**
	 * Tests the joining of two facts which share one flat variable as well as one variable which is
	 * featured in an object attribute in one branch. 
	 */
	public void testObjectJoinNode() {
		// make a network to determine which courses are particular student attends which are part of his major
		compileJoinTestNetwork("profiles(student = ?student, profiles = [ ?major | ?other ]), attending(student = ?student, course = Course( title = ?course, profile = ?major)).");
		
		// The test node does not trigger by adding a student on the left-hand side
		testFilterRejects("profiles(student = Stijn, profiles = [ SoftwareEngineering, AdaptiveSystems]).");
		
		// The test node does not trigger by adding an unrelated fact on the right-hand side
		testFilterRejects("attending(student = Jona,  course = Course( title = \"Machine Learning\", profile = AdaptiveSystems)).");
		
		// The test node does not trigger if only the ?student fact matches (i.e. the ?major variable in the
		// list from the right-hand side of the node is properly joined).
		testFilterRejects("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).");
		
		// The test node is however triggered when a matching fact is added from the right
		testFilterPasses("attending(student = Stijn,  course = Course( title = \"Compilers\", profile = SoftwareEngineering)).");
		
		// or from the left
		testFilterPasses("profiles(student = Jona, profiles = [ AdaptiveSystems, SoftwareEngineering]).");			
	}
	
	/**
	 * Tests the use of a negated join node, which is constructed whenever a negation is not used in the leftmost
	 * position. 
	 */
	public void testNegatedJoinNode() {
		// make a network to determine which courses a particular student attends which are not part of his major
		compileJoinTestNetwork("attending(student = ?student, course = Course( title = ?course, profile = ?major)), not profiles(student = ?student, profiles = [ ?major | ?other ]).");
		
		// First feed some facts in the left branch of the join node, to signal which courses I follow
		// These facts are all asserted since the major profile is not yet set on the right-hand side.
		testFilterAsserts("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).");
		testFilterAsserts("attending(student = Stijn, course = Course( title = \"Compilers\", profile = SoftwareEngineering)).");
		
		// Setting the major profile of other students has no effect
		testFilterRejects("profiles(student = Jona,   profiles = [ AdaptiveSystems, SoftwareEngineering]).");
		testFilterRejects("profiles(student = Joke,   profiles = [ ArtificialIntelligence, SoftwareEngineering]).");
				
		// Setting the major profile, will cause the retraction of a single fact
		testFilterRetracts("profiles(student = Stijn, profiles = [ SoftwareEngineering, AdaptiveSystems]).");
		
		// Adding a fact to the left-hand side which does not match the filter (i.e. it is a course within 
		// the major profile) does not trigger the test node.
		testFilterRejects("attending(student = Jona,  course = Course( title = \"Machine Learning\", profile = AdaptiveSystems)).");

		// If no tokens were produced by the assertion of a fact, removing it should not have any effect either
		testRemovalRejects("profiles(student = Joke,   profiles = [ ArtificialIntelligence, SoftwareEngineering]).");
		
		// On the other hand, removing fact which removed facts will ensure that these facts are reasserted
		testRemovalAsserts("profiles(student = Stijn,  profiles = [ SoftwareEngineering, AdaptiveSystems]).");
		
		// Removing a fact which passed the filter will result in a retract token being propagated
		testRemovalRetracts("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).");
				
		// Removing a course which did not produce any token (i.e. a course inside the main curriculum)
		// can be removed from the right-hand side of the negated join and does not produce a token.
		testRemovalRejects("attending(student = Jona,  course = Course( title = \"Machine Learning\", profile = AdaptiveSystems)).");
		
	}
	
	/**
	 * Tests the use of a findall node which is a beta layer node although it has only one input node.
	 */
	public void testFindallJoinNode() throws Exception {
		// make a network to determine which courses need to be organised since they are in the curriculum 
		// of at least one student
		compileJoinTestNetwork("findall(?course, [ attending(student = ?student, course = Course( title = ?course )) ], ?curriculum).");
		
		// Adding a single fact will yield the assertion of a single fact containing the accumulated 
		// curriculum, which is currently a list with a single element
		testFilterAsserts("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).");
		
		// Adding another fact for another student, will yield an update:
		// The previous token with the single course to be organised will be retracted
		// A new token listing both courses will be asserted
		testFilterExpands("attending(student = Jona,  course = Course( title = \"Compilers\", profile = SoftwareEngineering)).");
		
		// In accordance with the prolog findall semantics the list returned may contain doubles
		// hence adding a fact saying that another student is taking up a particular course does
		// add an new entry
		testFilterExpands("attending(student = Stijn,  course = Course( title = \"Compilers\", profile = SoftwareEngineering)).");
				
		// Retracting a fact for an existing student will shrink the curriculum if that student was 
		// the only student registered for a course.
		testFilterShrinks("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).");
	}
	
	/**
	 * Tests the use of a bagof node which is a beta layer node although it has only one input node.
	 */
	public void testBagofJoinNode() throws Exception {
		// make a network to determine the curriculum of a student
		compileJoinTestNetwork("bagof(?course, [ attending(student = ?student, course = Course( title = ?course )) ], ?curriculum).");
		
		// Adding a single fact will yield the assertion of a single fact containing the accumulated 
		// curriculum, which is currently a list with a single element
		testFilterAsserts("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).");
		
		// Adding another fact for the same student, will yield an update:
		// The previous token with the single-course curriculum will be retracted
		// A new token with a two-course curriculum will be asserted
		testFilterExpands("attending(student = Stijn,  course = Course( title = \"Compilers\", profile = SoftwareEngineering)).");
		
		// Adding a fact for another student does not affect the previous curriculum
		// A new fact is simply created, with a single-course curriculum
		testFilterAsserts("attending(student = Jona,  course = Course( title = \"Machine Learning\", profile = AdaptiveSystems)).");
		
		// Retracting a fact for an existing student will shrink the curriculum accordingly
		testFilterShrinks("attending(student = Stijn, course = Course( title = \"Compilers\", profile = SoftwareEngineering)).");
	}
	
	public void testLengthNode() throws Exception {
		// make a network to count find students who only attend one course
		compileJoinTestNetwork("bagof(?course, [ attending(student = ?student, course = Course( title = ?course )) ], ?curriculum), length(?curriculum, 1).");
		
		// Adding a single fact will yield in the assertion of a single token, since there is now a student which has
		// a curriculum consisting of a single course.
		testFilterAsserts("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).");
		
		// Adding a fact about an unrelated student will assert a new fact, as this student also has a single course 
		// registered in his curriculum.
		testFilterAsserts("attending(student = Jona,  course = Course( title = \"Machine Learning\", profile = AdaptiveSystems)).");
		
		// Adding a fact about a student which already had a course registered results in the retraction of a fact
		// since that student now has a curriculum consisting of more than one course.
		testFilterRetracts("attending(student = Stijn,  course = Course( title = \"Compilers\", profile = SoftwareEngineering)).");
		
		// Removing one of the two registered courses does cause the fact to be reasserted
		testRemovalAsserts("attending(student = Stijn, course = Course( title = \"Adaptive Systems 2\", profile = AdaptiveSystems)).");
	}
}
