package edu.vub.crime.rete.nodes;

import edu.vub.crime.exceptions.IncorrectAttributeIndexException;
import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.constant.Number;
import edu.vub.crime.grammar.attributes.datastructures.List;
import edu.vub.crime.grammar.utils.AttributePicker;
import edu.vub.crime.rete.data.DataFactory;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.util.Logging;

/** 
 * A LengthNode is used to determine (or restrict) the length of a list attribute. It is 
 * commonly used in combination with the FindAll and BagOf constructs which produce such
 * dynamic lists.
 * <p>
 * Consider the running example where we want to determine how many people in the living
 * room share my taste in music. This could be determined using the following rule:
 * <pre>
 * ShareTasteInMusic(?genre, ?number) :-
 *   MusicILike(genre = ?genre),
 *   findall(
 *     ?person, 
 *     [ IsPresent(person = ?person, location = `LivingRoom),
 *       Prefers(person = ?person, genre = ?genre)],
 *     ?peopleWithSimilarTastes),
 *   length(?peopleWithSimilarTastes, ?number); 
 * </pre>
 * <p>
 * Such a rule (which computes how many people share my taste in music for every genre I
 * marked as MusicILike) would be compiled to the following network:
 * 
 * <img src="ShareTasteInMusic.png" alt="<pre>
 *     FILTER: MusicILike           FILTER: IsPresent           FILTER: Prefers
 *             |                             |                         |
 *              \                    FILTER: LivingRoom                |
 *               \                           |                         |
 *                \                           \                       /
 *                 \                           \                     /   
 *                  \                          JOIN: ?person = ?person 
 *                   \                                   |
 *                    \                                 /
 *                     \                               / 
 *                      \                             /
 *                           JOIN: ?genre = ?genre
 *                                    |
 *                                    |
 *                                 FINDALL       
 *                                    |
 *                                 LENGTH
 * </pre>" />
 */ 
public class LengthNode extends Node {	

	private final AttributePicker	itsListPicker;
	private final AttributePicker	possiblyItsLengthPicker;
	
	private final boolean			wasLengthBound;
	
	public LengthNode(boolean isLengthBound, AttributePicker theListPicker, AttributePicker possiblyTheLengthPicker) {
		wasLengthBound			= isLengthBound;
		itsListPicker			= theListPicker;
		possiblyItsLengthPicker	= possiblyTheLengthPicker;
	}
		
	/**
	 * When a new token is inserted, the length node will isolate the list from the token
	 * and compute its length.  Subsequently,  it will determine whether the list has the
	 * correct length (if the length was bound at construction time of the node). Finally,
	 * a new fact is added to the token and propagated further down.
	 */
	public void insert(Token newToken) {
		try {
			List		thePickedList	= (List) itsListPicker.pickAttribute(newToken);
			int			theListSize		= thePickedList.length();
			
			if(wasLengthBound) {
				Number	thePickedLength	= (Number) possiblyItsLengthPicker.pickAttribute(newToken);
				
				// Only propagate if the length of the picked list is equal to the specified number.
				if(theListSize != thePickedLength.getValue()) return; 
			}
			
			Token		toPropagate		= DataFactory.token(newToken.getTag());
			newToken.copyFacts(toPropagate);
			
			Attribute	theLength		= ExpressionFactory.attribute(theListSize);
			GenericFact	theLengthFact	= ExpressionFactory.unqualifiedFact("length");
			
			theLengthFact.addNamedAttribute("list",	thePickedList);
			theLengthFact.addNamedAttribute("length",	theLength);
			toPropagate.addFact(theLengthFact);
			
			record(toPropagate);
			propagate(toPropagate);
		} catch (IncorrectAttributeIndexException e) {
			// If the list (and possibly the expected length) could not be picked out of the token,
			// report the error and do not propagate the token further down the network
			Logging.Engine_LOG.error("Could not pick an attribute to determine whether a token should be propagated", e);
		}
	}
}