package edu.vub.crime.rete.nodes;

import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.GenericFact;
import edu.vub.crime.rete.combiner.NegationCombiner;
import edu.vub.crime.rete.data.DataFactory;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.rete.data.TokenTag;

/** 
 * This node is constructed only if the first element of a rule's query contains a negation.
 * If this is the case, the negated fact can only contain constants, as all variables inside
 * a negation must always be bound prior to the occurrence of the negation itself.
 * <p>
 * When deploying a rule, which uses a negation as its first (and possibly only) clause,  it
 * is possible that the rule triggers in the absence of information. This behavior is encoded
 * in this class.
 * <p>
 * Occurrences of negations in a rule's query in any position other than the first are dealt
 * with using the {@link NegationCombiner} in a join node. 
 */
public class NegationFilterNode extends Node {
	
	private Token 		itsPresenceIndicator;
	private GenericFact	itsAbsenceIndicatorFact;
	
	public NegationFilterNode(){
		// Do not propagate an absence indicator at construction time.
		itsPresenceIndicator = null;
		itsAbsenceIndicatorFact = ExpressionFactory.unqualifiedFact(":absenceIndicator:" + this.toString());
	}
	
	private void assertAbsenceIndicator() {
		Token theAbsenceIndicator	= DataFactory.addToken(itsAbsenceIndicatorFact);
		itsPresenceIndicator 		= null;
		
		itsMemory.insert(theAbsenceIndicator);
		propagate(theAbsenceIndicator);
	}
	
	private void retractAbsenceIndicator(Token thePresenceIndicator) {
		itsMemory.clear();
		
		itsPresenceIndicator = thePresenceIndicator;
		
		propagate(DataFactory.removeToken(itsAbsenceIndicatorFact));
	}
	
	public void insert(Token token) {	     
		
		switch(token.getTag()) {
			case TokenTag.add: 
				// If a token is added, this means that the working memory contains a fact 
				// that matches the given specification.  As a consequence, the token that
				// was previously propagated to signal  the absence of such a fact must be
				// retracted.
				retractAbsenceIndicator(token);
				break;

			case TokenTag.remove:
				// If a token is removed, this means that the working memory no longer
				// contains a fact that matches the given specification. Hence, we can
				// propagate a token to signal the absence of such a fact.
				assertAbsenceIndicator();
				break;
		}	
	}  
	
	/**
	 * This method is called when building the network, after the negated filter node is connected. 
	 * If the working memory of the rete network would contain a token matching the negated filter,
	 * it will already have been added at this point.
	 * <p>
	 * If no token has been added, this filter node should emit a dummy token to signal to the rete
	 * network below that its conditions are met: i.e. there is not fact in the working memory that
	 * matches the given description. 
	 */
	public void triggerIfAbsent() {
		if(itsPresenceIndicator == null) 
			assertAbsenceIndicator();
	}
}