package edu.vub.crime.rete.nodes;

import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.rete.data.MemoryTable;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.rete.data.TokenTag;
import edu.vub.crime.rete.filter.Filter;
import edu.vub.crime.rete.filter.Join;

/**
 * A ReteNode is used to propagate Tokens, filtering and joining them in such a way that they
 * represent all valid supporting evidence to trigger the activation of a rule. As data flows
 * through the nodes, the intermediate results are cached, such that the addition or removal 
 * of data can be performed without recomputing derivations from previously available data.
 * <p>
 * Each node has a slot for Joins and Filters. The Joins verify the conformance of variable
 * bindings in two tokens whereas the Filters check whether the resulting tokens have a proper
 * attribute value at a given position.
 */
public abstract class Node {
	
	protected final Vector		itsJoins		= new Vector();
	protected final Vector		itsFilters  	= new Vector();
	protected final Vector		itsChildNodes	= new Vector();

	protected final MemoryTable itsMemory		= new MemoryTable();
	protected 		Node		itsParent		= null;
	
	public void addNewJoin(Join theNewJoin) {
		itsJoins.add(theNewJoin);
	}
	
	public void addNewFilter(Filter theNewFilter) {
		itsFilters.add(theNewFilter);
	}
	
	public final void addChild(Node theNewNode) {
		// Each node may only have one parent
		if(theNewNode.itsParent == null) {
			itsChildNodes.add(theNewNode);
			theNewNode.itsParent = this;
			
			// Initialize the child node with all the tokens in the memory;
			for(Iterator tokenIt = itsMemory.iterator(); tokenIt.hasNext(); ) {
				Token currentToken = (Token)tokenIt.next();
				
				theNewNode.insert(currentToken);
			}

		} else {
			throw new RuntimeException("Asked to be the parent of a node which already has a parent.");
		}
	}
			
	abstract public void insert(Token token);

	protected boolean join(Token theLeftToken, Token theRightToken) {
		for (Iterator joinIt = itsJoins.iterator(); joinIt.hasNext();) {
			Join currentJoin = (Join) joinIt.next();
			
			if(! currentJoin.join(theLeftToken, theRightToken)) {
				return false;
			}
		}
		
		return true;
	}
	
	protected boolean pass(Token theCombinedToken) {
		for (Iterator filterIt = itsFilters.iterator(); filterIt.hasNext();) {
			Filter currentFilter = (Filter) filterIt.next();
			
			if(! currentFilter.pass(theCombinedToken)) {
				return false;
			}
		}
		
		return true;
	}
	
	public void unhook(Node theChildNode) {
		itsChildNodes.remove(theChildNode);
		
		if(itsChildNodes.isEmpty() && itsParent != null) {
			itsParent.unhook(this);
		}
	}

	/**
	 * Depending on theToken's sign, it will either remove or add the token from this
	 * node's memory. This method is called for every token that is being propagated
	 * by a node.
	 */
	public void record(Token theToken) {
		switch (theToken.getTag()) {
		case TokenTag.add:
			itsMemory.insert(theToken);
			break;

		case TokenTag.remove:
			itsMemory.remove(theToken);
			break;
		}
	}
	
	protected void propagate(Token theToken) {
		for(Iterator childNodeIt = itsChildNodes.iterator(); childNodeIt.hasNext(); ) {
	    	Node currentChildNode = (Node) childNodeIt.next();
	    	
            currentChildNode.insert(theToken);
        }
	}
}