package edu.vub.crime.rete.nodes;

import edu.vub.crime.CrimeEngine;
import edu.vub.crime.grammar.Rule;
import edu.vub.crime.rete.agenda.RuleActivation;
import edu.vub.crime.rete.data.Token;

/** <pre>
 * A ProductionNode is a special kind of ReteNode.
 * It is used for building the Rete network.
 * The Rete network is built by parsing and transforming every rule.
 * A ProductionNode typically forms the last node of a rule.
 * When all the prerequisites of the rule (right-hand-side) are
 * parsed, transformed into the correct filter nodes
 * and pairwise combined using join nodes, the lowest join node
 * has a ProductionNode as its child.
 * This ProductionNode contains the rule and the actual CrimeEngine.
 * The node will insert a new activation, in this case a RuleActivation, 
 * into the engine's agenda.
 * 
 * The rule which is used for explanation is : 
 *
 *         likeMyMusic(?genre, ?number) :-
 *              genreILike(?genre),
 *              findall(?person, (
 *                         isPresent(?person, LivingRoom),
 *                         prefers(?person, ?genre)),
 *                      ?similarTastes),
 *              length(?similarTastes, ?number).
 *
 * This rule calculates the number of people who like the same genre as I am in favour of.
 * This will be done for each genre I prefer.
 * 
 * The Rete network for this rule looks like :
 *
 *     FILTER: genreILike           FILTER: isPresent           FILTER: prefers
 *             |                             |                          |
 *              \                    FILTER: LivingRoom                 |
 *               \                           |                          |
 *                \                           \                        /
 *                 \                           \                      /   
 *                  \                          JOIN: ?person = ? person 
 *                   \                                   |
 *                    \                                 /
 *                     \                               / 
 *                      \                             /
 *                           JOIN: ?genre = ?genre
 *                                    |
 *                                    |
 *                                 FINDALL       
 *                                    |
 *                                 LENGTH
 *                                    |
 *                                PRODUCTION
 * </pre> */
public class ProductionNode extends Node {

	/** <pre>
	 * rule_ : the rule of which this node is the production node
	 * engine_ : the actual crime engine
	 * </pre> */
	 private Rule rule_;
	 private CrimeEngine engine_;

	 /** <pre>
	  * Constructor.
	  * </pre> */
	 public ProductionNode(Rule rule, CrimeEngine engine) {
	     rule_ = rule;
	     engine_ = engine;
	 }

	 /** <pre>
	  * Inserting a token leads to inserting a new activation,
	  * a RuleActivation in this case, into the engine's agenda.
	  * </pre> */
	 public void insert(Token token) {
	     RuleActivation activation = new RuleActivation(token, rule_);
	     engine_.insertActivation(activation);
     }
}
