package edu.vub.crime.rete.nodes;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

import edu.vub.crime.grammar.ExpressionFactory;
import edu.vub.crime.grammar.Rule;
import edu.vub.crime.grammar.attributes.constant.Type;
import edu.vub.crime.rete.data.Token;
import edu.vub.crime.rete.filter.FilterFactory;
import edu.vub.util.MultiMap;

public class RootNode extends Node {

	/** 
	 * Maps fact labels onto the filter nodes interested in that type of fact.
	 */
	protected HashMap itsFilterNodes;
	
	/**
	 * Maps {@link Rule} instances to the corresponding {@link ProductionNode} so 
	 * they can be easily undeployed (to wit by unhooking the corresponding nodes).
	 */
	protected HashMap itsProductions;
	
	/**
	 * Keeps track of all Working Memory elements for which there is no corresponding
	 * filter node registered, so that future rules may still use it correctly.
	 */
	protected MultiMap itsWorkingMemory;
	
	/** Constructor. */
    public RootNode() {
    	itsFilterNodes 		= new HashMap();
    	itsProductions		= new HashMap();
    	itsWorkingMemory	= new MultiMap();
    	
    	// Add the "root" type, from which all types derive at startup
    	itsFilterNodes.put("root", NodesFactory.filterNode());
    }
    
    /**
     * Returns a filter node which filters out facts of a particular kind. This filter node
     * can be used as the parent node to attach more specific filters to. If no suitable 
     * filter node exists, one is automatically created and registered. 
     */
    public FilterNode getChild(String kind) {
    	// A. Check whether a corresponding filter node already exists
    	FilterNode filter = (FilterNode) itsFilterNodes.get(kind);
    	
    	// B. If not,
    	if(filter == null) {
    		// 1. Create a new filter node.
    		filter = addChild(kind);
    		
    		// 2. Store it in the map of filter nodes.
    		itsFilterNodes.put(kind, filter);
    		
    		// 3. Copy all facts which match the filter to the newly attached node.
    		Set theElements = (Set)itsWorkingMemory.get(kind);
    		
    		if(theElements != null) {
	    		for(Iterator tokenIt = theElements.iterator(); tokenIt.hasNext(); ) {
	    			Token currentToken = (Token)tokenIt.next();
	    			
	    			filter.insert(currentToken);
	    		}
    		}
    	}
    	
    	// C. Return the (new) filter node
    	return filter;
    }
    
    /**
     * Private method used to create a filter node for a previously unused label.
     */
    private FilterNode addChild(String kind) {
    	// A. Create a filter node which enforces that facts passed to it have the correct label
    	FilterNode theNewFilterNode	= NodesFactory.filterNode();
    	theNewFilterNode.addNewFilter(FilterFactory.filterKind(kind));
    	
    	// B. Pass matching facts on to the filters for all registered parent types
    	//    1. Get the type corresponding to the given label
    	Type thePossibleSubType		= ExpressionFactory.type(kind);
    	
    	//    2. Get the list of supertypes
    	Type[] theSuperTypes		= thePossibleSubType.getParentTypes();
    	
    	//    3. Add all filter nodes for a supertype as a valid child.
    	for (int i = 0; i < theSuperTypes.length; i++) {
			Type theSuperType = theSuperTypes[i];
			FilterNode theFilterNode = getChild(theSuperType.toString());
			
			// Save and restore the parent of the filter node
			Node theStoredParent = theFilterNode.itsParent;
			theFilterNode.itsParent = null;
			
			theNewFilterNode.addChild(theFilterNode);
			theFilterNode.itsParent = theStoredParent;
		}
    	
    	return theNewFilterNode;
    }
     
    /** 
     * Inserts a token containing a single fact into the Rete network. The token will 
     * be dispatched to a child filter node, whose type filter matches the label type
     * of the inserted fact. 
     * <p>
     * Due to the introduction of subtyping, one or more filter nodes may be added on
     * the fly to ensure that inserting a fact labeled with a subtype triggers rules
     * defined for facts labeled with a valid supertype.
     */
    public void insert(Token token) {
        String type = token.getFact(0).getLabel();
    	
    	if(hasMatchingFilter(type)) {
        	Node node = getChild(type);
            node.insert(token);
        }

        itsWorkingMemory.put(type, token);
    } 
    
    private boolean hasMatchingFilter(String type) {
    	if(itsFilterNodes.containsKey(type)) return true;
    	
    	Type 	theFactType 	= ExpressionFactory.type(type);
    	Type[]	theSuperTypes	= theFactType.getParentTypes();
    	
    	boolean foundRelevantSuperType = false;
    	
    	for (int i = 0; i < theSuperTypes.length && !foundRelevantSuperType; i++) {
			Type currentSuperType = theSuperTypes[i];
			
			if(hasMatchingFilter(currentSuperType.toString())) 
				foundRelevantSuperType = true;
		}
    	
    	return foundRelevantSuperType;
    }
    
    public void unhook(Node theChildNode) {
    	super.unhook(theChildNode);
    	
    	itsFilterNodes.values().remove(theChildNode);
    }
    
    public void deployed(Rule theRule, ProductionNode theNode) {
    	itsProductions.put(theRule, theNode);
    }
    
    public void undeploy(Rule theRule) {
    	Node theNode = (Node)itsProductions.remove(theRule);
    	
    	theNode.unhook(null);
    }
    
	public void initializeChild(Node theNewNode) {}

}
