package edu.vub.crime.rete.nodes.aux;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import java.util.Vector;

import edu.vub.crime.dataflow.DataFlowAttribute;
import edu.vub.crime.grammar.attributes.Attribute;
import edu.vub.crime.grammar.attributes.CompoundAttribute;
import edu.vub.crime.grammar.attributes.ConstantAttribute;

/**
 * Defines a concrete instance of a dataflow attribute for the purposes of testing the response of
 * filter nodes to such values.
 */
public class TestFlowAttribute extends DataFlowAttribute {
	
	private Attribute itsCurrentValue;

	public TestFlowAttribute(Attribute theInitialValue) {
		itsCurrentValue = theInitialValue;
	}
	
	public Attribute getCurrentValue() {
		return itsCurrentValue;
	}
	
	public void setCurrentValue(Attribute theNewValue) {
		// Dataflow attribute values can not be variables, nor compound structures containing variables.
		
		// This restriction is due to the fact that changes to the dataflow attribute's value are handled
		// by reevaluating the filte node containing the dataflow, rather than recompiling the rule. As a
		// consequence, the removal or introduction of variables cannot be handled correctly.
		
		// - Removing a variable that was previously stored in a dataflow attribute requires removing join
		//   clauses in dependent nodes of the network which refer to this variable. If this update cannot
		//   be performed (as is the case with the current implementation), the rule is overly restrictive
		//   and can reject valid matches.
		
		// - Adding new variables by setting the value of a dataflow attribute implies adding join clauses
		//   joining the variable with other occurrences in the rule.  If not performed properly, the rule
		//   becomes overly liberal, by passing on invalid tokens.
		
		// Note that CRIME supports dynamic addition and removal of rules, which permits recompiling rules
		// as described above. However, dataflow attributes are only conceived to deal with bound values.
		
		if(	theNewValue instanceof ConstantAttribute || 
			(	(theNewValue instanceof CompoundAttribute) && 
				((CompoundAttribute)theNewValue).getVariables().isEmpty())) {

			itsCurrentValue = theNewValue;
			
			// While propagating updates, new dependents may register. We do not consider them during this round.
			Vector theAlreadyRegisteredSubscribers = new Vector(itsSubscribers);
			
			for (Iterator subscriberIt = theAlreadyRegisteredSubscribers.iterator(); subscriberIt.hasNext();) {
				ActionListener currentSubscriber = (ActionListener) subscriberIt.next();
				
				currentSubscriber.actionPerformed(new ActionEvent(this, ActionEvent.ACTION_PERFORMED, "new value assigned to dataflow attribute"));
			}
			
		} else {
			throw new RuntimeException("Dataflow variables may only contain grounded attributes without any variables");					
		}
	}
	
	public boolean equals(Object o) {
		return itsCurrentValue.equals(o);
	}
	
	public void setIndex(Vector index) {
		super.setIndex(index);
		
		itsCurrentValue.setIndex(index);
	}
}
